/* FCHT.h
 * Copyright (C) 2010, Rodrigo Canovas & Miguel A. Martinez-Prieto
 * all rights reserved.
 *
 * FrontCoding Compression on Hu-Tucker Encoding
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *
 * Contacting the author:
 *   Rodrigo Canovas:  yigorc@gmail.com
 *   Miguel A Martinez-Prieto:  migumar2@infor.uva.es
 */


#include "FCHT.h"

FCHT::FCHT()
{
	tree = NULL;
	HTtree = NULL;
		
	delta = new uchar[2];
	delta[0] = (uchar)0;
	delta[1] = (uchar)0;
}
			
FCHT::FCHT(char *filename, size_t blocksize)
{		
	ifstream in(filename);
	if(in.good())
	{
		in.seekg(0,ios_base::end);
		size_t len = in.tellg();
		in.seekg(0,ios_base::beg);
		
		// Initialising values
		this->length = 0;
		this->max_len = 0;
			
		vector<uint> xblocks;
		this->blocksize = blocksize;
		this->nblocks = 0;

		this->uchars = 0;
		this->delta = NULL;
		this->HTtree = NULL;
		this->qstring = NULL;
			
		uint tsize = len/10 +1000;
		uchar *ctext = (uchar*)malloc(tsize*sizeof(uchar));
		uint chars = 0;
			
		string previous, current;
		uint lprev, lcurr;
		
		while (getline(in, current))
		{	
			lcurr = current.length();				
			if ((lcurr - 1) > max_len) max_len = lcurr;
			
			if ((length % blocksize) == 0)
			{					
				// This is the first element in the block: the full 
				// word is written in the text.					
				xblocks.push_back(chars);
				nblocks++;
				
				// Checking text size: realloc if necessary
				if (chars+lcurr+3 >= tsize)
				{
					tsize += len/10;
					ctext = (uchar*)realloc(ctext, tsize*sizeof(uchar));
				}
				
				chars += VByte_F::encode(lcurr, ctext+chars);
				
				for (uint i=0; i<lcurr; i++)
				{
					ctext[chars] = (uchar)current[i];
					chars++;
				}
			}
			else
			{
				// The word is encoded with the delta offset and the
				// different substring with respect to the previous
				// word.					
				uint delta = lcp((uchar*)(previous.c_str()), (uchar*)(current.c_str()), lprev, lcurr);
				
				// Checking text size: realloc if necessary
				// Hay un exceso de 2 considerando que el delta 
				// ocupa, a lo sumo, 2 bytes => revisar!
				if (chars+lcurr-delta+3 >= tsize)
				{
					tsize += len/10;
					ctext = (uchar*)realloc(ctext, tsize*sizeof(uchar));
				}
				
				chars += VByte_F::encode(delta, ctext+chars);
			
				for (uint i=delta; i<lcurr; i++)
				{
					ctext[chars] = (uchar)current[i];
					chars++;
				}
			}
			
			ctext[chars] = '\0';
			chars++;
			
			previous = current;
			lprev = lcurr;
			
			length++;
		}			
		
		// Puts the chars value in the last position of the blocks
		// vector
		xblocks.push_back(chars);
		in.close();
		
		// Calculating symbol frequencies (all values initialised to 1)
		int freqs[256];
		for (uint i=0; i<256; i++) freqs[i] = 1;
		for (uint i=0; i<chars; i++) freqs[(int)(ctext[i])]++;			

		HuTucker<int> ht(freqs,256);
		leafsHT = ht.getCodes(&HTcode, &tree);
					
		// Initialising the text structure
		tsize = len/10;
		this->text = (uchar*)malloc(tsize*sizeof(uchar));
		for (uint i=0; i<tsize; i++) text[i] = 0;
		
		// HuTucker encoding
		uint offset = 0, cblocks = 0;
		bool first = true;
		
		uint i = 0;
		
		while (i < chars)
		{				
			// Checking text size: realloc if necessary
			if (uchars+10 > tsize)
			{	
				tsize += len/10;
				text = (uchar*)realloc(text, tsize*sizeof(uchar));
					
				for (uint j=uchars+1; j<tsize; j++) text[j] = 0;
			}
			
			if (i == xblocks[cblocks])
			{
				first = true;
				
				// New Block: reseting 'chars' and 'offset' to obtain
				// byte-aligned blocks (if offset > 0)
				if (offset > 0)
				{					
					uchars++;
					offset = 0;
				}
				
				xblocks[cblocks] = uchars;
				cblocks++;
				
				// Me "salto" el VByte...
				while (true)
				{
					text[uchars] = ctext[i];
					i++;
					uchars++;
					
					if (ctext[i-1] > 128) break;
				}
			}
			
			encodeHT(HTcode[(int)ctext[i]].code, HTcode[(int)ctext[i]].cbits, &offset);
			
			if ((first) && (ctext[i] == '\0'))
			{
				first = false;
				
				if (offset > 0)
				{					
					uchars++;
					offset = 0;
				}
			}
			
			i++;
		}
		
		// Puts the chars value in the last position of the blocks
		// vector
		uchars++;
		xblocks[cblocks] = uchars;
		cblocks++;			
		
		free(ctext);
		
		// Transforming blocks vector into an array of log(uchars)
		// bits per element
		blocks = new Array_F(xblocks, bits_F(uchars));
	}
	else
	{
		abort();
	}
}

FCHT::FCHT(uchar ** words, uint n, size_t blocksize)
{		
	if(n>0)
	{
		size_t len = n;
		
		// Initialising values
		this->length = 0;
		this->max_len = 0;
			
		vector<uint> xblocks;
		this->blocksize = blocksize;
		this->nblocks = 0;

		this->uchars = 0;
		this->delta = NULL;
		this->HTtree = NULL;
		this->qstring = NULL;
			
		uint tsize = len +1000;
		uchar *ctext = (uchar*)malloc(tsize*sizeof(uchar));
		uint chars = 0;
			
		string previous, current;
		uint lprev, lcurr;
		
		//while (getline(in, current))
		uint ii;
		for (ii=0;ii<n;ii++)
		{	current = string((const char *)words[ii]);
			lcurr = current.length();		
			if ((lcurr - 1) > max_len) max_len = lcurr;
			
			if ((length % blocksize) == 0)
			{					
				// This is the first element in the block: the full 
				// word is written in the text.					
				xblocks.push_back(chars);
				nblocks++;
				
				// Checking text size: realloc if necessary
				if (chars+lcurr+3 >= tsize)
				{
					tsize += len;
					ctext = (uchar*)realloc(ctext, tsize*sizeof(uchar));
				}
				
				chars += VByte_F::encode(lcurr, ctext+chars);
				
				for (uint i=0; i<lcurr; i++)
				{
					ctext[chars] = (uchar)current[i];
					chars++;
				}
			}
			else
			{
				// The word is encoded with the delta offset and the
				// different substring with respect to the previous
				// word.					
				uint delta = lcp((uchar*)(previous.c_str()), (uchar*)(current.c_str()), lprev, lcurr);
				
				// Checking text size: realloc if necessary
				// Hay un exceso de 2 considerando que el delta 
				// ocupa, a lo sumo, 2 bytes => revisar!
				if (chars+lcurr-delta+3 >= tsize)
				{
					tsize += len;
					ctext = (uchar*)realloc(ctext, tsize*sizeof(uchar));
				}
				
				chars += VByte_F::encode(delta, ctext+chars);
			
				for (uint i=delta; i<lcurr; i++)
				{
					ctext[chars] = (uchar)current[i];
					chars++;
				}
			}
			
			ctext[chars] = '\0';
			chars++;
			
			previous = current;
			lprev = lcurr;
			
			length++;
		}			
		
		// Puts the chars value in the last position of the blocks
		// vector
		xblocks.push_back(chars);
		
		// Calculating symbol frequencies (all values initialised to 1)
		int freqs[256];
		for (uint i=0; i<256; i++) freqs[i] = 1;
		for (uint i=0; i<chars; i++) freqs[(int)(ctext[i])]++;			

		HuTucker<int> ht(freqs,256);
		leafsHT = ht.getCodes(&HTcode, &tree);
			
		// Initialising the text structure
		tsize = len+1000;
		this->text = (uchar*)malloc(tsize*sizeof(uchar));
		for (uint i=0; i<tsize; i++) text[i] = 0;
		
		// HuTucker encoding
		uint offset = 0, cblocks = 0;
		bool first = true;
		
		uint i = 0;
		
		while (i < chars)
		{				
			// Checking text size: realloc if necessary
			if (uchars+10 > tsize)
			{	
				tsize += len;
				text = (uchar*)realloc(text, tsize*sizeof(uchar));
					
				for (uint j=uchars+1; j<tsize; j++) text[j] = 0;
			}
			
			if (i == xblocks[cblocks])
			{
				first = true;
				
				// New Block: reseting 'chars' and 'offset' to obtain
				// byte-aligned blocks (if offset > 0)
				if (offset > 0)
				{					
					uchars++;
					offset = 0;
				}
				
				xblocks[cblocks] = uchars;
				cblocks++;
				
				// Me "salto" el VByte...
				while (true)
				{
					text[uchars] = ctext[i];
					i++;
					uchars++;
					
					if (ctext[i-1] > 128) break;
				}
			}
			
			encodeHT(HTcode[(int)ctext[i]].code, HTcode[(int)ctext[i]].cbits, &offset);
			
			if ((first) && (ctext[i] == '\0'))
			{
				first = false;
				
				if (offset > 0)
				{					
					uchars++;
					offset = 0;
				}
			}
			
			i++;
		}
		
		// Puts the chars value in the last position of the blocks
		// vector
		uchars++;
		xblocks[cblocks] = uchars;
		cblocks++;			
		
		free(ctext);
		
		// Transforming blocks vector into an array of log(uchars)
		// bits per element
		blocks = new Array_F(xblocks, bits_F(uchars));
	}
	else
	{
		abort();
	}
}




char 
FCHT::decodeHT(uchar *seq, uint *pos, uint *offset)
{
	uint node = 0;
	
	while (HTtree[node].symbol < 0)
	{
		bool bit = ((seq[*pos] >> (7-(*offset))) & 1);
		node = HTtree[node].children[bit];

		(*offset)++;
		
		if ((*offset) == 8)
		{				
			(*pos)++;
			(*offset) = 0;
		}
	}
	
	return (char)HTtree[node].symbol;
}

char
FCHT::decodeHT(uint *pos, uint *offset)
{
	uint node = 0;
	
	while (HTtree[node].symbol < 0)
	{
		bool bit = ((text[*pos] >> (7-(*offset))) & 1);
		node = HTtree[node].children[bit];

		(*offset)++;
		
		if ((*offset) == 8)
		{				
			(*pos)++;
			(*offset) = 0;
		}
	}
	
	return (char)HTtree[node].symbol;
}

void
FCHT::decompressWord(uint *pos, uint*offset, uint ptr)
{
	do
	{
		qstring[ptr] = decodeHT(pos, offset);
		ptr++;
	} 
	while (qstring[ptr-1] != '\0');
}

void
FCHT::decompressDelta(uint *pos, uint*offset)
{
	uint i = 0;
			
	do
	{
		delta[i] = (uchar)decodeHT(pos, offset);			
		i++;
	}
	while (delta[i-1] < 128);
}

void 
FCHT::encodeHT(uint code, uint len, uchar *seq, uint *pos, uint *offset)
{
	uchar uccode;
	uint uicode;	
	uint processed = 0;
	
	while ((len-processed) >= (8-(*offset)))
	{			
		uicode = code << (W_F-len+processed);
		uccode = (uchar)(uicode >> (W_F-(8-(*offset))));
		seq[*pos] = seq[*pos] | uccode;
		
		processed += 8-(*offset);
		(*pos)++;
		seq[*pos] = 0;
		(*offset) = 0;			
	}
	
	if (len-processed > 0)
	{
		uicode = code << (W_F-len+processed);
		uccode = (uchar)(uicode >> (W_F-(8-(*offset))));
		seq[*pos] = seq[*pos] | uccode;
		(*offset) += len-processed;
	}
}

void
FCHT::encodeHT(uint code, uint len, uint *offset)
{
	uchar uccode;
	uint uicode;	
	uint processed = 0;
	
	while ((len-processed) >= (8-(*offset)))
	{
		// Desplazo hacia la izquierda para "sacar fuera" los bits
		// ya procesados...
		uicode = code << (W_F-len+processed);
			
		// Desplazo hacia la derecha para quedarme con los que quiero
		uccode = (uchar)(uicode >> (W_F-(8-(*offset))));
			
		// Hago el OR de hcode con la posición actual
		text[uchars] = text[uchars] | uccode;
		
		processed += 8-(*offset);
		uchars++;
		(*offset) = 0;			
	}
		
	if (len-processed > 0)
	{
		uicode = code << (W_F-len+processed);
		uccode = (uchar)(uicode >> (W_F-(8-(*offset))));
		text[uchars] = text[uchars] | uccode;
		(*offset) += len-processed;
	}
}

size_t 
FCHT::locate(uchar *word, size_t len)
{
	// Binary searching for the block
	uint block;
	bool cmp = locateBlock(word, &block);
		
	if (cmp)
	{
		// The word is located at the first position of the block
		return (block*blocksize)+1;
	}
	else
	{
		// The block is sequentially scanned to find the word
		uint idblock = locateWordInBlock(block, word);
			
		// If idblock = 0, the word is not in the dictionary
		if (idblock != 0) return (block*blocksize)+idblock+1;
		else return idblock;
	}
		
	return 0;
}

uchar* 
FCHT::extract(size_t id_word)
{					
	if ((id_word > 0) && (id_word <= length))
	{					
		// Identifying the block and the offset inside it
		uint block = (id_word-1)/blocksize;
		uint offset = (id_word-1)%blocksize;
		
		return locateWordInBlock(block, offset);
	}
	else
	{
		return NULL;
	}
}

size_t
FCHT::getLength()
{
	return length;
}

size_t 
FCHT::getSize()
{
	return uchars*sizeof(uchar) + blocks->getSize() + tree->getSize() + (tree->getLength()/2)*sizeof(uint) + leafsHT*sizeof(Tcode) + sizeof(FCHT);
}

void 
FCHT::save(ofstream & fp)
{		
	saveValue<size_t>(fp, length);
	saveValue<size_t>(fp, max_len+1);
		
	saveValue<size_t>(fp, uchars);
	saveValue<uchar>(fp, text, uchars);		
		
	saveValue<size_t>(fp, blocksize);
	saveValue<size_t>(fp, nblocks);
	blocks->save(fp);
	tree->save(fp);
	
	saveValue<uint>(fp, leafsHT);
	for (uint i=0; i<leafsHT; i++)
	{
		saveValue<uint>(fp, HTcode[i].code);
		saveValue<uint>(fp, HTcode[i].cbits);
	}
}

FCHT* 
FCHT::load(ifstream & fp)
{						
	FCHT *dicc = new FCHT();
		
	dicc->length = loadValue<size_t>(fp);
	dicc->max_len = loadValue<size_t>(fp);
	
	dicc->uchars = loadValue<size_t>(fp);
	dicc->text = loadValue<uchar>(fp, dicc->uchars);
		
	dicc->blocksize = loadValue<size_t>(fp);
	dicc->nblocks = loadValue<size_t>(fp);		
	dicc->blocks = new Array_F(fp);
	dicc->tree = new BitString_F(fp);
		
	// Building HTtree
	dicc->HTtree = new Node[dicc->tree->getLength()/2];

	vector<uint> traversing;
	uint node = 0, symbol = 0;
		
	for (uint i=0; i<dicc->tree->getLength(); i++)
	{						
		if (dicc->tree->getBit(i) == 0)
		{				
			dicc->HTtree[node].children[0] = -1;
			dicc->HTtree[node].children[1] = -1;
			dicc->HTtree[node].symbol = 0;
				
			if (traversing.size() > 0)
			{
				uint parent = traversing[traversing.size()-1];
						
				if (dicc->HTtree[parent].symbol == 0) dicc->HTtree[parent].children[0] = node;
				else dicc->HTtree[parent].children[1] = node;

				dicc->HTtree[parent].symbol++;
			}
			
			traversing.push_back(node);
			node++;
		}
		else
		{
			uint last = traversing[traversing.size()-1];
			traversing.pop_back();
				
			if (last == (node-1))
			{
				dicc->HTtree[last].symbol = symbol;
				symbol++;
			}
			else
			{
				dicc->HTtree[last].symbol = -1;
			}
		}
	}
	
	dicc->leafsHT = loadValue<uint>(fp);
	dicc->HTcode = new Tcode[dicc->leafsHT];
	
	for (uint i=0; i<dicc->leafsHT; i++)
	{
		dicc->HTcode[i].code = loadValue<uint>(fp);
		dicc->HTcode[i].cbits = loadValue<uint>(fp);
	}
	
	dicc->qstring = new uchar[dicc->max_len];	
		
	return dicc;
}

uint 
FCHT::lcp(uchar* str1, uchar* str2, uint lstr1, uint lstr2)
{
	uint delta = 0;
	
	uint length = lstr1;
	if (length > lstr2) length = lstr2;
	
	for (uint i=0; i<length; i++)
	{
		if (str1[i] == str2[i]) delta++;
		else break;
	}
	
	return delta;
}

		
bool 
FCHT::locateBlock(uchar *word, uint *block)
{
	// Codificamos el patrón
	uint plength = strlen((char*)word), dlength;
	uchar *encoded = new uchar[2*plength];
		
	uint encpos = 0;
	uint encoffset = 0;
	encoded[encpos] = 0;
		
	for (uint i=0; i<plength; i++) encodeHT(HTcode[word[i]].code, HTcode[word[i]].cbits, encoded, &encpos, &encoffset);
	encpos++;
		
	uint l = 0, r = nblocks-1, c;
	int cmp;
		
	while (l <= r)
	{
		c = (l+r)/2;
		uint pos = blocks->getField(c);
			
		// Leemos la longitud del primer elemento
		pos += VByte_F::decode(&dlength, text+pos);			
		cmp = memcmp((text+pos), encoded, encpos);
			
		if (cmp > 0)
		{
			// The given word is in any preceding block
			r = c-1;
		}
		else
		{
			if (cmp < 0)
			{
				// The given word is in any subsequent block
				l = c+1;
			}
			else
			{
				if (plength == dlength)
				{					
					// The given word is the first one in the c-th block
					*block = c;	
	
					delete [] encoded;			
					return true;
				}
				else
				{
					r = c-1;
				}
			}
		}
	}
	
	// If (cmp < 0) the word is in the current block (c)
	// If (cmp > 0) the word is in the preceding block (c-1)
	if (cmp < 0) *block = c;
	else *block = c-1;
	
	delete [] encoded;
	return false;
}

uchar*
FCHT::locateWordInBlock(uint block, uint offset)
{				
	// Extracting the first word
	uint ptr = 0;
	uint pos = blocks->getField(block), off = 0;
		
	// Salto el delta
	do { pos++; } while (text[pos-1] < 128);
		
	decompressWord(&pos, &off, ptr);
	
	if (off != 0)
	{
		pos++;
		off=0;
	}
		
	for (uint j=0; j<offset; j++)
	{
		// Reading delta
		decompressDelta(&pos, &off);
		VByte_F::decode(&ptr, delta);
							
		// Reading the suffix
		decompressWord(&pos, &off, ptr);
	}
		
	return qstring;
}
			
uint 
FCHT::locateWordInBlock(uint block, uchar *word)
{
	uint pos = blocks->getField(block);
	uint end = blocks->getField(block+1);
		
	uint clcp_len = 0, plcp_len = 0;
	uint string_len = 0;
	uint word_len = strlen((char*)word);
		
	uint idblock = 0;
	
	// Extracting the first word
	uint ptr = 0, off = 0;
		
	// Salto el delta
	do { pos++; } while (text[pos-1] < 128);
			
	decompressWord(&pos, &off, ptr);
	string_len = strlen((char*)qstring);
	clcp_len = lcp(qstring, word, string_len, word_len);
		
	if (off != 0)
	{
		pos++;
		off=0;
	}
	
	// Scanning the block until the word is reached. If it is not in
	// the dictionary, the process finishes 1) when the block ends 
	// or 2) when a subsequent word is reached.
	while (pos < end)
	{					
		idblock++;
		
		if (idblock == blocksize) break;
		
		// Reading delta
		decompressDelta(&pos, &off);
		VByte_F::decode(&ptr, delta);
		
		// Reading the suffix
		decompressWord(&pos, &off, ptr);
		string_len = strlen((char*)qstring);

		if (ptr >= clcp_len)
		{
			// Current delta is greater than the current long common
			// preffix: 'clcp' is updated!
			plcp_len = clcp_len;
			clcp_len += lcp(qstring+clcp_len, word+clcp_len, string_len-clcp_len, word_len-clcp_len);
		}
		else
		{
			return 0;
		}

		// This is the required word
		if ((clcp_len == word_len) && (word_len == string_len))
		{
			return idblock;
		}

		// The word is not in the dictionary
		if (clcp_len < plcp_len) return 0;
	}
		
	return 0;
}
			
FCHT::~FCHT()
{
	delete blocks;
	
	if (qstring != NULL)
	{
		delete [] text;
		delete [] qstring;
	}
	else
	{
		free(text);
	}
	
	if (delta != NULL) delete [] delta;
	if (tree != NULL) delete tree;
	if (HTtree != NULL) delete [] HTtree;
	if (HTcode != NULL) delete [] HTcode;
}
