/* buildFacade.c
 * Copyright (C) 2011, Antonio Fariña, all rights reserved.
 *
 * BuildFacade.c: Implementation of the functions that the word-based self-index
 *   must implement following the "../pizza/interface.h" guidelines.
 *
 * See more details in:
 * Antonio Fariña, Nieves Brisaboa, Gonzalo Navarro, Francisco Claude, Ángeles Places, 
 * and Eduardo Rodríguez. Word-based Self-Indexes for Natural Language Text. ACM 
 * Transactions on Information Systems (TOIS), 2012. 
 * http://vios.dc.fi.udc.es/indexing/wsi/publications.html
 * http://www.dcc.uchile.cl/~gnavarro/ps/tois11.pdf	  
 * 
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */


#include "buildFacade.h"
#include "utils/errors.c"
#include "utils/canonicalsHuff.c"
#include "utils/variants.c"
#include "utils/separators.c"
#include "utils/huffVariants.c"

double getTime2 (void)
{
	double usertime, systime;
	struct rusage usage;

	getrusage (RUSAGE_SELF, &usage);

	usertime = (double) usage.ru_utime.tv_sec +
		(double) usage.ru_utime.tv_usec / 1000000.0;
	systime = (double) usage.ru_stime.tv_sec +
		(double) usage.ru_stime.tv_usec / 1000000.0;

	return (usertime + systime);
}



/** Building the index */

    /* Creates index from text[0..length-1]. Note that the index is an 
      opaque data type. Any build option must be passed in string 
      build_options, whose syntax depends on the index. The index must 
      always work with some default parameters if build_options is NULL. 
      The returned index is ready to be queried. */

int build_index (uchar *text, ulong length, char *build_options, void **index) {
	 //return build_WCSA (text, length, stopwordsfile, NULL, build_options);	

	int returnvalue;

     printf("\n parameters: \"%s\"\n",build_options); fflush(stderr);
        
    returnvalue = build_WCSA (text, length, build_options, index);
 
    if (!returnvalue) 
    	returnvalue = build_iCSA (build_options,*index);
    		   
    return returnvalue;    
}


/**  Saves index on disk by using single or multiple files, having 
	proper extensions. */
int save_index (void *index, char *filename) {
	char *basename = filename;
	twcsa *wcsa=(twcsa *) index;

	uint i,j;
	char *outfilename;
	int file;
	char c;

	printf("\n\t Saving structures to disk: %s.*",filename);			
	outfilename = (char *)malloc(sizeof(char)*(strlen(basename)+10));

	/**File with some constants (bSize and tohSize); */
	{ uint number;
		strcpy(outfilename, basename);
		strcat(outfilename, ".");
		strcat(outfilename, CONSTANTS_FILE_EXT);
		unlink(outfilename);
		if( (file = open(outfilename, O_WRONLY|O_CREAT,S_IRWXG | S_IRWXU)) < 0) {
			printf("Cannot open file %s\n", outfilename);
			exit(0);
		}	
		write(file, &(wcsa->tohSize), sizeof(uint));
		write(file, &(wcsa->sourceTextSize), sizeof(uint));		
		write(file, &(wcsa->seSize), sizeof(uint));				
		write(file, &(wcsa->samplePeriodB), sizeof(uint));			
		write(file, &(wcsa->samplePeriodZ), sizeof(uint));	
#ifdef COMPRESS_FCHT_CANONICALS
		uint tempval = wcsa->blocksizeFCHuTucker;
		write(file, &tempval, sizeof(uint));	
#endif			

				
		close(file);		
	}	

	/** A copy of the stopwords-list */
	{  char *filename;	
		filename = (char *)malloc(sizeof(char) * (strlen(basename)+10));
		{ 	strcpy(filename, basename);
			strcat(filename, ".");
			strcat(filename, STOPWORDS_FILE_EXT);
		}
		copyFile(wcsa->stopwordsfilename,filename);
		free(filename);
	}	


/** The canonical words, their variants and huffman trees */
	/** File with the canonical words. (sorted alphanumerically) */

#ifdef KEEP_CANONICALS_UNCOMPRESSED

	/** The Words in the vocabulary of words  (sorted alphabetically)*/
	{	strcpy(outfilename, basename);
		strcat(outfilename, ".");
		strcat(outfilename, CANONICAL_WORDS_FILE_EXT);
		unlink(outfilename);
		if( (file = open(outfilename, O_WRONLY|O_CREAT,S_IRWXG | S_IRWXU)) < 0) {
			printf("Cannot open file %s\n", outfilename);
			exit(0);
		}

		uint n = wcsa->numberOfCanonicalWords;
		uint elemSize = wcsa->wordsData.elemSize;
		write(file, &n, sizeof(uint));
		write(file, &elemSize, sizeof(uint));
		write(file, &(wcsa->wordsData.wordsZoneMem.size), sizeof(uint));
		
		//the number of canonical words
		write(file, (char *)wcsa->wordsData.wordsZoneMem.zone, wcsa->wordsData.wordsZoneMem.size * sizeof(byte));
		write(file, (char *)wcsa->wordsData.words, ((((n+1)* (elemSize))+W-1) /W) * (sizeof(uint)) );

		close(file);
	}	
#endif
#ifdef COMPRESS_HUFFMAN_CANONICALS
	/** File with the canonical words (sorted by alphabetically)... compressed with huffman.*/
	{
		strcpy(outfilename, basename);
		strcat(outfilename, ".");
		strcat(outfilename, CANONICAL_WORDS_FILE_EXT2);
		unlink(outfilename);
		if( (file = open(outfilename, O_WRONLY|O_CREAT,S_IRWXG | S_IRWXU)) < 0) {
			printf("Cannot open file %s\n", outfilename);
			exit(0);
		}	
		
		write(file, &(wcsa->numberOfCanonicalWords), sizeof(uint));				
		saveCanonicals(&(wcsa->HC),file);		
		close(file);	
	}
#endif

#ifdef COMPRESS_FCHT_CANONICALS
	/** File with the canonical words (sorted by alphabetically)... compressed with huffman.*/
	{
		strcpy(outfilename, basename);
		strcat(outfilename, ".");
		strcat(outfilename, CANONICAL_WORDS_FILE_EXT3);
		unlink(outfilename);
		saveCanonicalsFCHT(wcsa->fcht,(char*)outfilename);
	}
#endif



	/** File with the variants of each canonical word. (sorted by frequency)*/
	{
		strcpy(outfilename, basename);
		strcat(outfilename, ".");
		strcat(outfilename, VARIANTS_OF_WORDS_FILE_EXT2);
		unlink(outfilename);
		if( (file = open(outfilename, O_WRONLY|O_CREAT,S_IRWXG | S_IRWXU)) < 0) {
			printf("Cannot open file %s\n", outfilename);
			exit(0);
		}	

		saveVariants(&(wcsa->variants),file);		
		close(file);
	
	}

	/**  File with the data of the Huffman trees of each canonical word (and variants) */
	{ uint depth;
		strcpy(outfilename, basename);
		strcat(outfilename, ".");
		strcat(outfilename, HUFFWORD_CANONICAL_FILE_EXT);
		unlink(outfilename);
		if( (file = open(outfilename, O_WRONLY|O_CREAT,S_IRWXG | S_IRWXU)) < 0) {
			printf("Cannot open file %s\n", outfilename);
			exit(0);
		}	
		
		//saving all the depths first
		saveHuffVariants(&(wcsa->HV), file);
		close(file);		
	}

	/** The separators and their huffman tree (compacted) */
	// File with the separators. (sorted by frequency)  
	{ uint number,depth;
		strcpy(outfilename, basename);
		strcat(outfilename, ".");
		strcat(outfilename, ASCII_OF_SEPARATORS_FILE_EXT);
		unlink(outfilename);
		if( (file = open(outfilename, O_WRONLY|O_CREAT,S_IRWXG | S_IRWXU)) < 0) {
			printf("Cannot open file %s\n", outfilename);
			exit(0);
		}	
		
		number= wcsa->separators2.count;
		write(file, &number, sizeof(uint)); //the number of separators
		//the depth of the huffman tree
		depth = wcsa->separators2.huffman.depth;
		write(file, &depth, sizeof(uint));
		//the num and fst arrays.
		write(file, wcsa->separators2.huffman.num, sizeof(uint) * (depth+1));
		write(file, wcsa->separators2.huffman.fst, sizeof(uint) * (depth+1));
		saveSeparators(&(wcsa->separators2),file);	
		close(file);		
	}

	/** TOH: Compressed text  */
	{	strcpy(outfilename, basename);
		strcat(outfilename, ".");
		strcat(outfilename, TOH_FILE_EXT);
		unlink(outfilename);
		if( (file = open(outfilename, O_WRONLY|O_CREAT,S_IRWXG | S_IRWXU)) < 0) {
			printf("Cannot open file %s\n", outfilename);
			exit(0);
		}
		write(file, wcsa->TOH, sizeof(uint) * (((wcsa->tohSize)+W-1)/W));
		close(file);
	}	

	// B: Samples of offsets in TOH where the code of the ith valid word appears
	{	strcpy(outfilename, basename);
		strcat(outfilename, ".");
		strcat(outfilename, B_COMPR_FILE_EXT);
		unlink(outfilename);
		if( (file = open(outfilename, O_WRONLY|O_CREAT,S_IRWXG | S_IRWXU)) < 0) {
			printf("Cannot open file %s\n", outfilename);
			exit(0);
		}
		write(file, wcsa->B, sizeof(uint) * (wcsa->bSize));       //the samples
		close(file);
	}	

	// Z: Samples of the offset in the original text of each integer in SE
	{	strcpy(outfilename, basename);
		strcat(outfilename, ".");
		strcat(outfilename, Z_SAMPLES_FILE_EXT);
		unlink(outfilename);
		if( (file = open(outfilename, O_WRONLY|O_CREAT,S_IRWXG | S_IRWXU)) < 0) {
			printf("Cannot open file %s\n", outfilename);
			exit(0);
		}
		write(file, wcsa->Z, sizeof(uint) * (wcsa->zSize));       //the samples
		close(file);
	}					
	free(outfilename);

/*
	#ifdef CSA_ON
	// ******** saves index on integers (bottom) ******
	//Storing the CSA
	//storeStructsCSA(wcsa->myicsa,basename);
	saveIntIndex((void *) wcsa->myicsa, basename);
	#endif

	#ifndef CSA_ON
	saveSEfile(basename,wcsa->se, wcsa->seSize+1);
	free(wcsa->se);
	#endif
*/
	if (wcsa->myicsa) {
		/******** saves index on integers (bottom) ******/
		//Storing the CSA
		//storeStructsCSA(wcsa->myicsa,basename);
		saveIntIndex((void *) wcsa->myicsa, basename);		
	}	
	
	if (wcsa->se) {
		saveSEfile(basename,wcsa->se, wcsa->seSize+1);
		//free(wcsa->se);		
	}

	return 0;
}


    /**  Loads index from one or more file(s) named filename, possibly 
      adding the proper extensions. */
int load_index(char *filename, void **index){
	twcsa *wcsa;
	wcsa = loadWCSA (filename);
	(*index) = (void *) wcsa;
	return 0;
}

	/** Frees the memory occupied by index. */
int free_index(void *index){
	twcsa *wcsa=(twcsa *) index;
	ulong size;
	index_size(index,&size);
	printf("\n\t[destroying index] ...Freed %lu bytes... RAM", size);	

	stdestroy_hash(wcsa->stopwordshash);
	if (wcsa->stopwordsfilename) free (wcsa->stopwordsfilename);

	//frees the array SE.
	if (wcsa->se)
		free (wcsa->se);

	//destroyStructsCSA(wcsa->myicsa);
	if (wcsa->myicsa) {
		int err = freeIntIndex((void *) wcsa->myicsa);
	}
	
	destroyStructsFacade(wcsa);			
	free(wcsa);
	return 0;
}




/** Gives the memory occupied by index in bytes. */
int index_size(void *index, ulong *size) {
	twcsa *wcsa=(twcsa *)index;
	*size=0;
	*size += sizeof(twcsa);
	*size += stmemused(wcsa->stopwordshash);
	
	uint totaltmp=0;  //separators
	totaltmp += 2 * ((wcsa->separators2.huffman.depth+1)*sizeof(uint));  //Huffman tree of separators
	totaltmp += sizeSeparators(&(wcsa->separators2));
	*size += totaltmp;

	totaltmp=0;

#ifdef KEEP_CANONICALS_UNCOMPRESSED	
	//words: canonical, variants and huffman trees
	//totaltmp += sizeof(twords) * (wcsa->numberOfCanonicalWords);
	//totaltmp += wcsa->canonicalsZoneMem.size;
	totaltmp += ((((wcsa->numberOfCanonicalWords+1)*
								 (wcsa->wordsData.elemSize))+W-1) /W) * (sizeof(uint));  //the pointers
	totaltmp += wcsa->wordsData.wordsZoneMem.size * sizeof(byte); //the characters of the words.
	fflush(stdout);
	//fprintf(stderr,"\n \n *** canonicals size = %d bytes **\n",totaltmp);fflush(stderr);
	*size += totaltmp;		
#endif
#ifdef COMPRESS_HUFFMAN_CANONICALS	
	//canonical words compressed with huffman.
	totaltmp = sizeCanonicals(&(wcsa->HC)); 
	*size +=totaltmp;
#endif

#ifdef COMPRESS_FCHT_CANONICALS	
	//canonical words compressed with FTHC.
	totaltmp = sizeCanonicalsFCHT(wcsa->fcht); 
	*size +=totaltmp;
#endif
	
	//variants
	totaltmp=0;
	totaltmp += sizeVariants(&(wcsa->variants)); /** variants structure compacted*/
	totaltmp += sizeHuffVariants(&(wcsa->HV));   /**Huffman of variants */
	*size += totaltmp;

	*size += sizeof(uint) * (((wcsa->tohSize)+W-1)/W);
	*size += sizeof(uint) * (wcsa->bSize );
	*size += sizeof(uint) * (wcsa->zSize );
	
	if (wcsa->myicsa) {
		uint nbytes;
		int err = sizeIntIndex((void *) wcsa->myicsa, &nbytes);
		*size += nbytes;
		//*size += CSA_size(wcsa->myicsa);		
	}
	
/*	#ifdef CSA_ON
	{uint nbytes;

		int err = sizeIntIndex((void *) wcsa->myicsa, &nbytes);
	*size += nbytes;
	//*size += CSA_size(wcsa->myicsa);
	}
	#endif
*/	
	return 0;	
}


/** Querying the index =============================================================*/
	
	/* Writes in numocc the number of occurrences of the substring 
	   pattern[0..length-1] found in the text indexed by index. */
int count (void *index, uchar *pattern, ulong length, ulong *numocc){
	uint integerPatterns[MAX_INTEGER_PATTERN_SIZE];
	uint integerPatternSize;
	ulong l,r;	

	twcsa *wcsa=(twcsa *) index;
	parseTextIntoIntegers(wcsa, pattern, length, integerPatterns, &integerPatternSize);	
	if (!integerPatternSize) {*numocc=0; return 0;} //not found
	
	//*numocc = countCSA(wcsa->myicsa, integerPatterns, integerPatternSize, &l, &r);
	int err = countIntIndex((void *) wcsa->myicsa, integerPatterns, integerPatternSize, numocc,  &l, &r);
	return 0;
}

	/* Writes in numocc the number of occurrences of the substring 
	  pattern[0..length-1] in the text indexed by index. It also allocates
	  occ (which must be freed by the caller) and writes the locations of 
	  the numocc occurrences in occ, in arbitrary order.  */	  
int locate(void *index, uchar *pattern, ulong length, ulong **occ, ulong *numocc){
	uint integerPatterns[MAX_INTEGER_PATTERN_SIZE];
	uint integerPatternSize;
	ulong occurrences,l,r;	
	twcsa *wcsa=(twcsa *) index;
	
	parseTextIntoIntegers(wcsa, pattern, length, integerPatterns, &integerPatternSize);	
	if (!integerPatternSize) {*numocc=0; return 0;} //not found 

	ulong *seOffsets, *sourceOffsets;
  
  	//obtains the indexes in vector SE where the pattern appears.
	//seOffsets = locateCSA(wcsa->myicsa, integerPatterns, integerPatternSize, &occurrences);
	int err = locateIntIndex((void *)wcsa->myicsa, integerPatterns, integerPatternSize, &seOffsets, &occurrences);
	
	//obtains the offsets in the source text of the pattern (sourceOffsets)
	//sourceOffsets = (uint *) malloc (sizeof(uint)*occurrences);
	sourceOffsets=seOffsets;
	locateFacade(wcsa, (uint *)sourceOffsets,(uint *)seOffsets,occurrences);			
	*numocc = occurrences; 
	(*occ) = (ulong *)sourceOffsets;

	return 0;
}

   /* Gives the length of the text indexed */
int get_length(void *index, ulong *length) {
	twcsa *wcsa=(twcsa *) index;
	*length = wcsa->sourceTextSize;
	return 0;
}



/** ***********************************************************************************
  * Accessing the indexed text  
  * ***********************************************************************************/


	/**  Allocates snippet (which must be freed by the caller) and writes 
	  the substring text[from..to] into it. Returns in snippet_length the 
	  length of the text snippet actually extracted (that could be less 
	  than to-from+1 if to is larger than the text size). 			   */
int extract (void *index, ulong from, ulong to, uchar **snippet, ulong *snippet_length) {    		
    twcsa *wcsa=(twcsa *) index;    

	//    *snippet = displayFacadeMalloc(wcsa, from, to, snippet_length);    	
	//    return 0;        
	ulong n = wcsa->sourceTextSize;
	int pos;
	if (to >= n) to=n-1;
	if (from > to) {
		*snippet = NULL;
		*snippet_length=0;
	} else {  //uncompress  from -> to
		ulong j;
		ulong len =to-from+1;		
		*snippet = (uchar *) malloc((len)*sizeof(uchar));
		if (!(*snippet)) return 1;
		
		displayFacade(wcsa, from, to, snippet_length, *snippet);
		//(*snippet_length)=len;
	}
   return 0;
}

  /** Displays the text (snippet) surrounding any occurrence of the 
    substring pattern[0..length-1] within the text indexed by index. 
    The snippet must include numc characters before and after the 
    pattern occurrence, totalizing length+2*numc characters, or less if 
    the text boundaries are reached. Writes in numocc the number of 
    occurrences, and allocates the arrays snippet_text and 
    snippet_lengths (which must be freed by the caller). The first is a 
    character array of numocc*(length+2*numc) characters, with a new 
    snippet starting at every multiple of length+2*numc. The second 
    gives the real length of each of the numocc snippets. */

int display (void *index, uchar *pattern, ulong length, ulong numc, 
        ulong *numocc, uchar **snippet_text, ulong **snippet_lengths) {

	uint integerPatterns[MAX_INTEGER_PATTERN_SIZE];
	uint integerPatternSize;
	ulong occurrences,l,r;	
	ulong *seOffsets, *sourceOffsets;	
	ulong *occ;
	uchar *text_aux;
  	ulong i, j, from, to, len, x;  
	
	twcsa *wcsa=(twcsa *) index;

	parseTextIntoIntegers(wcsa, pattern, length, integerPatterns, &integerPatternSize);	
	if (!integerPatternSize) {*numocc=0; return 0;} //not found 
  
  	//obtains the indexes in vector SE where the pattern appears.
	//seOffsets = locateCSA(wcsa->myicsa, integerPatterns, integerPatternSize, &occurrences);
	int err = locateIntIndex((void *)wcsa->myicsa, integerPatterns, integerPatternSize, &seOffsets, &occurrences);
	
	*numocc = occurrences; 

	if (!occurrences) {
		*snippet_text =NULL;
		*snippet_lengths =NULL;
		return 0;
	}
	
	//obtains the offsets in the source text of the pattern (sourceOffsets)
	//sourceOffsets = (uint *) malloc (sizeof(uint)*occurrences);
	sourceOffsets=seOffsets;
	locateFacade(wcsa, (uint *)sourceOffsets, (uint *)seOffsets, occurrences);
  	//*numocc = occurrences; 
  	//free(seOffsets);	
  	//
	*snippet_lengths = (ulong *) malloc((*numocc)*sizeof(ulong));
	if (!(*snippet_lengths)) return 1;
	*snippet_text = (uchar *) malloc((*numocc)*(length+2*numc)*sizeof(uchar) +1) ;  //(the last "1" is for '\0');
	if (!(*snippet_text)) return 1;
	text_aux=*snippet_text;	


	for (i=0;i<(occurrences);i++) {
		x=sourceOffsets[i];
		if (x>numc) from = x-numc;
		else from=0;
		to= ((x+length+numc-1)<(wcsa->sourceTextSize-1)?(x+length+numc-1):(wcsa->sourceTextSize-1));
		
		displayFacade(wcsa,from,to,&len,text_aux); //asumo que machaca el �ltimo byte con un \0, pero despu�s se actualiza
			
		text_aux+=length+2*numc;
		(*snippet_lengths)[i] = len;
	}
	free(sourceOffsets);			
	
	return 0;
}

    /**  Obtains the length of the text indexed by index. */

int length (void *index, ulong *length) {
	return (get_length(index,length));
}




/** ***********************************************************************************
  * WORD-ORIENTED QUERY FUNCTIONS: LocateWord and DisplayWord
  * ***********************************************************************************/  
	/* Writes in numocc the number of occurrences of the substring 
	  pattern[0..length-1] in the text indexed by index. It also allocates
	  occ (which must be freed by the caller) and writes the locations of 
	  the numocc occurrences in occ, in arbitrary order. These occurrences
	  refer to the offsets in TOH where the caller could start a display
	  operation. So locateWord implies synchronization using B.
	  Moreover, positions occ[numocc.. 2*numocc-1] is set with the rank in SE of the
	  words whose codes begin in TOH in the positions in occ[0... numocc-1]
	  ** Parameter kbefore sets locateWord not to obtain the offset in TOH of the
	     searched word, but the offset in TOH of k-before words before.	  
	*/	  



int locateWord(void *index, uchar *pattern, ulong length, ulong **occ, ulong *numocc, uint kbefore){
	twcsa *wcsa=(twcsa *) index;
	
	uint integerPatterns[MAX_INTEGER_PATTERN_SIZE];
	uint integerPatternSize;
	ulong occurrences,l,r;	

	
	parseTextIntoIntegers(wcsa, pattern, length, integerPatterns, &integerPatternSize);	

//#define CHECKPATTERNS	
#ifdef CHECKPATTERNS	
static int countt=1;
static int sum =0;
static int numzeros=0;
  FILE *ff = fopen("patterns.statistics","a");
  //fprintf(ff,"%d\t%u\t",countt++,integerPatternSize);
  sum += integerPatternSize;
  if(!integerPatternSize) numzeros++;
  fprintf(ff,"%d-th pattern, numwords=%u, currNumZeroes= %d, currAvgSize = %2.3f\n",countt,integerPatternSize, numzeros,((double)sum)/countt );
  fprintf(stderr,"%d-th pattern, numwords=%u, currNumZeroes= %d, currAvgSize = %2.3f\n",countt,integerPatternSize, numzeros,((double)sum)/countt );  countt++;
  fclose(ff);

#endif

	if (!integerPatternSize) {*numocc=0; return 0;} //not found 
	  
	  
	ulong *seOffsets, *sourceOffsets;
  
  	//obtains the indexes in vector SE where the pattern appears.
	//seOffsets = locateCSA(wcsa->myicsa, integerPatterns, integerPatternSize, &occurrences);
	int err = locateIntIndex((void *)wcsa->myicsa, integerPatterns, integerPatternSize, &seOffsets, &occurrences);
		
	*numocc = occurrences;
 	
	if (!occurrences) {(*occ)=NULL;return 0;}

	seOffsets = (ulong *) realloc (seOffsets, 2 * sizeof(ulong) * occurrences); //to store both ptr and SE offsets
	sourceOffsets=seOffsets;
	{
		register ulong i,pos;
		for (i=0;i<occurrences;i++) {
			
			pos = (seOffsets[i]<kbefore)? 0 :(seOffsets[i]-kbefore);
/*			if (seOffsets[i] < kbefore) {
				pos =0;
				fprintf(stdout,"\n !!!!!!!!!!! MENOR QUE 0 seOff = %d, kbefore = %d",seOffsets[i], kbefore);
			}
			else pos = seOffsets[i]-kbefore;*/
			sourceOffsets[i]= BSelect(wcsa,pos);
			{ /** to show the first word from there */
				uint  indexSE;
				indexSE = pos;
				sourceOffsets[i+occurrences]=indexSE;
				
// 				{ /** decodes 1 word from there */
//					uint posSEValue,variant;		
// 					uint tmplen, j;
// 				 	byte *src;
// 					posSEValue =displayCSA(wcsa->myicsa,indexSE); 
// 					//sourceOffsets[i+occurrences]=posSEValue;
// 					uint ptr=sourceOffsets[i];				
// 					ptr= decodeHuffDecFuncVariantWordPos2bits(wcsa->HV, &variant, wcsa->TOH, ptr, posSEValue-1); 						
// 
// 				 	src=asciiVariantLen(&(wcsa->variants), posSEValue-1, variant, &tmplen);
// 				 	//fprintf(stderr,"\nWORD = ");
// 				 	//for (j=0;j<tmplen;j++) {fprintf(stderr,"%c",src[j]);}
// 				 	for (j=0;j<tmplen;j++) {uint c = src[j];}
// 				}				
			}
		}
		(*occ) = (ulong *)sourceOffsets;
	}
	return 0;
}  


  /** Displays the text (snippet) surrounding any occurrence of the 
    substring pattern[0..length-1] within the text indexed by index. 
    The snippet must include numc characters before and after the 
    pattern occurrence, totalizing length+2*numc characters, or less if 
    the text boundaries are reached. Writes in numocc the number of 
    occurrences, and allocates the arrays snippet_text and 
    snippet_lengths (which must be freed by the caller). The first is a 
    character array of numocc*(length+2*numc) characters, with a new 
    snippet starting at every multiple of length+2*numc. The second 
    gives the real length of each of the numocc snippets. */

 int displayWords (void *index, uchar *pattern, ulong length, ulong numc, 
         ulong *numocc, uchar **snippet_text, ulong **snippet_lengths, uint kbefore) {

 	ulong tohptr;
 	ulong *tohOffsets;
 	ulong *indexesInSE;
 	ulong occurrences;
 	uint bytesPerSnippet;
 	byte *text_aux;
	twcsa *wcsa=(twcsa *) index;
 	
 	locateWord(index, pattern, length, (ulong **)&tohOffsets, &occurrences, kbefore); 	
 	(*numocc) = occurrences; 
 	indexesInSE = &(tohOffsets[occurrences]); /** with #occurrences integers */  

 	if (!occurrences) {
 		*snippet_text =NULL;
 		*snippet_lengths =NULL;
 		return 0;
 	} 	

 	bytesPerSnippet = length+2*numc;
  	*snippet_lengths = (ulong *) malloc((*numocc)*sizeof(ulong));
  	if (!(*snippet_lengths)) return 1;
  	*snippet_text = (uchar *) malloc((*numocc)*(bytesPerSnippet)*sizeof(uchar) +1) ;  //(the last "1" is for '\0');
  	if (!(*snippet_text)) return 1;
  	
  	text_aux=*snippet_text;	  	
  	{
  		int err;
  		uint i, j, tmplen, variant;
  		uint ptr, maxptr;
  		byte *src, *dst;
  		uint snippetLen;
  		uint posSEValue; ulong indexSE;
  		uint alternate ;
		THuffDec *huff = &(wcsa->separators2.huffman);  		
  		maxptr = wcsa->tohSize;

  		for (i=0;i<occurrences;i++) {
			/** decodes words from there */			
			snippetLen=0;alternate =1;			
			ptr=tohOffsets[i];				
			indexSE = indexesInSE[i];			 
			
			dst = text_aux;					
			while (ptr < maxptr) { /** decoding bytesPerSnippet chars (if not at the end) */
				if (alternate) { //a valid word
					alternate=0;	
					//posSEValue =displayCSA(wcsa->myicsa,indexSE); indexSE++;
					err= displayIntIndex((void *)wcsa->myicsa,indexSE, &posSEValue);indexSE++;
					
					if (hasUniqueVariant (&(wcsa->variants),posSEValue -1)){
						variant = 0;
						//hasUniqueVariant
					}
					else 	
						ptr= decodeHuffDecFuncVariantWordPos2bits(wcsa->HV, &variant, wcsa->TOH, ptr, posSEValue-1); 						
			 		
			 		src=asciiVariantLen(&(wcsa->variants), posSEValue-1, variant, &tmplen);
		 		}	
				else {			  //a separator
					alternate =1;	
					decodeHuffDecMacro(huff,variant,wcsa->TOH,ptr);
					src = asciiSeparatorLen(&(wcsa->separators2),variant, &tmplen);
				}			 		
				/* at the end ?? */
		 		if  ((tmplen+snippetLen)>bytesPerSnippet) {
		 			tmplen =(bytesPerSnippet - snippetLen);
		 			ptr=maxptr; //so while loop ends;
		 		}	
				 				 			
		 		for (j=0;j<tmplen;j++) {*dst++ = *src++;}	
		 		snippetLen +=tmplen;		 		
		 	}

			text_aux += bytesPerSnippet;
			(*snippet_lengths)[i] = snippetLen;		 				
		}				
		free(tohOffsets);
	}
	return 0; 	     	
}




/** simulates extration of text process, but do not actually returns anything at all 
   Extracts upto <=2K words from K=wordsbefore words before each occurrence of a pattern.
   Less than 2K words can be extracted if more than numc characters have been already obtained.
   Do nothing else... do not return the text */
int  displayTextOcurrencesNoShow(void *index, uchar *pattern, ulong length, uint wordsbefore, uint maxnumc) {
	
	ulong *indexesInSE;
	ulong *tohOffsets;	
 	ulong occurrences;
 	byte *text_aux;
 	
	twcsa *wcsa=(twcsa *) index;
 	
 	locateWord(index, pattern, length, (ulong **)&tohOffsets, &occurrences, wordsbefore); 	
 	indexesInSE = &(tohOffsets[occurrences]); /** with #occurrences integers */  
 	if (!occurrences) {
 		return 0;
 	} 	
 	
 	ulong maxsnippetLen  = maxnumc;
	ulong extractedbytes = 0;

 	text_aux = (byte *) malloc (maxsnippetLen+1);

  	{
  		int err;
  		uint i, j, tmplen, variant;
  		uint ptr, maxptr;
  		uint alternate;
  		byte *src, *dst;
  		uint snippetLen;
  		//ulong posSEValue,indexSE;
  		uint posSEValue; ulong indexSE;

			THuffDec *huff = &(wcsa->separators2.huffman);  		
  		maxptr = wcsa->tohSize;
  		
  		uint numWordsToExtract = 2 * wordsbefore;  //the number of valid words that are extracted (at most).
  		uint z;
  		printf("\n occurrences = %lu",occurrences);
			
  		for (i=0;i<occurrences;i++) {

			/** decodes words from there */			
			snippetLen=0;alternate =1;							
			ptr=tohOffsets[i];				
			indexSE = indexesInSE[i];		 

			dst = text_aux;	
			z=0;				
			while ((z<numWordsToExtract) && (ptr < maxptr)){ /** extracting words (if not at the end) */
				if (alternate) { //a valid word
					alternate=0;	
					//posSEValue =displayCSA(wcsa->myicsa,indexSE); indexSE++;
					err= displayIntIndex((void *)wcsa->myicsa,indexSE, &posSEValue);indexSE++;
					if (hasUniqueVariant (&(wcsa->variants),posSEValue -1)) variant = 0;
					else ptr = decodeHuffDecFuncVariantWordPos2bits(wcsa->HV, &variant, wcsa->TOH, ptr, posSEValue-1); 									 			
					src=asciiVariantLen(&(wcsa->variants), posSEValue-1, variant, &tmplen);
					z++;
				}
				else {			  //a separator
					alternate =1;	
					decodeHuffDecMacro(huff,variant,wcsa->TOH,ptr);
					src = asciiSeparatorLen(&(wcsa->separators2),variant, &tmplen);
				}			 		
				/* at the end ?? */
				if  ((tmplen+snippetLen)>maxsnippetLen) {
					tmplen =(maxsnippetLen - snippetLen);
					ptr=maxptr; //so while loop ends;
				}				 			
				for (j=0;j<tmplen;j++) {*dst++ = *src++;}	
				snippetLen +=tmplen;		 		
				//fprintf(stderr,"\n %d",z);
							
			}//while

			extractedbytes += snippetLen;
				
		}	//for
	}
	//	printWord(text_aux,extractedbytes);		

	if (occurrences) free(tohOffsets);
	if (text_aux) free (text_aux);
	return extractedbytes; 
}




/**  Allocates text (which must be freed by the caller) and recovers the
  the substring of text starting from the "fromword"-th word up to the
  "toWord"-th words. Returns in text the text, and in "text_lenght" the 
  length of the text  actually extracted. Text is allocated. 
  Actually extracts SE[fromWord .. toWord) ... not the last word.    */

int extractWords (void *index, ulong fromWord, ulong toWord, uchar **text, 
        ulong *text_length){
        	
	twcsa *wcsa=(twcsa *) index;
    uint initTextLen=10000;
    uint avgWordLen =7;

	uint i, j;//, tmplen;
	uint prevValid;
	byte *src, *dst, *buff;
	uint tmplen =0;
	uint variant;
	
	THuffDec *huff = &(wcsa->separators2.huffman); 

	uint buffBytes = 1000;
	uint leng=0; //curr pos in buffer that was occupied.
	
	//fprintf(stderr,"\n toword = %d, wcsa->seSize= %d",toWord, wcsa->seSize);
	
	if (toWord > wcsa->seSize) toWord = wcsa->seSize;
	if (fromWord >= wcsa->seSize) fromWord = wcsa->seSize-1;
	if (buffBytes < ( (toWord-fromWord)* avgWordLen)) buffBytes = ((toWord-fromWord)* avgWordLen);
		
	buff = (uchar *) malloc (buffBytes * sizeof(char));
	if (!buff) return 1; //out of memory.
	dst = buff;
	
	register uint indexSE=fromWord;
	uint posSEValue=0;
	register uint ith;

	uint alternate =1;
	register uint ptr;		
	ptr = BSelect(wcsa, indexSE);	//position in toh				
		
	while  ( (indexSE < toWord) ){ /** extracting words (if not at the end) */

		//obtains a pointer to the separator or decompresses the variant and gets the pointer to it.
		if (alternate) { //a valid word
			alternate=0;	
			//posSEValue =displayCSA(wcsa->myicsa,indexSE); indexSE++;
			int err= displayIntIndex((void *)wcsa->myicsa,indexSE, &posSEValue);indexSE++;			
			if (hasUniqueVariant (&(wcsa->variants),posSEValue -1)) variant = 0;
			else ptr= decodeHuffDecFuncVariantWordPos2bits(wcsa->HV, &variant, wcsa->TOH, ptr, posSEValue-1); 									
			src=asciiVariantLen(&(wcsa->variants), posSEValue-1, variant, &tmplen);
		}	
		else {			  //a separator
			alternate =1;	
			decodeHuffDecMacro(huff,variant,wcsa->TOH,ptr);
			src = asciiSeparatorLen(&(wcsa->separators2),variant, &tmplen);
		}			 		
									
		//for (j=0;j<tmplen;j++) {*dst++ = *src++;}	
		//snippetLen +=tmplen;
		 				 	
		if ( buffBytes < (leng + tmplen+1) ) {
			buffBytes *=2;
			while (buffBytes < (leng + tmplen+1)) buffBytes *=2;
			buff = (uchar*) realloc(buff, buffBytes);
			if (!buff) return 1; //out of memory.
			dst = buff + leng;
		}																					
											
		for (j=0;j<tmplen;j++) {*dst++ = *src++;}	  //copies word to the output buffer
		leng +=tmplen;
	}//while

	*text_length =leng;
	*dst='\0';
	*text = buff;
	return 0;
}











/** ***********************************************************************************
	 CONSTRUCTION OF THE INDEX WCSA
    ***********************************************************************************/

	//Some functions used only inside build_WCSA function.
	/**------------------------------------------------------------------
	* initilizes Separators Structure
	------------------------------------------------------------------ */
	void initilizeSeparatorsBuild(uint NWords, tSepsBuild *seps) {
		seps->slots = (uint *) malloc (sizeof(uint) * NWords);
		seps->count = 0;
		seps->totallen=0;
		seps->hash = initialize_hash (NWords);		
	}
		
	/**------------------------------------------------------------------
	Compares two slots (by frequency). For qsort of variants && separators
	------------------------------------------------------------------ */			   
	int auxSlotsCompareFreq(const void *arg1, const void *arg2) {
		tauxslots *a1 = (tauxslots *) arg1;
		tauxslots *a2 = (tauxslots *) arg2;
		return a1->freq >= a2->freq;
	}			   
	
	/**------------------------------------------------------------------
	Compares two slots (alphanumericaly). For qsort of canonical words 
	------------------------------------------------------------------ */			   
	int canonicalWordsCompareAlpha(const void *arg1, const void *arg2) {
		tposInHT *a1 = (tposInHT *) arg1;
		tposInHT *a2 = (tposInHT *) arg2;
		return strcmp((char*)a1->word, (char *)a2->word);	
	}


/**
  * BUILDS THE WCSA INDEX
  */

int build_WCSA (uchar *text, ulong length, char *build_options, void **index) {
	
 	unsigned long numberOfCanonicalWords;  //number of different canonical words.
 	stt_hash stopwordshash; 	
 	char stopwordsfile[255]="nonDefined.stopwordsfile";
 	tSepsBuild sepsBuild;  //Structure than keeps the word-separators during construction

	t_hash hash;   		// the hash table to store both variants and canonical words.				      	
	tposInHT *posInHT;	// structure for canonicals and variants+huffmans 

 	uint *TOH;    //the compressed data.
 	uint tohSize; //it's size
 	uint *B;      // indicates the beginning of the code of a valid word in TOH 
 	uint bSize;   //it's size
 	uint *Z;      //samples of the original text
 	uint zSize;   //it's size = (tohSize + (DEFAULT_SAMPLE_PERIOD_Z -1)) /DEFAULT_SAMPLE_PERIOD_Z
 	uint samplePeriodZ;
 	uint samplePeriodB;
 	uint blocksizeFCHuTucker;
 	uint sourceTextSize;  

 	uint seSize;  //it's size == "numberOfValidWords". 

 	uint *SE;     //Integers vector. (represents the rank of the valid words in the source text).

	uint totallenCanonicals=0; //The numberOfBytes that occupy canonical words (their ascii version) in memory
	uint totallenVariants=0;   //The numberOfBytes occupied by variants  (their ascii version) in memory.
	uint totallenFstAndNum=0;   //The Sum of the depths of all the Huffman trees of variants.

	uint tohSizeSeparators, tohSizeValid; //the part in TOH (in bits) of valid words and separators.
	unsigned long bytesFile,bytesFileReal;
	long sizeNValue;

	/* used during first pass */	
			
	unsigned long ihash, addrInTH,addrInTHCanonical;
 	tauxslots *auxslots;   					 //used to sort all the words by frequency.
 	unsigned long numberOfDistinctValidWords=0;  //number of different words that appear in the text.
 	unsigned long numberOfValidWords=0;          //number of valid words in the text.
 	
	byte* inputBuffer = text;	
	bytesFileReal= bytesFile = length;

	sourceTextSize=length;
	
	/**     */
	twcsa *wcsa;
	wcsa = (twcsa *) malloc (sizeof (twcsa) * 1);

	samplePeriodZ = DEFAULT_SAMPLE_PERIOD_Z; //can be changed by an optional parameter sZ
	samplePeriodB = DEFAULT_SAMPLE_PERIOD_B; //can be changed by an optional parameter sB
	blocksizeFCHuTucker = DEFAULT_BLOCKSIZEFCHUTUCKER; //block size for front-coding-Hu-Tucker compression of canonicals.
	numberOfCanonicalWords=0;	


	{ //parsing the parameters of the index....
	
		/** processing the parameters */
		char delimiters[] = " =;";
		int j,num_parameters;
		char ** parameters;
		int value;
		
		if (build_options != NULL) {
		parse_parameters(build_options,&num_parameters, &parameters, delimiters);
		for (j=0; j<num_parameters;j++) {
		  
			if ((strcmp(parameters[j], "sZ") == 0 ) && (j < num_parameters-1) ) {
				samplePeriodZ=atoi(parameters[j+1]);			
			} 
			if ((strcmp(parameters[j], "sB") == 0 ) && (j < num_parameters-1) ) {
				samplePeriodB=atoi(parameters[j+1]);			
			} 
#ifdef COMPRESS_FCHT_CANONICALS			
			if ((strcmp(parameters[j], "blocksizefcht") == 0 ) && (j < num_parameters-1) ) {
				blocksizeFCHuTucker=atoi(parameters[j+1]);			
			} 
#endif			
			else if  ((strcmp(parameters[j], "stopfile") == 0 ) && (j < num_parameters-1) ) {
				strcpy(stopwordsfile,parameters[j+1]);
			}
		
			j++;
		}
		free_parameters(num_parameters, &parameters);
		}		
		printf("\n\tparameters of Presentation Layer: sampleZ=%u, sampleB=%u, blocksizeFCHuTucker=%u, \n\t\tstopWordsFile = %s\n",samplePeriodZ,samplePeriodB, blocksizeFCHuTucker,stopwordsfile);		
	}



	/**      */

	//Stimation (Using Heap's law) of the number of different "meaningful" words.
	//sizeNValue=N_value;
	if(bytesFile<5000000) bytesFile = 5000000;
	sizeNValue = (unsigned long) floor(3.9* pow(bytesFile,0.70) );
	

	// Inicializes the arrays used to detect if a char is valid or not.
	StartValid();
	// Inicializes the arrays used translated a char into lowercase.
	StartToLow();
		
	//initializes the hash table of stopwords
	//printf("\nLoading stop-words list...");
	loadStopWords(&stopwordshash,stopwordsfile);
	
	// **********************************************************************************
	//STARTING THE FIRST PASS. 
	// **********************************************************************************
	printf("\n\t ** STARTING THE FIRST PASS... **");
	
	posInHT = (tposInHT *) malloc(sizeof(tposInHT) * sizeNValue); 
	hash = initialize_hash (sizeNValue); //hash to cointain both canonical and normal words
	auxslots =(tauxslots *) malloc(sizeof(tauxslots) * sizeNValue);

	initilizeSeparatorsBuild(sizeNValue*2*2, &sepsBuild);	
	
	//-----------------------------------------------------------------	
	//1st pass (processing the file)
	{ 
		byte *pbeg,*pend,*wordstart,*aWord, *aSeparator;
		register unsigned long size;
		uint sizeTmp;
		unsigned int valid;
		uint i;
		
		tCanonical canonical;		

		pbeg = inputBuffer;
		pend = inputBuffer+bytesFileReal;
		sizeTmp = 0;
				
		while (pbeg <pend) {  
			
			//parsing either a word or phrase of separators.
			//parsing a word or separator
			size=0;
			wordstart = pbeg;
			if (_Valid[*pbeg]) {   //alphanumerical data
				while ( (size<MAX_SIZE_OF_WORD) && (pbeg<pend)&& ( _Valid[*pbeg] )) {size++;pbeg++;}
		    }
	      	else { // a separator
				while ( (size<MAX_SIZE_OF_GAP) && (pbeg<pend) &&  (!_Valid[*pbeg] )) {size++;pbeg++;}
			}
				
			//The parsed word/separator is  is "wordstart", and its length is "size"...
			aWord=wordstart;					
			aSeparator = wordstart;
									
			getCanonical(&canonical, aWord,size, stopwordshash);				
		 	valid = _Valid[*aWord] && (!canonical.type == IS_STOPWORD);
		 	while ((!valid) && (pbeg < pend)) {  //parses a separator-phrase
				wordstart = pbeg;
				sizeTmp = 0;
				if (_Valid[*pbeg]) {   //alphanumerical data
					while ( (size<MAX_SIZE_OF_WORD) && (pbeg<pend)&& ( _Valid[*pbeg] )) {sizeTmp++;size++;pbeg++;}

					getCanonical(&canonical, wordstart,sizeTmp, stopwordshash);
					valid = _Valid[*wordstart] && (!canonical.type == IS_STOPWORD);						
			    }
		      	else { // a separator
					while ( (size<MAX_SIZE_OF_GAP) && (pbeg<pend)&& (!_Valid[*pbeg]) ) {sizeTmp++;size++;pbeg++;}
				}							 		
		 	}			
		 	
		 	//Two cases:
		 	// 1. sizeTmp =0 ==> aWord points to a valid word with size= "size"
		 	// 2. else: aSeparator points to a non valid word (of size "size-sizeTmp") and
		 	//          aWord points to a valid word of size sizeTmp;
				
			if ((sizeTmp > 0)|| ((pbeg>=pend)&&(!_Valid[*aWord]))) { //hay frase de separador
				size -=sizeTmp;
				//printSeparator(aSeparator,size);
				//processment of the separator, which is pointed by "aSeparator" and
				//  whose size is "size".
				
				if (! inHashTable(sepsBuild.hash,aSeparator, size, &addrInTH )) {
					insertElement(sepsBuild.hash,aSeparator,size,&addrInTH );
					sepsBuild.totallen +=size +1;
					sepsBuild.hash->hash[addrInTH].posInVoc = sepsBuild.count;
					sepsBuild.slots[sepsBuild.count++]= addrInTH;									
				}
				else sepsBuild.hash->hash[addrInTH].freq ++;
														
				//....
				//end of processment of the separator.
				aWord=wordstart; size=sizeTmp; sizeTmp=0;
				if (!_Valid[*aWord])	{continue;} //	
			}	
					
			//Processement done for a valid word			
			//printfWord(aWord,size);			

			{ // a "meaningfull word" (a non-stopword)
				i = inHashTable(hash,aWord, size, &addrInTH );
				
				if (i) { // word  is already in the hash table.
					if (! hash->hash[addrInTH].freq) { //a canonical not appeared before (FREQ == 0)
						posInHT[hash->hash[addrInTH].posInVoc].number ++;
						
						//adds the new word to the arrays of "appearing valid-words"
						auxslots[numberOfDistinctValidWords].slot=addrInTH; //to permit listing all the words.
						numberOfDistinctValidWords++;										
						totallenVariants+=size+1;
					}
					hash->hash[addrInTH].freq++;
				}
				else { //word not in hash table																
					i = inHashTable(hash,canonical.word,canonical.size,&addrInTHCanonical);								
					if (!i)  {
						insertElement (hash,canonical.word, canonical.size, &addrInTHCanonical);
						hash->hash[addrInTHCanonical].freq=0;   //<-- adds with freq = 0
						posInHT[numberOfCanonicalWords].slot=addrInTHCanonical;
						posInHT[numberOfCanonicalWords].number =0;
						hash->hash[addrInTHCanonical].posInVoc = numberOfCanonicalWords;
						numberOfCanonicalWords++;
						totallenCanonicals += canonical.size +1;									
					}
					
					if (strcomp(canonical.word, aWord, canonical.size, size)!= 0){ //aWord is not a canonical form.
						i = inHashTable(hash,aWord, size, &addrInTH );	
						insertElement (hash, aWord, size, &addrInTH);
						hash->hash[addrInTH].posInVoc = hash->hash[addrInTHCanonical].posInVoc; //set posInVoc to the canonical form.
						posInHT[hash->hash[addrInTHCanonical].posInVoc].number ++;
					}
					else {																
						posInHT[hash->hash[addrInTHCanonical].posInVoc].number ++;	
						hash->hash[addrInTHCanonical].freq ++;
					} 
						
					//adds the new word to the array of "appearing valid-words"
					auxslots[numberOfDistinctValidWords].slot=addrInTH; //to permit listing all the words.
					numberOfDistinctValidWords++;
					totallenVariants+=size+1;

				}														
				////fprintf(stderr,"\n The canonical word [%s],%d, for word:",canonical.word, canonical.size);printWord(aWord,size);//fprintf(stderr,"\n");
			}				
		}//inner while
	}

	//for (i=0; i< sepsBuild.count;i++) {
	//	fprintf(stderr,"\n separator[%6d]:freq=%7d data= <<%s>>",i,sepsBuild.hash->hash[sepsBuild.slots[i]].freq,sepsBuild.hash->hash[sepsBuild.slots[i]].word);	}
		
	
	//sorting all the separators (that appear in the text) by frequency.		
	{	uint i;	
		{/**Sorting **/
			//copying to auxiliar structure for sorting
			tauxslots *auxseps = (tauxslots *) malloc (sizeof(tauxslots) * sepsBuild.count);
			for (i=0; i<sepsBuild.count;i++) {
				uint hashpos=sepsBuild.slots[i];
				auxseps[i].freq= sepsBuild.hash->hash[hashpos].freq;
				auxseps[i].slot= hashpos;
			}
			qsort(auxseps, sepsBuild.count, sizeof(tauxslots), auxSlotsCompareFreq);	
			//updating sepsBuild from the sorted auxiliar structure.
			for (i=0; i<sepsBuild.count;i++) {
				sepsBuild.slots[i] = auxseps[i].slot;
			}
			free(auxseps);
		}		

		uint *freqs;
		freqs = (uint *) malloc (sizeof(uint)*sepsBuild.count);		
		for (i=0;i<sepsBuild.count;i++) {
			freqs[i] = sepsBuild.hash->hash[sepsBuild.slots[i]].freq;		
			sepsBuild.hash->hash[sepsBuild.slots[i]].posInVoc = i;	  //restore ptr from hash -> vocab after qsort		
		}
		
		sepsBuild.huffman=createHuff (freqs,sepsBuild.count-1,SORTED);
		free(freqs);
		
		/** some statistics **/		
		//		fprintf(stderr,"\n TOTAL HUFFMAN OF SEPARATORS = %d bytes, #sep=%d",(sepsBuild.huffman.total+7)/8,sepsBuild.count);
		//		uint size=0;
		//		for (i=0;i<sepsBuild.count;i++) size += 1+strlen((char*)sepsBuild.hash->hash[sepsBuild.slots[i]].word);
		//		f(stderr,"\n TOTAL SIZE VOCAB OF SEPARATORS = %d bytes / %d",size, sepsBuild.totallen);
		//		for (i=0;i<numberOfCanonicalWords;i++) size += strlen((char*)hash->hash[posInHT[i].slot].word);
		//		fprintf(stderr,"\n TOTAL SIZE VOCAB OF VALID words = %d bytes",size);				
		/** end some statistics **/
	}
	
	printf("\n\tlenCanonicals %d, lenVariants %d", totallenCanonicals,totallenVariants);
	printf("\n\tTOTAL distinct words: %lu, distinct SEPS: %d",numberOfDistinctValidWords,sepsBuild.count);

	// **********************************************************************************
	// Generates the structure that will contain all the canonical words and their variants.
	// And generates the codes for each variant of a word and the Huffman for the separators 
	// **********************************************************************************

	//sorting all the words (that appear in the text) by frequency.		
	{	uint i;
		//copy data to auxiliar auxslots[i].freq field.
		for (i=0;i<numberOfDistinctValidWords;i++) auxslots[i].freq = hash->hash[auxslots[i].slot].freq;
		qsort(auxslots, numberOfDistinctValidWords, sizeof(tauxslots), auxSlotsCompareFreq);	
	}

	//allocating memory to store the variants of each canonical word.
	{
		register unsigned long i,number;
		for (i=0;i<numberOfCanonicalWords;i++) {
			posInHT[i].word = hash->hash[posInHT[i].slot].word;  //just setting the pointer.
			number =posInHT[i].number;
			posInHT[i].variants	= (byte **) malloc (number * sizeof (byte *));
			posInHT[i].freqs = (unsigned long *) malloc (number * sizeof (unsigned long));
			posInHT[i].slots = (unsigned long *) malloc (number * sizeof (unsigned long));   //needed for alf-N sorting
			posInHT[i].number = 0; //will be fulfilled later (during the copy from the hash table)
									     //so it is also used as the "pos" where to insert next variant.
		}
	}

	//Setting all the variants of each canonical word. NOTE: since auxslots is sorted by
	//frequency, all the variants for each canonical word are already inserted "sorted by
	//frequency", so no further sorting is needed to have all the variants of a canonical
	//word ordered by frequency, which is needed to apply word-based Huffman.
	//(MAYBE NOT THE FASTEST WAY??, BUT THE SIMPLEST).
	{
		register unsigned long p,i,pos,w;
		for (w=0;w<numberOfDistinctValidWords;w++) {
			p=auxslots[w].slot;
			if (hash->hash[p].freq > 0) {			
				i=hash->hash[p].posInVoc;
				pos = posInHT[i].number; posInHT[i].number++;
				posInHT[i].variants[pos]=hash->hash[p].word;
				posInHT[i].freqs[pos]=hash->hash[p].freq;
				posInHT[i].slots[pos]=p;				
				numberOfValidWords+= hash->hash[p].freq;	
				
				hash->hash[p].numVariant = pos;			
			}			
		}	
	}   //So we have all the variants of each canonical word sorted by frequency (ascending).
	
	free(auxslots);  //It is no longer needed.


//	//PROBA: �CANTAS PALABRAS CAN�NICAS CON S� 1 VARIANTE ? �E CANTAS OCURRENCIAS?
//	{	uint k,i;
//		uint maxNumberOfVariants=0;
//		uint numeroCon1Variante=0;
//		uint freqCon1Variante=0;
//		for (i=0; i<numberOfCanonicalWords;i++) {
//			k=posInHT[i].number;
//			if (k>maxNumberOfVariants) maxNumberOfVariants=k;
//			if (k==1) {
//				numeroCon1Variante ++;
//				freqCon1Variante += posInHT[i].freqs[0];
//			}
//		}
//		
//		
//		fprintf(stderr,"\n HAI %d de %d palabras can�nicas que te�en unha �nica variante",numeroCon1Variante, numberOfCanonicalWords);
//		fprintf(stderr,"\n O n�mero de ocurrencias delas � %d ", freqCon1Variante);
//		fprintf(stderr,"\n O numero total de palabras validas no texto � %d", numberOfValidWords );
//		fprintf(stderr,"\n O m�ximo NUMERO DE VARIANTES por palabra � %d", maxNumberOfVariants );
//		
//		
//		uint numeroConMais1Variante=0;
//		uint z;
//		uint freqConMais1Variante=0;
//		for (i=0; i<numberOfCanonicalWords;i++) {
//			k=posInHT[i].number;
//			if (k>1) {
//				numeroConMais1Variante ++;
//				for (z=0;z<k;z++) {
//					freqConMais1Variante += posInHT[i].freqs[z];
//				}
//			}
//		}
//	}
				
//		fprintf(stderr,"\n HAI %d de %d palabras can�nicas que te�en >1 variantes",numeroConMais1Variante, numberOfCanonicalWords);
//		fprintf(stderr,"\n O n�mero de ocurrencias delas � %d ", freqConMais1Variante);
//		fprintf(stderr,"\n O numero total de palabras validas no texto � %d", numberOfValidWords );		
//		exit(0);
//CR		
// O m�ximo NUMERO DE VARIANTES por palabra � 5
// TOTAL HUFFMAN OF SEPARATORS = 4167938 bytes, #sep=127004
// TOTAL SIZE VOCAB OF SEPARATORS = 1896328 bytes
// TOTAL SIZE VOCAB OF VALID words = 2528239 bytes
//TOTAL SIZEWORDS: 31580966, SIZESEPS: 19504579, TOTAL: 51085545
//TOTAL NUM WORDS: 4964139, SEPS: 4964140, TOTAL: 9928279
//TOTAL distinct words: 115829, distinct SEPS: 127004
// HAI 71485 de 90748 palabras can�nicas que te�en unha �nica variante
// O n�mero de ocurrencias delas � 782590
// O numero total de palabras validas no texto � 4964139
// HAI 19263 de 90748 palabras can�nicas que te�en >1 variantes
// O n�mero de ocurrencias delas � 4181549
// O numero total de palabras validas no texto � 4964139


//AP.
//O m�ximo NUMERO DE VARIANTES por palabra � 7
// TOTAL HUFFMAN OF SEPARATORS = 22313609 bytes, #sep=410954
// TOTAL SIZE VOCAB OF SEPARATORS = 6685819 bytes
// TOTAL SIZE VOCAB OF VALID words = 8121711 bytes
//TOTAL SIZEWORDS: 151119640, SIZESEPS: 99514546, TOTAL: 250634186
//TOTAL NUM WORDS: 25363843, SEPS: 25363843, TOTAL: 50727686
//TOTAL distinct words: 217727, distinct SEPS: 410954
// HAI 148938 de 181705 palabras can�nicas que te�en unha �nica variante
// O n�mero de ocurrencias delas � 3938512
// O numero total de palabras validas no texto � 25363843
// HAI 32767 de 181705 palabras can�nicas que te�en >1 variantes
// O n�mero de ocurrencias delas � 21425331
// O numero total de palabras validas no texto � 25363843


	//Sorting the canonical words alphanumerically (over posInHT)
	//and setting the new posInVoc in the hash table for each variant.
	//Note that this is only needed if we want that any variant (in hash)
	//can know the position of its "canonical version" in the sorted arrays of canonical words.
	{	register unsigned long i,j;	
		//sorting canonical words ...
		qsort(posInHT, numberOfCanonicalWords, sizeof(tposInHT), canonicalWordsCompareAlpha);		
		
		//setting in hash the new positions of the canonical words.
		for (i=0;i<numberOfCanonicalWords;i++) {	
			hash->hash[posInHT[i].slot].posInVoc = i;
			for (j=0;j<posInHT[i].number;j++) {
				hash->hash[posInHT[i].slots[j]].posInVoc = i;
			}			
		}
		//so in hash, all the variants of a canonical word W (and W itself) point to the same posInVoc.
	}	
	//from now on, posInHT[i].slots array is no longer neccesary.	
	//             the canonical words are sorted alphanumerically in posInHT
	//             all the variants for a canonical word are sorted by freq (ascending)		
		
	
	//%%%%%%%%%%%%%% GENERATING CODES for VALID WORDS ...........................................	
	{	//creates all the Huffman trees that include all the variants of each canonical word.
		tohSizeValid=0;
		register unsigned long i;
		for (i=0;i<numberOfCanonicalWords;i++) {
			posInHT[i].huffman = createHuff ((uint *)posInHT[i].freqs,posInHT[i].number-1,SORTED);
			/**DEPTH*/
			//totallenFstAndNum +=(posInHT[i].huffman.depth +1); setting fst[dept+1]=num[dept+1] =0
			totallenFstAndNum +=(posInHT[i].huffman.depth);   //using depth during decoding
			
			//tohSizeValid+= posInHT[i].huffman.total;  /*****@@*****/
			tohSizeValid+=  (posInHT[i].number > 1) ?  ( posInHT[i].huffman.total ) : 0 ; 

		}					


//		/** solo para debugging **/
//		{ uint suma=0;
//		for (i=0;i<numberOfCanonicalWords;i++) {
//			 if (posInHT[i].number > 1){
//			 		suma +=posInHT[i].huffman.total ;
//			 		 fprintf(stderr,"\nTOH SIZE needed for word %d is %d bits y suma actual %d bits",i,posInHT[i].huffman.total, suma);
//			 		 
//			 	}
//			
//		}
//		}
//		/** fin solo para debuggin **/		
		
		tohSizeSeparators =sepsBuild.huffman.total;
		tohSize	=tohSizeValid + tohSizeSeparators;				
	}

	//%%%%%%%%%%%%%% INITIALIZING bitmaps ......................................
	{
		seSize= numberOfValidWords;
		bSize = (seSize + (samplePeriodB -1)) /samplePeriodB;
		zSize = (seSize + (samplePeriodZ -1)) /samplePeriodZ;		
		 
		TOH = (uint *) malloc (((tohSize+W-1)/W) * sizeof (uint));    //it is a bit vector
		//TOH[((tohSize+W-1)/W) -2]=0000;
		TOH[((tohSize+W-1)/W) -1]=0000;  //Just to clear the last uint, that could remain with trash (uninitialized)
										 //So that valgrind says nothing when writting to disk.
		SE  = (uint *) malloc ((seSize+1)*sizeof (uint));
		

		B   = (uint *) malloc ( bSize * sizeof (uint));    
		Z   = (uint *) malloc ( zSize * sizeof (uint));
		
		printf("\n\t ===================================================:");				
		printf("\n\t Summary of first pass from parsing:");		
		printf("\n\t   Number of canonical words = %ld",numberOfCanonicalWords);
		printf("\n\t   Number of valid words = %lu",numberOfValidWords);
		printf("\n\t   Size of vectors: TOH= %u, TOHSeparators = %u, TOHsizeValid = %u",tohSize,tohSizeSeparators, tohSizeValid);
		printf("\n\t   Size of  B= %u, SE = %u, Z=%u, validWords =%ld",bSize,seSize,zSize,numberOfValidWords);
		printf("\n\t ===================================================:");		
		printf("\n");		
	}


	/** **********************************************************************************
	  *  STARTING THE SECOND PASS.
	  * **********************************************************************************/

	printf("\n\t ** STARTING THE SECOND PASS... ** ");
	//2nd pass (processing the file)
	{ 
		byte *pbeg,*pend,*wordstart,*aWord,*aSeparator;
		register unsigned long size;
		unsigned int valid;
		uint sizeTmp;
		sizeTmp = 0;
		
		register uint offset =0;   //byte in the source text of the beginning of the current word/sep.
		register uint offsetCT=0; //bit in the compressed text.
		register uint countValidWords =0; //counts the valid words that have appeared to detect when
										  // a new sample has to be added to Z.
		tCanonical canonical;		

		pbeg = inputBuffer;
		pend = inputBuffer+bytesFileReal;
				
		while (pbeg <pend) {  
			
			//parsing either a word or phrase of separators.
			//parsing a word or separator
			size=0;
			wordstart = pbeg;
			if (_Valid[*pbeg]) {   //alphanumerical data
				while ( (size<MAX_SIZE_OF_WORD)&& (pbeg<pend) && ( _Valid[*pbeg] )) {size++;pbeg++;}
		    }
	      	else { // a separator
				while ( (size<MAX_SIZE_OF_GAP) && (pbeg<pend)&& (!_Valid[*pbeg]) ) {size++;pbeg++;}
			}
				
			//The parsed word/separator is  is "wordstart", and its length is "size"...
			aWord=wordstart;					
			aSeparator = wordstart;
									
			getCanonical(&canonical, aWord,size, stopwordshash);				
		 	valid = _Valid[*aWord] && (!canonical.type == IS_STOPWORD);
		 	while ((!valid) && (pbeg < pend)) {  //parses a separator-phrase
				wordstart = pbeg;
				sizeTmp = 0;
				if (_Valid[*pbeg]) {   //alphanumerical data
					while ( (size<MAX_SIZE_OF_WORD) && (pbeg<pend)&& ( _Valid[*pbeg] )) {sizeTmp++;size++;pbeg++;}

					getCanonical(&canonical, wordstart,sizeTmp, stopwordshash);
					valid = _Valid[*wordstart] && (!canonical.type == IS_STOPWORD);						
			    }
		      	else { // a separator
					while ( (size<MAX_SIZE_OF_GAP) && (pbeg<pend)&& (!_Valid[*pbeg]) ) {sizeTmp++;size++;pbeg++;}
				}							 		
		 	}			
		 	
		 	//Two cases:
		 	// 1. sizeTmp =0 ==> aWord points to a valid word with size= "size"
		 	// 2. else: aSeparator points to a non valid word (of size "size-sizeTmp") and
		 	//          aWord points to a valid word of size sizeTmp;
						 		
			{
				register unsigned long i;				
				//register byte *w;			
				register uint variantNumb;				
				
				if ((sizeTmp > 0) || ((pbeg>=pend)&&(!_Valid[*aWord]))) { //hay frase de separador
					size -=sizeTmp;
					//printSeparator(aSeparator,size);
					//processment of the separator, which is pointed by "aSeparator" and
					//  whose size is "size".
					//....				
					i= inHashTable(sepsBuild.hash,aSeparator, size, &addrInTH ); 
					variantNumb = sepsBuild.hash->hash[addrInTH].posInVoc;				
							
					offsetCT=encodeHuff(sepsBuild.huffman, variantNumb, TOH, offsetCT);															
					//end of processment of the separator.

					offset+=size;
					aWord=wordstart; size=sizeTmp; sizeTmp=0;
					if (!_Valid[*aWord]) {continue;} // --> not processing the valid word till next iteration	
				}	
				
				// a "meaningfull word" (a non-stopword)
				i = inHashTable(hash,aWord, size, &addrInTH );	
				variantNumb = hash->hash[addrInTH].numVariant;

				if (!(countValidWords % samplePeriodB)) B[countValidWords / samplePeriodB] = offsetCT;			
							 
				if (posInHT[hash->hash[addrInTH].posInVoc].number > 1) {
						//fprintf(stderr,"\n******************ENCODING word %d, TOH SIZE needed for word %d is %d bits",hash->hash[addrInTH].posInVoc,hash->hash[addrInTH].posInVoc,posInHT[hash->hash[addrInTH].posInVoc].huffman.total);
						offsetCT=encodeHuff(posInHT[hash->hash[addrInTH].posInVoc].huffman, variantNumb, TOH, offsetCT);  /*****@ + de 1 variante@*****/
				}
				
				if (!(countValidWords % samplePeriodZ)) Z[countValidWords / samplePeriodZ] = offset;
							
				SE[countValidWords]=hash->hash[addrInTH].posInVoc+1;
				countValidWords++;		
				offset+=size;																				
			}																
		}//inner while		
		
	//END OF SECOND PASS	
	printf("\n\t END OF SECOND PASS, bits on compressed Huffm seq. = %d\n",offsetCT);
	SE[countValidWords] = 0;   //needed for a S.A.
				
	}/** ====== End of 2nd pass **************************************************** */
	
	//freeing the source text (it is no longer needed).

	printf("\n\t ** Freeing memory **");
	free(inputBuffer); //the text	

	/** Now Setting the data of the index **/
	wcsa->numberOfCanonicalWords = numberOfCanonicalWords;
	wcsa->samplePeriodZ = samplePeriodZ;
	wcsa->samplePeriodB = samplePeriodB;

#ifdef COMPRESS_FCHT_CANONICALS	
	wcsa->blocksizeFCHuTucker = blocksizeFCHuTucker;
#endif

	wcsa->sourceTextSize = sourceTextSize;
	wcsa->stopwordshash = stopwordshash;

	wcsa->stopwordsfilename = (char *) malloc (sizeof(char) * strlen(stopwordsfile)+1);
	strcpy(wcsa->stopwordsfilename, stopwordsfile);
	wcsa->tohSize = tohSize;
	wcsa->TOH = TOH;
	wcsa->bSize = bSize;
	wcsa->B = B;
	wcsa->zSize = zSize;
	wcsa->Z = Z;
	wcsa->seSize = seSize;

	{/** the separators and the huffman tree of separators */
		uint i, depth;
		/** preparing the Huffman tree of the separators for decoding (before saving) */
		prepareToDecode(&(sepsBuild.huffman));
		depth = wcsa->separators2.huffman.depth = sepsBuild.huffman.depth;

		/** the num array and fst array.*/		
		wcsa->separators2.huffman.num = (uint *) malloc ((depth+1)*sizeof(uint));
		wcsa->separators2.huffman.fst = (uint *) malloc ((depth+1)*sizeof(uint));
		for (i=0;i<=depth; i++) {
			wcsa->separators2.huffman.num[i] = sepsBuild.huffman.num[i];
			wcsa->separators2.huffman.fst[i] = sepsBuild.huffman.fst[i];
		}
		
		tZoneMem asciiSeps;
 		asciiSeps.zone = (byte *) malloc (sizeof(byte) * (sepsBuild.totallen));
 		asciiSeps.size = sepsBuild.totallen;

		//moves data from sepsBuild to a temporal "tzoneMem: asciiSeps" structure
		byte *zoneMem,*src;
		wcsa->separators2.count = sepsBuild.count;
 		zoneMem = asciiSeps.zone;				
		for(i = 0; i < wcsa->separators2.count; i++) {
			src = sepsBuild.hash->hash[sepsBuild.slots[i]].word; //copying the separator
			while (*src) {*zoneMem++ = *src++;} //moving data until '\0'
			*zoneMem='\0'; zoneMem++;          //copies also the '\0'
		}
		//freeing structures used during construction
		freeHuff(sepsBuild.huffman);
		destroy_hash(sepsBuild.hash);
		free(sepsBuild.slots);	

		buildSeparators(&(wcsa->separators2), &(asciiSeps));
		free(asciiSeps.zone);
	}	

	/**-----------*/
	{/** the canonical words and their variants and Huffman trees. */
		byte *zoneMem,*src;
		uint i;
								 
				//the canonical words sorted alphabetically. Moving data from posInHT to WCSA structure
		//		wcsa->wordsData = (twords *) malloc(sizeof(twords) * numberOfCanonicalWords); 
		// 		wcsa->canonicalsZoneMem.zone = (byte *) malloc (sizeof(byte) * totallenCanonicals);
		// 		wcsa->canonicalsZoneMem.size = totallenCanonicals;
		// 		zoneMem = wcsa->canonicalsZoneMem.zone;				
		//		for(i = 0; i < wcsa->numberOfCanonicalWords; i++) {
		//			src = posInHT[i].word;				 //copying the canonical word
		//			wcsa->wordsData[i].word = zoneMem;   //setting the pointer
		//			while (*src) {*zoneMem++ = *src++;}  //moving data until '\0'
		//			*zoneMem='\0'; zoneMem++;            //copies also the '\0'
		//		}
			


		{  
		/** the canonical words sorted alphabetically. 
		    Moving data from posInHT to WCSA structure */
		  uint n = wcsa->numberOfCanonicalWords;
			uint *tmpOffsets = (uint *) malloc (sizeof(uint) * (n+1) );  //1 extra uint (to point to the virtual "n+1" ^th word.
			uint tmpOffset =0;
			 
			byte *zoneMem,*src;
			uint i;
									 
			//Moving data from posInHT to WCSA structure
	 		wcsa->wordsData.wordsZoneMem.size = totallenCanonicals - n; //without '\0' bytes (end-tag).
	 		wcsa->wordsData.wordsZoneMem.zone = (byte *) malloc ( wcsa->wordsData.wordsZoneMem.size * sizeof(byte) );  
	 		zoneMem = wcsa->wordsData.wordsZoneMem.zone;					
			for(i = 0; i < n; i++) {
				src = posInHT[i].word;				 //copying the canonical word
				//wcsa->wordsData.words[i].word = zoneMem;   //setting the pointer
				tmpOffsets[i]=tmpOffset;  //offset in zoneMem
				while (*src) {*zoneMem++ = *src++; tmpOffset++;}  //moving data until '\0'			
			}
			tmpOffsets[n]=tmpOffset; //setting pointer to the "virtual" word {n+1}^{th}
			
			//kbit encoding of the offsets
			uint elemSize = bits(tmpOffset);
			wcsa->wordsData.elemSize = elemSize;
			wcsa->wordsData.words = (uint *) malloc (((((n+1) *elemSize)+W-1) /W) * sizeof(uint));  //with 1 extra slot !.
			wcsa->wordsData.words[((((n +1)*elemSize)+W-1) /W)   -1 ] =0000;
			//		fprintf(stderr,"\n ElemSize = %d, maxOffset = %d",elemSize,tmpOffset);
	
			tmpOffset=0;
			for (i=0; i<=n; i++) {  //setting "n+1" offsets
					bitwrite(wcsa->wordsData.words, tmpOffset, elemSize, tmpOffsets[i]);
					tmpOffset+=elemSize; 
			}
			
			//frees temporary array of offsets into zoneMem.
			free(tmpOffsets);
#ifdef COMPRESS_HUFFMAN_CANONICALS					
			buildHuffCanonicals (&(wcsa->HC), posInHT, wcsa->numberOfCanonicalWords);
			
			//showCanonicals(&(wcsa->HC));
			//comparing all the canonical words with its compressed representation.
			/*
			{uint x;
				byte *dst;
				uint len;
				for (x=0;x<wcsa->numberOfCanonicalWords;x++) {
					dst = asciiCanonical(&wcsa->HC, x);
					len = lenCanonical(&wcsa->HC, x) ;
					fprintf(stderr,"\n word %u, %s <--> %s!!!",x,dst,posInHT[x].word);
					if (strcmp((char *)posInHT[x].word,(char*)dst)) {
						fprintf(stderr,"\n Canonical comparison failed:  word %u, %s <--> %s!!!",x,dst,posInHT[x].word);
						break;
					}
				}
			}
			*/
#endif			
#ifdef COMPRESS_FCHT_CANONICALS	
			
			{
				byte **cwords = (byte **) malloc ( sizeof (byte *) * wcsa->numberOfCanonicalWords);
				uint x;
				for (x=0;x<wcsa->numberOfCanonicalWords;x++) {
					cwords[x] = (byte *)strdup((char *)posInHT[x].word);
				}

				buildCanonicalsFCHT(&(wcsa->fcht), cwords, wcsa->numberOfCanonicalWords ,(size_t) wcsa->blocksizeFCHuTucker);

				for (x=0;x<wcsa->numberOfCanonicalWords;x++) {
					free(cwords[x]);
				}
				free(cwords);
			}		
							
			
			//showCanonicals(&(wcsa->HC));
			//comparing all the canonical words with its compressed representation.
			/*
			{uint x;
				byte *dst;
				uint len;
				for (x=0;x<wcsa->numberOfCanonicalWords;x++) {
					dst = getWordFCHT(wcas->fthc, (size_t) x);
					len = strlen(dst);
									
					fprintf(stderr,"\n word %u, %s <--> %s!!!",x,dst,posInHT[x].word);
					if (strcmp((char *)posInHT[x].word,(char*)dst)) {
						fprintf(stderr,"\n Canonical comparison failed:  word %u, %s <--> %s!!!",x,dst,posInHT[x].word);
						break;
					}
				}
			}
			*/
#endif	



		}	 //end canonicals	

		//the variants...
		buildVariants(&(wcsa->variants), posInHT, numberOfCanonicalWords,
				totallenVariants,  numberOfDistinctValidWords);

		//showVariants(&(wcsa->variants));
		
		//The Huffman trees of all the words
		//prepares Huffman trees for decoding (mandatory before storing Huffman's tree)
		for (i=0;i<numberOfCanonicalWords;i++) prepareToDecode(&(posInHT[i].huffman));

		//build the compacted structure wcsa->HT 			
		buildHuffVariants(&(wcsa->HV), posInHT, numberOfCanonicalWords,totallenFstAndNum);

	}//end canonicals, and variants and Huffman Trees of variants.	

	/**-----------*/
	//frees memory from facade's structures.
	{
		uint i,total;
		destroy_hash(hash);
		
		for (i=0;i<numberOfCanonicalWords;i++) {
			 freeHuff(posInHT[i].huffman); 
			 free(posInHT[i].freqs); free(posInHT[i].variants);
			 free(posInHT[i].slots);			
		}
		free(posInHT);
	}			
	
	/********* creates the self-index on ints (bottom layer) *********/
/*
 * 	//creating CSA from Edu's code...
	#ifdef CSA_ON
	{
		uint total;
		fprintf(stderr,"\n **** CREATING CSA from Edu's Code *****");
		//ticsa *myicsa;		
		//myicsa = createIntegerCSA(&SE,seSize+1, build_options);
		//wcsa->myicsa= myicsa;
		void *bottomIntIndex;
		int err =  buildIntIndex(SE,seSize+1, build_options,(void **)&bottomIntIndex);				
		//wcsa->myicsa= (ticsa *) bottomIntIndex;
		wcsa->myicsa = bottomIntIndex;

		//total = CSA_size(wcsa->myicsa);
		err = sizeIntIndex((void *) wcsa->myicsa, &total);

		free(SE);  //comes from facade data structures, but used in CSA.

		printf("\n\t**** [iCSA built on %d words. Size = %ld bytes... RAM",seSize,total);
	}	
	#endif
	
	*/ 
	//#ifndef CSA_ON
		wcsa->se = SE;
		wcsa->myicsa = NULL;
	//#endif
	
	printf("\n\t ** Building done! **\n");

	*index = wcsa;
	return 0;
}


int build_iCSA (char *build_options, void *index)
{
	twcsa *wcsa = (twcsa *) index;
	/********* creates the self-index on ints (bottom layer) *********/
	//creating CSA from Edu's code...
	//#ifdef CSA_ON
	{
		uint total;
		fprintf(stderr,"\n **** CREATING CSA from Edu's Code *****");
		//ticsa *myicsa;		
		//myicsa = createIntegerCSA(&SE,seSize+1, build_options);
		//wcsa->myicsa= myicsa;
		void *bottomIntIndex;
		int err =  buildIntIndex(wcsa->se,wcsa->seSize+1, build_options,(void **)&bottomIntIndex);				
		//wcsa->myicsa= (ticsa *) bottomIntIndex;
		wcsa->myicsa = bottomIntIndex;

		//total = CSA_size(wcsa->myicsa);
		err = sizeIntIndex((void *) wcsa->myicsa, &total);

		//free(wcsa->se);  //comes from facade data structures, but used in CSA.

		printf("\n\t**** [iCSA built on %u words. Size = %u bytes... RAM",wcsa->seSize,total);
	}	
	//#endif	
	return 0;
}

/** ********************************************************************
  * Loading from disk
  **********************************************************************/ 

/**-----------------------------------------------------------------  
 * LoadWCSA.                                                       
 * Loads all the data structures of WCSA (included the icsa)     
 ----------------------------------------------------------------- */ 

twcsa *loadWCSA(char *filename) {
	twcsa *wcsa;
	wcsa = (twcsa *) malloc (sizeof (twcsa) * 1);
	wcsa->numberOfCanonicalWords=0;
	wcsa->se = NULL;

	// Inicializes the arrays used to detect if a char is valid or not.
	StartValid();
	// Inicializes the arrays used translated a char into lowercase.
	StartToLow();

	//wcsa->myicsa = loadCSA(filename);		
	int err = loadIntIndex(filename, (void **)&wcsa->myicsa);
						  	
	//wcsa->samplePeriodZ = DEFAULT_SAMPLE_PERIOD_Z;  //will be changed by loadStructs...
	//wcsa->samplePeriodB = DEFAULT_SAMPLE_PERIOD_B;
	loadStructs(wcsa,filename);

	//err = sourceLenIntIndex( (void *) wcsa->myicsa, &(wcsa->seSize));	 							    
	//wcsa->seSize = wcsa->myicsa->suffixArraySize;  //not needed, only for prints.	

	return wcsa;
}

/** ------------------------------------------------------------------
 * LoadStructs.
 *	Reads files and loads all the data needed for searcherFacade
 ----------------------------------------------------------------- */ 
 void loadStructs(twcsa *wcsa, char *basename) {
	uint i,j;
	char *filename;
	int file;
	uint sizeFile;
	char c;
	uint numberOfCanonicalWords;
		
	filename = (char *)malloc(sizeof(char)*(strlen(basename)+10));
	fprintf(stderr,"Loading Index from file %s.*\n", basename);
	
	//** SOME CONSTANTS: tohSize, bSize and sourceTextSize
	{	strcpy(filename, basename);
		strcat(filename, ".");
		strcat(filename, CONSTANTS_FILE_EXT);
		
		if( (file = open(filename, O_RDONLY)) < 0) {
			printf("Cannot open file %s\n", filename);
			exit(0);
		}
				
		read(file, &(wcsa->tohSize), sizeof(uint));
		read(file, &(wcsa->sourceTextSize), sizeof(uint));	
		read(file, &(wcsa->seSize), sizeof(uint));
		read(file, &(wcsa->samplePeriodB), sizeof(uint));	
		read(file, &(wcsa->samplePeriodZ), sizeof(uint));
	#ifdef COMPRESS_FCHT_CANONICALS
		uint tempval;
		read(file, &(tempval), sizeof(uint));
		wcsa->blocksizeFCHuTucker = (size_t) tempval;
	#endif
		close(file);
	}		
	
	fprintf(stderr,"\n@@ loaded value of seSize = %d", wcsa->seSize);
	
	//** THE FILE WITH THE STOPWORDS...
	{
		strcpy(filename, basename);
		strcat(filename, ".");
		strcat(filename, STOPWORDS_FILE_EXT);
		loadStopWords(&(wcsa->stopwordshash),filename);
		wcsa->stopwordsfilename =NULL; //not needed during searches !! (only for save_index)
	}	

#ifdef KEEP_CANONICALS_UNCOMPRESSED		
	/** File with the canonical words. (sorted alphanumerically) */
	{	byte *zoneMem;
		uint n;
		strcpy(filename, basename);
		strcat(filename, ".");
		strcat(filename, CANONICAL_WORDS_FILE_EXT);
		//sizeFile= fileSize(filename)-sizeof(uint);

		if( (file = open(filename, O_RDONLY)) < 0) {
			printf("Cannot open file %s\n", filename);
			exit(0);
		}

		//		//the number of canonical words
		//		read(file, &numberOfCanonicalWords, sizeof(uint));
		//		wcsa->wordsData = (twords *) malloc(sizeof(twords) * numberOfCanonicalWords); 
		//		wcsa->numberOfCanonicalWords = numberOfCanonicalWords;
		//
		//		//allocating the memory needed for all words and reading them		
		// 		wcsa->canonicalsZoneMem.zone = (byte *) malloc (sizeof(byte) *sizeFile);
		// 		wcsa->canonicalsZoneMem.size = sizeFile;
		// 		zoneMem = wcsa->canonicalsZoneMem.zone;
		// 		 			
		//		//reading the words 
		//		//P.S. Can be optimized by reading just once (whole file) and then setting pointers for the words ...
		//		for(i = 0; i < numberOfCanonicalWords; i++) {
		//			wcsa->wordsData[i].word=zoneMem;
		//			while (1) {read(file, zoneMem, sizeof(byte)); if (!(*zoneMem++)) break;} //reads word
		//		}
		//		close(file);
		//	}	

		//the number of canonical words
		read(file, &n, sizeof(uint));
		wcsa->numberOfCanonicalWords = n;
		read(file, &(wcsa->wordsData.elemSize), (sizeof(uint)));
		read(file, &(wcsa->wordsData.wordsZoneMem.size), (sizeof(uint)));
		
		//allocating the memory needed for all words and reading them	//(ascii) << no \0 chars are needed>>.	
		wcsa->wordsData.wordsZoneMem.zone = (byte *) malloc(wcsa->wordsData.wordsZoneMem.size * sizeof(byte));			
	 	read(file, (wcsa->wordsData.wordsZoneMem.zone),     wcsa->wordsData.wordsZoneMem.size * sizeof(byte));
	
		//reading the offsets of the words (kbitArray that points to offsets in zoneMem of words.
		wcsa->wordsData.words = (uint *) malloc (((((n+1)* (wcsa->wordsData.elemSize))+W-1) /W) * sizeof(uint));	
		wcsa->wordsData.words[ ((((n+1)*(wcsa->wordsData.elemSize))+W-1) /W)   -1 ] =0000;
		read(file, (wcsa->wordsData.words),     ((((n+1)* (wcsa->wordsData.elemSize))+W-1) /W) * (sizeof(uint)));		
		close(file);
	}	
#endif
		
#ifdef COMPRESS_HUFFMAN_CANONICALS
	/** File with the canonical words (sorted by alphabetically)... compressed with huffman.*/
	{	
		wcsa->wordsData.wordsZoneMem.zone = NULL;
		wcsa->wordsData.words = NULL;
		
		/**version 2 of CANONICALS (compressed with huffman) */
		strcpy(filename, basename);
		strcat(filename, ".");
		strcat(filename, CANONICAL_WORDS_FILE_EXT2);

		if( (file = open(filename, O_RDONLY)) < 0) {
			printf("Cannot open file %s\n", filename);
			exit(0);
		}

		read(file, &(wcsa->numberOfCanonicalWords), sizeof(uint));		
		loadCanonicals(&(wcsa->HC),file);
		close(file);
	}
#endif

#ifdef COMPRESS_FCHT_CANONICALS
	/** File with the canonical words (sorted by alphabetically)... compressed with huffman.*/
	{
		wcsa->wordsData.wordsZoneMem.zone = NULL;
		wcsa->wordsData.words = NULL;
		
		/**version 3 of CANONICALS (compressed with huffman) */
		strcpy(filename, basename);
		strcat(filename, ".");
		strcat(filename, CANONICAL_WORDS_FILE_EXT3);
		
		loadCanonicalsFCHT(&wcsa->fcht, (char *) filename);
	}
#endif

	
	/** File with the variants of each canonical word. (sorted by frequency) */
	{	
		/**version 2 of variants */
		strcpy(filename, basename);
		strcat(filename, ".");
		strcat(filename, VARIANTS_OF_WORDS_FILE_EXT2);
		sizeFile= fileSize(filename) - sizeof(byte)*numberOfCanonicalWords;

		if( (file = open(filename, O_RDONLY)) < 0) {
			printf("Cannot open file %s\n", filename);
			exit(0);
		}
		
		loadVariants(&(wcsa->variants),file);
		close(file);

	}
				
	/** File with the data of the Huffman trees of each canonical word (and variants) */
	{ uint depth;
		strcpy(filename, basename);
		strcat(filename, ".");
		strcat(filename, HUFFWORD_CANONICAL_FILE_EXT);

		if( (file = open(filename, O_RDONLY)) < 0) {
			printf("Cannot open file %s\n", filename);
			exit(0);
		}
		loadHuffVariants(&(wcsa->HV), file);	
		close(file);	
	}	

		
	/** File with the separators. (sorted by frequency) .. compacted with char-based huffman */
	{	byte *zoneMem;
		uint number,depth;
		strcpy(filename, basename);
		strcat(filename, ".");
		strcat(filename, ASCII_OF_SEPARATORS_FILE_EXT);
		sizeFile= fileSize(filename) - sizeof(uint);
		if( (file = open(filename, O_RDONLY)) < 0) {
			printf("Cannot open file %s\n", filename);
			exit(0);
		}
		
		//the number of symbols in the vocabulary of separators 
		read(file, &number, sizeof(uint));
		wcsa->separators2.count = number;
		read(file, &depth, sizeof(uint));
		wcsa->separators2.huffman.depth = depth;
		wcsa->separators2.huffman.num = (uint *) malloc ((depth+1)*sizeof(uint));
		wcsa->separators2.huffman.fst = (uint *) malloc ((depth+1)*sizeof(uint));
		read(file, wcsa->separators2.huffman.num, sizeof(uint) * (depth+1));
		read(file, wcsa->separators2.huffman.fst, sizeof(uint) * (depth+1));

		loadSeparators(&(wcsa->separators2),file);
		close(file);		
	}	
	
	//** TOH: Compressed text 
	{	strcpy(filename, basename);
		strcat(filename, ".");
		strcat(filename, TOH_FILE_EXT);
		sizeFile= fileSize(filename)-sizeof(uint);
		
		if( (file = open(filename, O_RDONLY)) < 0) {
			printf("Cannot open file %s\n", filename);
			exit(0);
		}

		wcsa->TOH = (uint *) malloc (sizeof(uint) * ((sizeFile +4)/4));
		read(file, wcsa->TOH, sizeof(uint) * ((sizeFile +4)/4));		
		close(file);		
	}	

	
	//** B: Samples of offsets to the beginning of a code in toh
	{	strcpy(filename, basename);
		strcat(filename, ".");
		strcat(filename, B_COMPR_FILE_EXT);
		sizeFile= fileSize(filename);
		
		if( (file = open(filename, O_RDONLY)) < 0) {
			printf("Cannot open file %s\n", filename);
			exit(0);
		}
		wcsa->B = (uint *) malloc (sizeFile);
		wcsa->bSize = sizeFile / sizeof(uint);	//read(file, &zSize, sizeof(uint));          //size
		read(file, wcsa->B,  sizeFile);       	//the samples
		close(file);

		//wcsa->samplePeriodB = DEFAULT_SAMPLE_PERIOD_B; //loaded from constants...			
	}		

	//** Z: Samples of the offset in the original text of each integer in SE
	{	strcpy(filename, basename);
		strcat(filename, ".");
		strcat(filename, Z_SAMPLES_FILE_EXT);
		sizeFile= fileSize(filename);
		
		if( (file = open(filename, O_RDONLY)) < 0) {
			printf("Cannot open file %s\n", filename);
			exit(0);
		}
		wcsa->Z = (uint *) malloc (sizeFile);
		wcsa->zSize = sizeFile / sizeof(uint);	//read(file, &zSize, sizeof(uint));          //size
		read(file, wcsa->Z,  sizeFile);       	//the samples
		close(file);

		//wcsa->samplePeriodZ = DEFAULT_SAMPLE_PERIOD_Z; //loaded from constants...			
	}		
	wcsa->se= NULL;	

	free(filename);
}














/** ****************************************************************
  * Querying the index WCSA
  * ***************************************************************/
///////////////////////////////////////////////////////////////////////////////////////
//					   FUNCTIONS NEEDED FOR SEARCHING A PATTERN    					 //
///////////////////////////////////////////////////////////////////////////////////////

/*
void parseTextIntoIntegersPreviousToHuffman (twcsa *wcsa, byte *textPattern, uint patLen, uint *integerPattern, uint *sizeIntegers) {
	
	byte *pbeg,*pend,*wordstart,*aWord;
	register unsigned long size;
	tCanonical canonical;	
	uint index =0; 
		 	
	pbeg = textPattern; 
	pend = pbeg + patLen;
							
	while (pbeg <pend) {  
		//parsing a word or separator
		size=0;
		wordstart = pbeg;
		if (_Valid[*pbeg]) {   //alphanumerical data
			while ( (size<MAX_SIZE_OF_WORD) && (pbeg<=pend) && ( _Valid[*pbeg] )) {size++;pbeg++;}
		 }
		else {  // a separator
			while ( (size<MAX_SIZE_OF_GAP) && (pbeg<=pend)&& (!_Valid[*pbeg]) ) {size++;pbeg++;}
		}
		
		if ( pbeg >= pend ) {  //needed in case of the last word been parsed further away of cend. 
			size -= (pbeg-pend);		  			
			pbeg=pend;
		}					
		  		
		//The parsed word is "aWord", and its length is "size"...
		aWord=wordstart;																		
		if (_Valid[*aWord]) { //either a word or a stopword		
			register uint min,max,p;	
			uint len, tmplen;
			byte *wcsaWord;			
			byte *canonicWord;		
			
			getCanonical(&canonical, aWord,size, wcsa->stopwordshash);
			canonicWord = canonical.word;					
			size = canonical.size; // fari 22.05.2009 /
			
			if (canonical.type != IS_STOPWORD)  { // a "meaningfull word" (a non-stopword)
				
				// Binary search on the canonical words (wordsData)
				min = 0;
				max = (wcsa->numberOfCanonicalWords) - 1;
				while(min < max) {
					p = (min+max)/2;
					//					if(strcmp((char*)canonicWord, (char*)wcsa->wordsData[p].word) > 0) min = p+1;
					//					else max = p;
					{//preparing for strcompL
						len    = bitread (wcsa->wordsData.words, (p+1)* wcsa->wordsData.elemSize , wcsa->wordsData.elemSize);
						tmplen = bitread (wcsa->wordsData.words, ( p )* wcsa->wordsData.elemSize  , wcsa->wordsData.elemSize);
						len -=tmplen;
						wcsaWord= (byte *) wcsa->wordsData.wordsZoneMem.zone + tmplen;
					}
					
					//if(strncmp((char*)aWord, (char*)wcsa->wordsData[p].word,size) > 0) min = p+1;
					if(strcompL(canonicWord, wcsaWord, size, len) > 0) min = p+1;
					else max = p;
				}

			{//preparing for strcompL
				len    = bitread (wcsa->wordsData.words, (min+1)* wcsa->wordsData.elemSize, wcsa->wordsData.elemSize);
				tmplen = bitread (wcsa->wordsData.words, ( min )* wcsa->wordsData.elemSize, wcsa->wordsData.elemSize);
				len -=tmplen;
				wcsaWord= (byte *) wcsa->wordsData.wordsZoneMem.zone + tmplen;
			}
			if(!strcompL(canonicWord, wcsaWord, size, len)) {					
					integerPattern[index++] = min +1 ;  //<--					
			}
			else {*sizeIntegers = 0; return;}  // a valid word that does not appear in the source text.

			//				if(!strcmp((char*)aWord, (char*)wcsa->wordsData[min].word)) integerPattern[index++] = min +1;
			//				else {*sizeIntegers = 0; return;}  // a valid word that does not appear in the source text.
			
			}								
		}		
	}// end while
	*sizeIntegers = index;	
	
//	//shows the parsed words:
//	{uint i;
//		printf("\n\n >>%s>> HA SIDO PARSEADO COMO:",textPattern);
//		for (i=0; i<index;i++) {
//				printf("<<%s>>",wcsa->wordsData[integerPattern[i] -1].word);
//		}
//		
//	}	

		
}
*/







/** ****************************************************************
  * Querying the index WCSA
  * ***************************************************************/
///////////////////////////////////////////////////////////////////////////////////////
//					   FUNCTIONS NEEDED FOR SEARCHING A PATTERN    					 //
///////////////////////////////////////////////////////////////////////////////////////
#ifndef COMPRESS_FCHT_CANONICALS
void parseTextIntoIntegers(twcsa *wcsa, byte *textPattern, uint patLen, uint *integerPattern, uint *sizeIntegers) {

	byte *pbeg,*pend,*wordstart,*aWord;
	register unsigned long size;
	tCanonical canonical;	
	uint index =0; 
		 	
	pbeg = textPattern; 
	pend = pbeg + patLen;
							
	while (pbeg <pend) {  
		//parsing a word or separator
		size=0;
		wordstart = pbeg;
		if (_Valid[*pbeg]) {   //alphanumerical data
			while ( (size<MAX_SIZE_OF_WORD) && (pbeg<=pend) && ( _Valid[*pbeg] )) {size++;pbeg++;}
		 }
		else {  // a separator
			while ( (size<MAX_SIZE_OF_GAP) && (pbeg<=pend)&& (!_Valid[*pbeg]) ) {size++;pbeg++;}
		}
		
		if ( pbeg >= pend ) {  //needed in case of the last word been parsed further away of cend. 
			size -= (pbeg-pend);		  			
			pbeg=pend;
		}					
		  		
		//The parsed word is "aWord", and its length is "size"...
		aWord=wordstart;																		
		if (_Valid[*aWord]) { //either a word or a stopword		
			register uint min,max,p;	
			uint len, tmplen;
			byte *wcsaWord;			
			byte *canonicWord;		
			
			getCanonical(&canonical, aWord,size, wcsa->stopwordshash);
			canonicWord = canonical.word;					
			size = canonical.size; /** fari 22.05.2009 **/
			
			if (canonical.type != IS_STOPWORD)  { // a "meaningfull word" (a non-stopword)
				
				// Binary search on the canonical words (wordsData)
				min = 0;
				max = (wcsa->numberOfCanonicalWords) - 1;
				while(min < max) {
					p = (min+max)/2;
					//					if(strcmp((char*)canonicWord, (char*)wcsa->wordsData[p].word) > 0) min = p+1;
					//					else max = p;
#ifdef KEEP_CANONICALS_UNCOMPRESSED					
					{//preparing for strcompL
						len    = bitread (wcsa->wordsData.words, (p+1)* wcsa->wordsData.elemSize , wcsa->wordsData.elemSize);
						tmplen = bitread (wcsa->wordsData.words, ( p )* wcsa->wordsData.elemSize  , wcsa->wordsData.elemSize);
						len -=tmplen;
						wcsaWord= (byte *) wcsa->wordsData.wordsZoneMem.zone + tmplen;
					}
										
					if(strcompL(canonicWord, wcsaWord, size, len) > 0) min = p+1;					
					else max = p;
#endif				
#ifdef COMPRESS_HUFFMAN_CANONICALS	
 					if(strcompHuff(&(wcsa->HC), p, canonicWord , size) > 0 ) min = p+1;					
					else max = p;
#endif 					

				}

#ifdef KEEP_CANONICALS_UNCOMPRESSED
			{//preparing for strcompL
				len    = bitread (wcsa->wordsData.words, (min+1)* wcsa->wordsData.elemSize, wcsa->wordsData.elemSize);
				tmplen = bitread (wcsa->wordsData.words, ( min )* wcsa->wordsData.elemSize, wcsa->wordsData.elemSize);
				len -=tmplen;
				wcsaWord= (byte *) wcsa->wordsData.wordsZoneMem.zone + tmplen;
			}
			if(!strcompL(canonicWord, wcsaWord, size, len)) {					
					integerPattern[index++] = min +1 ;  //<--					
			}
			else {*sizeIntegers = 0; return;}  // a valid word that does not appear in the source text.
#endif
#ifdef COMPRESS_HUFFMAN_CANONICALS
			if(!strcompHuff(&(wcsa->HC), min, canonicWord , size) ) {
				integerPattern[index++] = min +1 ;  //<--
			}					
			else {*sizeIntegers = 0; return;} 
#endif

			}								
		}		
	}// end while
	*sizeIntegers = index;	
		
	//	//shows the parsed words:
	//	{uint i;
	//		printf("\n\n >>%s>> HA SIDO PARSEADO COMO:",textPattern);
	//		for (i=0; i<index;i++) {
	//				printf("<<%s>>",wcsa->wordsData[integerPattern[i] -1].word);
	//		}
	//		
	//	}	
	
}

#endif

#ifdef COMPRESS_FCHT_CANONICALS
void parseTextIntoIntegers(twcsa *wcsa, byte *textPattern, uint patLen, uint *integerPattern, uint *sizeIntegers) {

	byte *pbeg,*pend,*wordstart,*aWord;
	register unsigned long size;
	tCanonical canonical;	
	uint index =0; 
		 	
	pbeg = textPattern; 
	pend = pbeg + patLen;
							
	while (pbeg <pend) {  
		//parsing a word or separator
		size=0;
		wordstart = pbeg;
		if (_Valid[*pbeg]) {   //alphanumerical data
			while ( (size<MAX_SIZE_OF_WORD) && (pbeg<=pend) && ( _Valid[*pbeg] )) {size++;pbeg++;}
		 }
		else {  // a separator
			while ( (size<MAX_SIZE_OF_GAP) && (pbeg<=pend)&& (!_Valid[*pbeg]) ) {size++;pbeg++;}
		}
		
		if ( pbeg >= pend ) {  //needed in case of the last word been parsed further away of cend. 
			size -= (pbeg-pend);		  			
			pbeg=pend;
		}					
		  		
		//The parsed word is "aWord", and its length is "size"...
		aWord=wordstart;																		
		if (_Valid[*aWord]) { //either a word or a stopword		
			register uint min,max,p;	
			uint len, tmplen;
			byte *wcsaWord;			
			byte *canonicWord;		
			
			getCanonical(&canonical, aWord,size, wcsa->stopwordshash);
			canonicWord = canonical.word;					
			size = canonical.size; /** fari 22.05.2009 **/
			
			if (canonical.type != IS_STOPWORD)  { // a "meaningfull word" (a non-stopword)
			
				size_t id = getIDFCHT(wcsa->fcht, canonicWord, (ulong) size);				
				
				if (id != -1) {	
					integerPattern[index++] = id+1 ;  //<--
				}					
				else {*sizeIntegers = 0; return;} 

			}								
		}		
	}// end while
	*sizeIntegers = index;	
	
}
#endif













	/** ------------------------------------------------------------------
	 * Returns the number of occurrences of a given text pattern
	 *------------------------------------------------------------------ */
int countTextOcurrences(twcsa *wcsa, byte *textPattern) {
	ulong left, right;
	uint integerPatterns[MAX_INTEGER_PATTERN_SIZE];
	uint integerPatternSize, min, max;
	
	uint lenpat = strlen((char*)textPattern);
	parseTextIntoIntegers(wcsa, textPattern, lenpat, integerPatterns, &integerPatternSize);	
	if (!integerPatternSize) return -1;
		
//	#ifdef DEBUG_ON		
//		uint i;
//		printf("\n %d Integers to search for:",integerPatternSize );
//		for (i=0;i<integerPatternSize;i++) {
//			printf(" \n [%u] from word [%s]",integerPatterns[i], wcsa->wordsData[integerPatterns[i]-1].word);			
//		}	
//		printf("\n");
//	#endif
	

	ulong numocc;
	int err = countIntIndex((void *) wcsa->myicsa, integerPatterns, integerPatternSize, &numocc,  &left, &right);
	return numocc;
	//return countCSA(wcsa->myicsa, integerPatterns, integerPatternSize, &left, &right);
}


	/** ------------------------------------------------------------------
	 * locateTextOcurrences: 
	 * Returns the offsets of the source text where a word/phrase appears
	 * Returns also the number of occurrences.
	 *------------------------------------------------------------------ */
uint *locateTextOcurrences(twcsa *wcsa, byte *textPattern, int *numberOccurrences) {
	uint integerPatterns[MAX_INTEGER_PATTERN_SIZE];
	uint integerPatternSize, min, max;
	
	uint lenpat = strlen((char*)textPattern);
	parseTextIntoIntegers(wcsa, textPattern, lenpat, integerPatterns, &integerPatternSize);		
	if (!integerPatternSize) {*numberOccurrences = -1; return NULL;}
		
//	#ifdef DEBUG_ON		
//		uint i;
//		printf("\n %d Integers to search for:",integerPatternSize );
//		for (i=0;i<integerPatternSize;i++) {
//			printf(" \n [%u] from word [%s]",integerPatterns[i], wcsa->wordsData[integerPatterns[i]-1].word);			
//		}	
//		printf("\n");
//	#endif
		
	ulong occurrences, left, right;
	ulong *seOffsets;
	ulong *sourceOffsets;
	
	//obtains the indexes in vector SE where the pattern appears.
	//seOffsets = locateCSA(wcsa->myicsa, integerPatterns, integerPatternSize, &occurrences);
	int err = locateIntIndex((void *)wcsa->myicsa, integerPatterns, integerPatternSize, &seOffsets, &occurrences);
	
	//sourceOffsets = (uint *) malloc (sizeof(uint)*occurrences);
	
	sourceOffsets=seOffsets;
	//obtains the offsets in the source text of the pattern (sourceOffsets)
	locateFacade(wcsa, (uint *)sourceOffsets, (uint *)seOffsets,occurrences);

	#ifdef DEBUG_ON		
		fprintf(stderr,"\n*** %s appears in the source text in positions:\n\t",textPattern);
		for (i=0;i<occurrences;i++) 
			fprintf(stderr,"[%u]",sourceOffsets[i]);
		fflush(stderr);
	#endif

	//free(seOffsets);
	
	*numberOccurrences = occurrences;
	return (uint *)sourceOffsets;
}

	/** ------------------------------------------------------------------
	 * displayTextOcurrences:
	 * Shows in stdout, the text around the occurrences of a word/phrase
	 * Returns also the number of occurrences.
	 *------------------------------------------------------------------ */
int displayTextOcurrences(twcsa *wcsa, byte *textPattern, uint radixDisplay) {
	int err;
	uint integerPatterns[MAX_INTEGER_PATTERN_SIZE];
	uint integerPatternSize, min, max;
	
	uint lenpat = strlen((char*)textPattern);
	parseTextIntoIntegers(wcsa, textPattern, lenpat, integerPatterns, &integerPatternSize);		
	if (!integerPatternSize) {return -1;}
		
//	#ifdef DEBUG_ON		
//		uint i;
//		printf("\n %d Integers to search for:",integerPatternSize );
//		for (i=0;i<integerPatternSize;i++) {
//			printf(" \n [%u] from word [%s]",integerPatterns[i], wcsa->wordsData[integerPatterns[i]-1].word);			
//		}	
//		printf("\n");
//	#endif
		
	ulong  left, right, occurrences;
	ulong *seOffsets, *sourceOffsets;	
	
		//occurrences = countCSA(wcsa->myicsa, integerPatterns, integerPatternSize, &left, &right);
		//seOffsets  = locateCSA(wcsa->myicsa, integerPatterns, integerPatternSize, (uint *)&occurrences);
	err = countIntIndex ((void *)wcsa->myicsa, integerPatterns, integerPatternSize, &occurrences, &left, &right);
	err = locateIntIndex((void *)wcsa->myicsa, integerPatterns, integerPatternSize, &seOffsets, &occurrences);

	locateAllAndDisplay (wcsa, (uint *)seOffsets, occurrences, radixDisplay); 

	free(seOffsets);
	return occurrences;
}

	/** ------------------------------------------------------------------
	 * Locate Facade: 
	 * For given sePositions[], returns the sourceTextPositions
	 * where the those valid-words in se[sePositions[i]] occurr.
	 *------------------------------------------------------------------*/ 
int locateFacade (twcsa *wcsa, uint *sourceTextPositions,uint *sePositions, uint number) {
	uint i;	
	for (i=0;i<number;i++) {
		sourceTextPositions[i] = locateOne(wcsa, sePositions[i]);
	}
	return 0;
}


/** ------------------------------------------------------------------
	* DISPLAYFACADE:
	* Returns the subString from a starting offset to a final offset
	* in the source text. It does not allocate any memory, receives "dstptr"
	* Precondition: offsetIni >=0;
	------------------------------------------------------------------*/
 int displayFacade (twcsa *wcsa, uint offsetIni, uint offsetFin, ulong *length, byte *dstptr) {
	uint offsetZ;
	register uint ptr;   //index in TOH.
	uint alternate;
	THuffDec *huff;
	uint variant;
	
	int err;
	//uint indexSE;
	//int posSEValue;
	uint posSEValue; ulong indexSE;
	
	
	uint tptr,tptrold;  //offset in the "Original Plain Text"	
	uint i;
	
	byte *dst, *src;
	
	dst = dstptr;
	
	if (offsetIni < wcsa->Z[0]) {  //only NON-valid word"s" (a separator) before offsetIni
		ptr = 0;	
		huff = &(wcsa->separators2.huffman);	
		//ptr=decodeHuffDec(huff, &variant,wcsa->TOH,ptr); //decoding the separator before the first word (in Z[0]).		
		decodeHuffDecMacro(huff,variant,wcsa->TOH,ptr);
		{
			uint end;
			end = (offsetFin <wcsa->Z[0]) ? offsetFin : wcsa->Z[0]-1;

/** */			
			{ //src= (byte *)(offsetIni + wcsa->separators.words[variant]);
			src=asciiSeparator(&(wcsa->separators2), variant);
			}
/** */
			tptr=offsetIni;
			for ( ; tptr<=end; tptr++) *dst++ = *src++;			
			if (tptr>offsetFin) {//*dst='\0'; 
								 return 0;}  //offsetFin < Z[0]
		}
		// now: tptr =Z[0] //== end
		offsetZ = 0; //i=Z[0];  //the first valid word. (it is ranked "0").
		indexSE = 0;			//the first valid word. (it is ranked "0").
		offsetIni = wcsa->Z[0];	//the offset of  first valid word in the source text				
	}
	else { 
		offsetZ = binsearchUintArray(wcsa->Z, wcsa->zSize, offsetIni);
		indexSE = offsetZ * (wcsa->samplePeriodZ);
		tptr = wcsa->Z[offsetZ];		
		ptr = BSelect(wcsa, indexSE);  //ptr = (Bbitmap,indexSE+1);		
	}
	//printf("\n ptr = %d and tohsize = %d ",ptr,wcsa->tohSize);

	//now proceeds from Z[offsetZ] and skips text until offsetIni.
	alternate=1;
	tptrold=tptr;
	while (tptr < offsetIni) {
		if (alternate) {  // a valid word
			alternate =0;	
			tptrold=tptr;
			//posSEValue =displayCSA(wcsa->myicsa,indexSE); 
			err= displayIntIndex((void *)wcsa->myicsa,indexSE, &posSEValue);
			
			//huff = &(wcsa->wordsData[posSEValue-1].huffman);  
			//ptr = decodeHuffDec(huff, &variant,wcsa->TOH,ptr);				
		/**DEPTH2*///decodeHuffDecMacroVALIDWORDS(huff,variant,wcsa->TOH,ptr,huff->depth);				

				{//decodeHuffDecMacroVALIDWORDS(huff,variant,wcsa->TOH,ptr,wcsa->depthsHuffVariants[posSEValue-1]);
/**					uint posInZone = wcsa->HV.offsetNumAndFst[posSEValue-1];
					register uint hdepth = (wcsa->HV.offsetNumAndFst[posSEValue] - posInZone)/2;
					uint *posHuff= &(wcsa->HV.zoneNumFst[posInZone]);
					decodeHuffDecMacroVariantWord(posHuff, variant, wcsa->TOH, ptr,hdepth);
*/	
					//register uint posInZone = wcsa->HV.offsetNumAndFst[posSEValue-1];
					//register uint hdepth = (wcsa->HV.offsetNumAndFst[posSEValue] - posInZone)/2;							
					//decodeHuffDecMacroVariantWordPos(wcsa->HV.zoneNumFst, posInZone, variant, wcsa->TOH, ptr,hdepth); 
					
					//decodeHuffDecMacroVariantWordPos2bits(wcsa->HV, variant, wcsa->TOH, ptr, posSEValue-1);
					if (hasUniqueVariant (&(wcsa->variants),posSEValue -1)){
						variant = 0;
						//hasUniqueVariant
					}
					else 					
						ptr= decodeHuffDecFuncVariantWordPos2bits(wcsa->HV, &variant, wcsa->TOH, ptr, posSEValue-1); 
				}

				//decodeHuffDecMacroVALIDWORDS(huff,variant,wcsa->TOH,ptr);
				
				//tptr += strlen((char*)wcsa->wordsData[posSEValue-1].variants[variant]); 
			//ADDLEN(wcsa->wordsData[posSEValue-1].variants[variant],tptr);	
			//indexSE++;
			//src = wcsa->wordsData[posSEValue-1].variants[variant];

			{uint tmplen;
				//src=asciiVariant(&(wcsa->variants), posSEValue-1, variant);
				//ADDLEN(src, tptr);
				src=asciiVariantLen(&(wcsa->variants), posSEValue-1, variant, &tmplen);
				tptr+=tmplen;
			indexSE++;
			}
		}
		else {			  //a separator
			alternate =1;	
			tptrold=tptr;
			huff = &(wcsa->separators2.huffman);	
			//ptr = decodeHuffDec(huff, &variant,wcsa->TOH,ptr);
			decodeHuffDecMacro(huff,variant,wcsa->TOH,ptr);
				//tptr+=strlen((char*)wcsa->separators.words[variant]);
			//ADDLEN(wcsa->separators.words[variant],tptr);				 
			//src=wcsa->separators.words[variant];
/** */			
			{	uint tmplen;
				//ADDLEN(wcsa->separators.words[variant],tptr);				 
				//src=wcsa->separators.words[variant];
				src=asciiSeparatorLen(&(wcsa->separators2), variant,&tmplen);
				tptr+=tmplen;
			}
/** */

		}	
	}
	
	//now it holds either (tptr == offsetIni) or (tptr >offsetIni)
	//tptr is the position in the text AFTER:
	//		 the last decoded "word" or
	//		 the last decoded "char" (from a non valid word)
	//oldptr is the position in the text of the last "valid word".

	//  if tptr == offsetIni --> all data before offsetIni has already been skiped.
	//  if tptr >  offsetIni --> offsetIni belongs to a valid word, so we
	//       have to process it to extract all chars from offsetIni	
	//       (having into account than offsetFin could also be inside the same valid word)	

	if (tptr > offsetIni) { //copies a part of either the word or separator from offsetIni.
			//fprintf(stderr,"\entro 111: tprt = %u, texto = <<%s>>",tptr,src);
		byte *tmp; 
		uint counter; //number of bytes to copy from current word
		
		//src = wordsData[SE[indexSE-1]].variants[variant];
		//src = wordsData[displayCSA(indexSE-1) -1].variants[variant];      /*****/  //14 agosto: src set before !!
		
		src += (offsetIni - tptrold);	 //moves pointer inside the current variant.
		counter = (tptr>offsetFin)? offsetFin-offsetIni+1 : tptr-offsetIni ;
		for (i=1; i<=counter; i++) 	*dst++ = *src++; //copies that part.		
	}
	
	
	//if offsetFin has not already been overpassed.. 
	//... we continue decoding and outputting data until offsetFin.
	uint len; 
	while (tptr <=offsetFin) {		
		//if ((indexSE >=wcsa->seSize)||(ptr >= wcsa->tohSize)) break;
		if (alternate) {  // a valid word
			alternate =0;	
			tptrold=tptr;
			//posSEValue =displayCSA(wcsa->myicsa,indexSE); // SE[indexSE]
			err= displayIntIndex((void *)wcsa->myicsa,indexSE, &posSEValue);

			//huff = &(wcsa->wordsData[posSEValue-1].huffman);
			//ptr = decodeHuffDec(huff, &variant,wcsa->TOH,ptr);
				
		/**DEPTH2*///decodeHuffDecMacroVALIDWORDS(huff,variant,wcsa->TOH,ptr,huff->depth);
				{//decodeHuffDecMacroVALIDWORDS(huff,variant,wcsa->TOH,ptr,wcsa->depthsHuffVariants[posSEValue-1]);
/**					uint posInZone = wcsa->HV.offsetNumAndFst[posSEValue-1];
					register uint hdepth = (wcsa->HV.offsetNumAndFst[posSEValue] - posInZone)/2;
					uint *posHuff= &(wcsa->HV.zoneNumFst[posInZone]);
					decodeHuffDecMacroVariantWord(posHuff, variant, wcsa->TOH, ptr,hdepth);
*/	
					//register uint posInZone = wcsa->HV.offsetNumAndFst[posSEValue-1];
					//register uint hdepth = (wcsa->HV.offsetNumAndFst[posSEValue] - posInZone)/2;							
					//decodeHuffDecMacroVariantWordPos(wcsa->HV.zoneNumFst, posInZone, variant, wcsa->TOH, ptr,hdepth); 
					//decodeHuffDecMacroVariantWordPos2bits(wcsa->HV, variant, wcsa->TOH, ptr, posSEValue-1);					
					if (hasUniqueVariant (&(wcsa->variants),posSEValue -1)){
						variant = 0;
						//hasUniqueVariant
					}
					else											
						ptr= decodeHuffDecFuncVariantWordPos2bits(wcsa->HV, &variant, wcsa->TOH, ptr, posSEValue-1); 
				}
				//decodeHuffDecMacroVALIDWORDS(huff,variant,wcsa->TOH,ptr);
				
			//len = strlen ((char*)wcsa->wordsData[posSEValue-1].variants[variant]);  
			//STRLEN( wcsa->wordsData[posSEValue-1].variants[variant], len);
			len = lenVariant(&(wcsa->variants), posSEValue-1, variant);
			tptr += len;
			
			if (tptr >offsetFin) len = offsetFin-tptrold+1; //copies only until offsetFin
				//src = wcsa->wordsData[posSEValue-1].variants[variant];	

			asciiVariantCopyBytes (dst,&(wcsa->variants), posSEValue-1, variant, len); 
			dst+=len;
			//src=asciiVariant(&(wcsa->variants), posSEValue-1, variant);
			//for (i=0; i<len; i++) *dst++ = *src++; //copies the word.		
			indexSE++;			
		}		
		else {			  //a separator
			alternate =1;	
			tptrold=tptr;
			huff = &(wcsa->separators2.huffman);	
			//ptr = decodeHuffDec(huff, &variant,wcsa->TOH,ptr);			
			decodeHuffDecMacro(huff,variant,wcsa->TOH,ptr);
				//len = strlen((char*)wcsa->separators.words[variant]);
			//STRLEN( wcsa->separators.words[variant], len);
			//tptr+= len;

			len = lenSeparator(&(wcsa->separators2), variant);
			tptr += len;

			if (tptr >offsetFin) len = offsetFin-tptrold+1; //copies only until offsetFin			

			asciiSeparatorCopyBytes (dst,&(wcsa->separators2), variant, len); 
			dst+=len;

			//src=wcsa->separators.words[variant];
			//for (i=0; i<len; i++) *dst++ = *src++; //copies the word.							
		}				
	}	
	//*dst='\0'; 
	*length= dst-dstptr;
	return 0;		
}


	/**------------------------------------------------------------------
	 * DISPLAYFacadeMalloc:
	 * Returns the subString from a starting offset to a final offset
	 * in the source text. It allocates Memory !!
	 * NOT CURRENTLY USED
	 ------------------------------------------------------------------*/
 byte *displayFacadeMalloc (twcsa *wcsa, uint offsetIni, uint offsetFin, ulong *length) {
	uint offsetZ;
	uint ptr;   //index in TOH.
	uint alternate;
	THuffDec *huff;
	uint variant;
	
	//uint indexSE; int posSEValue;
	uint posSEValue; ulong indexSE;

		
	uint tptr,tptrold;  //offset in the "Original Plain Text"	
	uint i;
	
	byte *dstptr;
	byte *dst;
	byte *src;
	
	dstptr= (byte *) malloc (sizeof(byte)* (offsetFin+1-offsetIni+1));
	dst = dstptr;
	int err;
	
	if (offsetIni < wcsa->Z[0]) {  //only NON-valid word"s" (a separator) before offsetIni
		ptr = 0;	
		huff = &(wcsa->separators2.huffman);	
		//ptr=decodeHuffDec(huff, &variant,wcsa->TOH,ptr); //decoding the separator before the first word (in Z[0]).		
		decodeHuffDecMacro(huff,variant,wcsa->TOH,ptr);
		{
			uint end;
			end = (offsetFin <wcsa->Z[0]) ? offsetFin : wcsa->Z[0]-1;

/** */			
			{ //src= (byte *)(offsetIni + wcsa->separators.words[variant]);
			src=asciiSeparator(&(wcsa->separators2), variant);
			}
/** */
			tptr=offsetIni;
			for ( ; tptr<=end; tptr++) *dst++ = *src++;			
			if (tptr>offsetFin) {//*dst='\0'; 
								 return 0;}  //offsetFin < Z[0]
		}
		// now: tptr =Z[0] //== end
		offsetZ = 0; //i=Z[0];  //the first valid word. (it is ranked "0").
		indexSE = 0;			//the first valid word. (it is ranked "0").
		offsetIni = wcsa->Z[0];	//the offset of  first valid word in the source text				
	}
	else { 
		offsetZ = binsearchUintArray(wcsa->Z, wcsa->zSize, offsetIni);
		indexSE = offsetZ * (wcsa->samplePeriodZ);
		tptr = wcsa->Z[offsetZ];		
		ptr = BSelect(wcsa, indexSE);  //ptr = bselect(Bbitmap,indexSE+1);
	}

	//now proceeds from Z[offsetZ] and skips text until offsetIni.
	alternate=1;
	tptrold=tptr;
	while (tptr < offsetIni) {
		if (alternate) {  // a valid word
			alternate =0;	
			tptrold=tptr;
			//posSEValue =displayCSA(wcsa->myicsa,indexSE); 
			err= displayIntIndex((void *)wcsa->myicsa,indexSE, &posSEValue);

			
			//huff = &(wcsa->wordsData[posSEValue-1].huffman);  
			//ptr = decodeHuffDec(huff, &variant,wcsa->TOH,ptr);
				
		/**DEPTH2*///decodeHuffDecMacroVALIDWORDS(huff,variant,wcsa->TOH,ptr,huff->depth);
				{//decodeHuffDecMacroVALIDWORDS(huff,variant,wcsa->TOH,ptr,wcsa->depthsHuffVariants[posSEValue-1]);
/**					uint posInZone = wcsa->HV.offsetNumAndFst[posSEValue-1];
					register uint hdepth = (wcsa->HV.offsetNumAndFst[posSEValue] - posInZone)/2;
					uint *posHuff= &(wcsa->HV.zoneNumFst[posInZone]);
					decodeHuffDecMacroVariantWord(posHuff, variant, wcsa->TOH, ptr,hdepth);
*/	
					//register uint posInZone = wcsa->HV.offsetNumAndFst[posSEValue-1];
					//register uint hdepth = (wcsa->HV.offsetNumAndFst[posSEValue] - posInZone)/2;							
					//decodeHuffDecMacroVariantWordPos(wcsa->HV.zoneNumFst, posInZone, variant, wcsa->TOH, ptr,hdepth); 
					//decodeHuffDecMacroVariantWordPos2bits(wcsa->HV, variant, wcsa->TOH, ptr, posSEValue-1);
					if (hasUniqueVariant (&(wcsa->variants),posSEValue -1)){
						variant = 0;
						//hasUniqueVariant
					}
					else 	
						ptr= decodeHuffDecFuncVariantWordPos2bits(wcsa->HV, &variant, wcsa->TOH, ptr, posSEValue-1); 

				}				//decodeHuffDecMacroVALIDWORDS(huff,variant,wcsa->TOH,ptr);
				
				//tptr += strlen((char*)wcsa->wordsData[posSEValue-1].variants[variant]); 
			//ADDLEN(wcsa->wordsData[posSEValue-1].variants[variant],tptr);	
			//indexSE++;
			//src = wcsa->wordsData[posSEValue-1].variants[variant];

			{uint tmplen;
				//src=asciiVariant(&(wcsa->variants), posSEValue-1, variant);
				//ADDLEN(src, tptr);
				src=asciiVariantLen(&(wcsa->variants), posSEValue-1, variant, &tmplen);
				tptr+=tmplen;
			indexSE++;
			}
		}
		else {			  //a separator
			alternate =1;	
			tptrold=tptr;
			huff = &(wcsa->separators2.huffman);	
			//ptr = decodeHuffDec(huff, &variant,wcsa->TOH,ptr);
			decodeHuffDecMacro(huff,variant,wcsa->TOH,ptr);
				//tptr+=strlen((char*)wcsa->separators.words[variant]);
			//ADDLEN(wcsa->separators.words[variant],tptr);				 
			//src=wcsa->separators.words[variant];
/** */			
			{	uint tmplen;
				//ADDLEN(wcsa->separators.words[variant],tptr);				 
				//src=wcsa->separators.words[variant];
				src=asciiSeparatorLen(&(wcsa->separators2), variant,&tmplen);
				tptr+=tmplen;
			}
/** */

		}	
	}
	
	//now it holds either (tptr == offsetIni) or (tptr >offsetIni)
	//tptr is the position in the text AFTER:
	//		 the last decoded "word" or
	//		 the last decoded "char" (from a non valid word)
	//oldptr is the position in the text of the last "valid word".

	//  if tptr == offsetIni --> all data before offsetIni has already been skiped.
	//  if tptr >  offsetIni --> offsetIni belongs to a valid word, so we
	//       have to process it to extract all chars from offsetIni	
	//       (having into account than offsetFin could also be inside the same valid word)	

	if (tptr > offsetIni) { //copies a part of either the word or separator from offsetIni.
			//fprintf(stderr,"\entro 111: tprt = %u, texto = <<%s>>",tptr,src);
		byte *tmp; 
		uint counter; //number of bytes to copy from current word
		
		//src = wordsData[SE[indexSE-1]].variants[variant];
		//src = wordsData[displayCSA(indexSE-1) -1].variants[variant];      /*****/  //14 agosto: src set before !!
		
		src += (offsetIni - tptrold);	 //moves pointer inside the current variant.
		counter = (tptr>offsetFin)? offsetFin-offsetIni+1 : tptr-offsetIni ;
		for (i=1; i<=counter; i++) 	*dst++ = *src++; //copies that part.		
	}
	
	
	//if offsetFin has not already been overpassed.. 
	//... we continue decoding and outputting data until offsetFin.
	uint len; 
	while (tptr <=offsetFin) {		
		if (alternate) {  // a valid word
			alternate =0;	
			tptrold=tptr;
			//posSEValue =displayCSA(wcsa->myicsa,indexSE); // SE[indexSE]
			err= displayIntIndex((void *)wcsa->myicsa,indexSE, &posSEValue);

			
			//huff = &(wcsa->wordsData[posSEValue-1].huffman);
			//ptr = decodeHuffDec(huff, &variant,wcsa->TOH,ptr);
				
		/**DEPTH2*///decodeHuffDecMacroVALIDWORDS(huff,variant,wcsa->TOH,ptr,huff->depth);
				{//decodeHuffDecMacroVALIDWORDS(huff,variant,wcsa->TOH,ptr,wcsa->depthsHuffVariants[posSEValue-1]);
/**					uint posInZone = wcsa->HV.offsetNumAndFst[posSEValue-1];
					register uint hdepth = (wcsa->HV.offsetNumAndFst[posSEValue] - posInZone)/2;
					uint *posHuff= &(wcsa->HV.zoneNumFst[posInZone]);
					decodeHuffDecMacroVariantWord(posHuff, variant, wcsa->TOH, ptr,hdepth);
*/	
					//register uint posInZone = wcsa->HV.offsetNumAndFst[posSEValue-1];
					//register uint hdepth = (wcsa->HV.offsetNumAndFst[posSEValue] - posInZone)/2;							
					//decodeHuffDecMacroVariantWordPos(wcsa->HV.zoneNumFst, posInZone, variant, wcsa->TOH, ptr,hdepth); 
					//decodeHuffDecMacroVariantWordPos2bits(wcsa->HV, variant, wcsa->TOH, ptr, posSEValue-1);

					if (hasUniqueVariant (&(wcsa->variants),posSEValue -1)){
						variant = 0;
						//hasUniqueVariant
					}
					else 	
						ptr= decodeHuffDecFuncVariantWordPos2bits(wcsa->HV, &variant, wcsa->TOH, ptr, posSEValue-1); 
				}
				//decodeHuffDecMacroVALIDWORDS(huff,variant,wcsa->TOH,ptr);
				
			//len = strlen ((char*)wcsa->wordsData[posSEValue-1].variants[variant]);  
			//STRLEN( wcsa->wordsData[posSEValue-1].variants[variant], len);
			len = lenVariant(&(wcsa->variants), posSEValue-1, variant);
			tptr += len;
			
			if (tptr >offsetFin) len = offsetFin-tptrold+1; //copies only until offsetFin
				//src = wcsa->wordsData[posSEValue-1].variants[variant];	

			asciiVariantCopyBytes (dst,&(wcsa->variants), posSEValue-1, variant, len); 
			dst+=len;
			//src=asciiVariant(&(wcsa->variants), posSEValue-1, variant);
			//for (i=0; i<len; i++) *dst++ = *src++; //copies the word.		
			indexSE++;			
		}		
		else {			  //a separator
			alternate =1;	
			tptrold=tptr;
			huff = &(wcsa->separators2.huffman);	
			//ptr = decodeHuffDec(huff, &variant,wcsa->TOH,ptr);			
			decodeHuffDecMacro(huff,variant,wcsa->TOH,ptr);
				//len = strlen((char*)wcsa->separators.words[variant]);
			//STRLEN( wcsa->separators.words[variant], len);
			//tptr+= len;

			len = lenSeparator(&(wcsa->separators2), variant);
			tptr += len;

			if (tptr >offsetFin) len = offsetFin-tptrold+1; //copies only until offsetFin			

			asciiSeparatorCopyBytes (dst,&(wcsa->separators2), variant, len); 
			dst+=len;

			//src=wcsa->separators.words[variant];
			//for (i=0; i<len; i++) *dst++ = *src++; //copies the word.							
		}				
	}	
	*dst='\0'; 
	*length= dst-dstptr;
	return dstptr;		
}


///////////////////////////////////////////////////////////////
//  VARIOUS FUNCTIONS CALLED BY OTHERS IN THIS MODULE
///////////////////////////////////////////////////////////////
/*------------------------------------------------------------------ 
 * Binary search for a uint in a uint vector V
 * returns: an index such that V[index] <=value.
 *         or -1 if value < V[0].
  ------------------------------------------------------------------*/
int binsearchUintArray(uint *V, uint vSize, register uint value) {
	register uint l, r, p;
	
	if (value < V[0]) return -1;
    l = 0;
    r = vSize-1;
    while(l <= r) {
    	p = (l+r)/2;
		if (V[p]<value)
			l = p+1;
		else if (V[p]>value)
			r = p-1;
		else 
			return p;  //the value was found.
	}
	return r; //The value was not in A, so returns index such that Z[index]<"value"
}






















/** ****************************************************************
  * destroyStructsFacade.
  *	Frees the memory used by the Facade (Presentation Layer).
  ******************************************************************/
void destroyStructsFacade(twcsa *wcsa) {
	uint i;

	//the canonical words.

	//free (wcsa->canonicalsZoneMem.zone);
	//free (wcsa->wordsData); /** huge!! */
	if (wcsa->wordsData.wordsZoneMem.zone) free (wcsa->wordsData.wordsZoneMem.zone);
	if (wcsa->wordsData.words) free (wcsa->wordsData.words); /** huge!! */	
#ifdef COMPRESS_HUFFMAN_CANONICALS
	freeCanonicals (&(wcsa->HC));
#endif
#ifdef COMPRESS_FCHT_CANONICALS
	freeCanonicalsFCHT(wcsa->fcht);
#endif

	//the Huffman trees of each canonical
	freeHuffVariants(&(wcsa->HV));
	//All the variants
	freeVariants(&(wcsa->variants));
	//All the separators
	freeSeparators(&(wcsa->separators2));
	freeHuffDec(wcsa->separators2.huffman);
	
	//TOH, B, and Z
	free(wcsa->TOH);
	free(wcsa->B);
	free(wcsa->Z);
}



/** FUNCTIONS NEEDED DURING SEARCHES */

/** ------------------------------------------------------------------
	* BSelect
	* Gives the offset in TOH of the code of the i^th valid word
	* ----------------------------------------------------------------- */
uint BSelect(twcsa *wcsa, uint i){		
	uint remainder =i%(wcsa->samplePeriodB);
	uint sample    =i/(wcsa->samplePeriodB);
	if (!remainder) {
		return wcsa->B[sample];
	}		
	//uncompressing "remainder" valid words...
	{
		int err;
		register uint variant, ptr;
		uint alternate =0,counter=0;	
		uint posSEValue,  posSE;
		THuffDec *huff;
		
		posSE =sample * (wcsa->samplePeriodB);
		ptr=wcsa->B[sample];
		alternate=1;  //we start in a sampled word.		
		huff = &(wcsa->separators2.huffman);

		while (1) {
			if (alternate) {  // a valid word
				alternate =0;				
				counter++;
				if (counter >remainder) break;
				//posSEValue =displayCSA(wcsa->myicsa,posSE)-1; /*****/
				err= displayIntIndex((void *)wcsa->myicsa,posSE, &posSEValue); posSEValue--;

				//huff = &(wcsa->wordsData[posSEValue].huffman);				
				//ptr = decodeHuffDec(huff, &variant,wcsa->TOH,ptr);
				
				{				
					//decodeHuffDecMacroVariantWordPos(wcsa->HV.zoneNumFst, posInZone, variant, wcsa->TOH, ptr,hdepth); 
					//decodeHuffDecMacroVariantWordPos2bits(wcsa->HV, variant, wcsa->TOH, ptr, posSEValue);
					if (hasUniqueVariant (&(wcsa->variants),posSEValue )){
						variant = 0;
						//hasUniqueVariant
					}
					else 	
						ptr= decodeHuffDecFuncVariantWordPos2bits(wcsa->HV, &variant, wcsa->TOH, ptr, posSEValue); 
				}
				
				posSE++;				 				
			}
			else {			  //a separator
				alternate =1;	
				//huff = &(wcsa->separators2.huffman);	
				//ptr = decodeHuffDec(huff, &variant,wcsa->TOH,ptr);												
				decodeHuffDecMacro(huff,variant,wcsa->TOH,ptr);

				//if (ptr >=wcsa->tohSize) break; //not needed, as we always start BSelect at a valid word.	
			}
		}			
		return ptr;	
	}
}

/** ------------------------------------------------------------------
	* Starting at a given offset in TOH, skipping counting the first one, 
	* uncompresses the TOH stream up to the beginning of the 
	* n-ary "valid" word after startOffset.
	* Returns the number of "decoded bytes"
	* Warnings: it does not check "end-of-bitmap".
	* ------------------------------------------------------------------*/
 uint unCompressNValidWords (twcsa *wcsa, uint posSE, uint startOffset, int n){
	int counter=0; 
	uint variant;
	//int posSEValue;
	uint posSEValue;

	int counterBytes= 0;
	uint ptr;
	THuffDec *huff;	
	uint alternate;	
		
	int err;
	alternate=1; // starting in a codeword that belongs to a valid word
	ptr=startOffset;	
	while (1) {
		if (alternate) {  // a valid word
			alternate =0;				
			counter++;
			if (counter > n) break;
				
			//posSEValue =displayCSA(wcsa->myicsa,posSE)-1; /*****/
			err= displayIntIndex((void *)wcsa->myicsa,posSE, &posSEValue); posSEValue--;
			
			//huff = &(wcsa->wordsData[posSEValue].huffman);				
			//ptr = decodeHuffDec(huff, &variant,wcsa->TOH,ptr);
				
		/**DEPTH2*///decodeHuffDecMacroVALIDWORDS(huff,variant,wcsa->TOH,ptr,huff->depth);
				{//decodeHuffDecMacroVALIDWORDS(huff,variant,wcsa->TOH,ptr,wcsa->depthsHuffVariants[posSEValue-1]);
/**					uint posInZone = wcsa->HV.offsetNumAndFst[posSEValue];
					register uint hdepth  = (wcsa->HV.offsetNumAndFst[posSEValue+1]-posInZone)/2;
					uint *posHuff= &(wcsa->HV.zoneNumFst[posInZone]);
					decodeHuffDecMacroVariantWord(posHuff, variant, wcsa->TOH, ptr,hdepth);
*/
					//register uint posInZone = wcsa->HV.offsetNumAndFst[posSEValue];
					//register uint hdepth = (wcsa->HV.offsetNumAndFst[posSEValue+1] - posInZone)/2;							
					//decodeHuffDecMacroVariantWordPos(wcsa->HV.zoneNumFst, posInZone, variant, wcsa->TOH, ptr,hdepth); 
					//decodeHuffDecMacroVariantWordPos2bits(wcsa->HV, variant, wcsa->TOH, ptr, posSEValue);

					if (hasUniqueVariant (&(wcsa->variants),posSEValue )){
						variant = 0;
						//hasUniqueVariant
					}
					else 	
						ptr= decodeHuffDecFuncVariantWordPos2bits(wcsa->HV, &variant, wcsa->TOH, ptr, posSEValue); 
				}

				//decodeHuffDecMacroVALIDWORDS(huff,variant,wcsa->TOH,ptr);
				
				//counterBytes+=strlen((char*)wcsa->wordsData[posSEValue].variants[variant]);				
			//ADDLEN(wcsa->wordsData[posSEValue].variants[variant],counterBytes);	
			counterBytes += lenVariant(&(wcsa->variants), posSEValue, variant);
			posSE++;				 				
		}
		else {			  //a separator
			alternate =1;	
			huff = &(wcsa->separators2.huffman);	
			//ptr = decodeHuffDec(huff, &variant,wcsa->TOH,ptr);				
			decodeHuffDecMacro(huff,variant,wcsa->TOH,ptr);
				//counterBytes+=strlen((char*)wcsa->separators.words[variant]);
			//ADDLEN(wcsa->separators.words[variant],counterBytes);	
			counterBytes += lenSeparator(&(wcsa->separators2), variant);			
			
			if (ptr >=wcsa->tohSize) break;
		}							
	}//while(1)	
	return counterBytes;
}


	/** ------------------------------------------------------------------
	 * LOCATEOne:
	 * Obtains the position (in the source text) of word ranked "indexInSE"
	 * in SE (a given valid word)
	 * ------------------------------------------------------------------*/
uint locateOne (twcsa *wcsa, uint indexInSE) {
	uint startTohOffset;
	register uint sampledZ,gapFromSampled;
	uint offset;
	
	sampledZ = indexInSE / (wcsa->samplePeriodZ);
	gapFromSampled = indexInSE % (wcsa->samplePeriodZ);
			
	if (!gapFromSampled) // it was a sampled word.
		return wcsa->Z[sampledZ];  		
	else {	//it was not a sampled word...		
			//1st obtains the offset in B of the code for the previous sampled word		
		startTohOffset = BSelect(wcsa,indexInSE-gapFromSampled);  
		offset = wcsa->Z[sampledZ]; //location of the previous sampled word.		 
			//2nd now uncompressing to count the number of bytes till the searched word.
		offset+= unCompressNValidWords (wcsa,indexInSE-gapFromSampled, startTohOffset , gapFromSampled);	
		return offset;	
	}	
}

	/** ------------------------------------------------------------------ 
	 * LOCATEALLandDISPLAY:
	 * Displays the text around an occurrence of the searched word in the source text.
	 * Assuming that $p$ is that position --> shows only chars in [p_radix-1,p_radix]
	 ------------------------------------------------------------------*/ 
int locateAllAndDisplay (twcsa *wcsa, uint *sePositions, uint number, int radix) {
	register uint i;
	int position;
	int ini,end;
	register uint sourceTextSize = wcsa->sourceTextSize;
	
	printf("\n\n ...** Displaying text around occurrences with radix %d ...\n",radix);
	for (i=0;i<number;i++) {
		position = locateOne(wcsa, sePositions[i]);
		ini= ((position -radix) >=0) ? position-radix : 0;
		end= ((position +radix) >= sourceTextSize) ? sourceTextSize-1 : position+radix;		
		fprintf(stderr,"\n call to DisplayFac centered in %d with radix %d: [%d,%d]\n",position,radix,ini,end);fflush(stderr);
		ulong length;
		byte *text = displayFacadeMalloc(wcsa,ini,end,&length);
//		fprintf(stderr,"\n[%2d][len=%3d] <<%s>>",i,strlen((char*)text),text);		fflush(stderr);
		fprintf(stderr,"\n[%2d][len=%3lu] <<%s>>",i,length,text);		fflush(stderr);
		free(text);					
	}
	printf("\n\n ... ** End LocateAllandDisplay.. \n");	
}


	/** ------------------------------------------------------------------
	 * recovers the source text by calling display(0,fileSize);
	 * ------------------------------------------------------------------ */
void recoverSourceText1(twcsa *wcsa, char *basename, char *ext, uint sourceTextSize) {		
	int start;int end;
	byte *cc;
	char *filename = (char *) malloc (strlen( basename)+ strlen(ext)+1);
	ulong length;
	
	strcpy( filename,  basename);
	strcat( filename, ext);
	filename[strlen( basename)+ strlen(ext)]='\0';
	fprintf(stderr,"\n uncompressing text into file -->%s ",filename);fflush(stderr);

	FILE *salida;
	unlink( filename);
	salida = fopen( filename,"w");
	start=0; end = sourceTextSize-1;
	
	//cc= displayFacadeMalloc(wcsa,start,end,&length);		
	//void *Index = (void *) wcsa;
	//extract (Index, start, end, &cc, &length);	 

	cc = (byte *) malloc (sourceTextSize* sizeof(uchar));
	ulong leng;
	displayFacade(wcsa,start,end,&leng,cc);
	fprintf(stderr,"\n sourceTextSize = %d, len = %lu",sourceTextSize,leng);
	//fprintf(salida,"%s",cc);
	fwrite(cc,sizeof(byte),leng,salida);
	fclose(salida);

	free(cc);
	free(filename);
}

	/** ------------------------------------------------------------------
	 * recovers the source text by calling display (1 byte at a time)
	 * ------------------------------------------------------------------ */		
void recoverSourceText2(twcsa *wcsa, char *basename, char *ext, uint sourceTextSize) {		
	FILE *salida; int i; int end;
	byte cc;
	ulong length;
	char *filename = (char *) malloc (strlen( basename)+ strlen(ext)+1);
	strcpy( filename,  basename);
	strcat( filename, ext);
	filename[strlen( basename)+ strlen(ext)]='\0';

	fprintf(stderr,"\n uncompressing text into file -->%s ",filename);fflush(stderr);	
	unlink( filename);
	salida = fopen( filename,"w");
	end = sourceTextSize-1;
	for (i=0;i<=end;i++) {		
		if (! (i% (1<<16))) {fprintf(stderr,"\ndecoding 1 byte at a time. progress %u of %u.",i,end);}
		displayFacade(wcsa,i,i,&length,&cc);
		fwrite(&cc,1,1,salida);
	}	
	fclose(salida);	
	free(filename);			
}




/** *******************************************************************************
  * Showing some statistics and info of the index
  * *******************************************************************************/
		/* Shows summary info of the index */
int printInfo(void *index) {
	twcsa *wcsa = (twcsa *) index;
	tVariants *variants = &(wcsa->variants);
	tSeparators2 * separators = &(wcsa->separators2);

	unsigned long indexSize;
	uint intIndexSize, presentationSize;
	int err;
	
	err = index_size(index, &indexSize);
	if (err!=0) return err;
	err = sizeIntIndex(wcsa->myicsa, &intIndexSize); 	
	if (err!=0) return err;
	
	presentationSize = indexSize - intIndexSize;
	
		printf("\n ===================================================:");		
		printf("\n Summary of Presentation layer:");		
		printf("\n   Number of valid words (SEsize) = %u",wcsa->seSize);
		printf("\n   Size of vectors: TOH= %u (%d bytes), B = %u (%d bytes)",wcsa->tohSize, sizeof(uint) * (((wcsa->tohSize)+W-1)/W), wcsa->bSize, sizeof(uint) * (wcsa->bSize ));
		printf("\n   Size of vectors :Z %u (%u bytes)\n",wcsa->zSize, sizeof(uint) * (wcsa->zSize ));
		printf("\n   CAnonical words = %ld",wcsa->numberOfCanonicalWords);		
		printf("\n   Sample periodZ = %u",wcsa->samplePeriodZ);

		printf("\n   WCSA structure = %d bytes", sizeof(twcsa));
		printf("\n   Hash table of stopwords = %lu bytes", stmemused(wcsa->stopwordshash));

#ifdef KEEP_CANONICALS_UNCOMPRESSED
		uint totalcanonicals = ((((wcsa->numberOfCanonicalWords+1)* (wcsa->wordsData.elemSize))+W-1) /W) * (sizeof(uint))+
								wcsa->wordsData.wordsZoneMem.size * sizeof(byte);

		printf("\n   Size Of canonicals structure (%u bytes):",totalcanonicals);
		printf("\n      [     pointers = %lu bytes || AsciiZone = %d bytes",
					((((wcsa->numberOfCanonicalWords+1)* (wcsa->wordsData.elemSize))+W-1) /W) * (sizeof(uint)),
					wcsa->wordsData.wordsZoneMem.size * sizeof(byte) );	
#endif
#ifdef COMPRESS_HUFFMAN_CANONICALS
		uint totalcanonicals = sizeCanonicals(&(wcsa->HC));
		printf("\n   Size Of canonicals structure, Compressed with Huffman-chars (%u bytes):",totalcanonicals);
		printf("\n      [     pointers = %u bytes || TOHzone = %d bytes",
					((((wcsa->HC.nwords+1)* (wcsa->HC.elemSize))+W-1) /W) * (sizeof(uint)),
					((wcsa->HC.sizeZoneHuff +W-1)/W * sizeof(uint) ));	
		printf("\n            maxLenCanonical= %13d + HuffTreeChars = %13d", 
				( ((wcsa->HC.maxLenAsciiCanonical) +1) * sizeof(byte)), sizeHuff (wcsa->HC.H)  ); 
#endif

#ifdef COMPRESS_FCHT_CANONICALS
		uint totalcanonicals = sizeCanonicalsFCHT(wcsa->fcht);
		printf("\n   Size Of canonicals structure, Compressed with Front-coding Hu-tucker (%u bytes):",totalcanonicals);
		printf("\n           Block-size for Front-coding Hu-tucker set to (%u words):",wcsa->blocksizeFCHuTucker);
#endif
				
		printf("\n   Size Of variants structure (%d bytes):",sizeVariants(variants));
		printf("\n      [     Dirb +     Subb +  HuffZone + maxLenVariant + HuffTreeChars]");
		printf("\n      [ %8d + %8d + %9d + %13d + %13d]", 
							( (((variants->dirbSize)+W-1)/W) * sizeof (uint)), 
							( (((variants->subbSize)+W-1)/W) * sizeof (uint)), 
							( ((variants->sizeZoneHuff + W-1)/W) * sizeof(uint)),
							( ((variants->maxLenAsciiVariant) +1) * sizeof(byte)),
							sizeHuff (variants->H));
	
		printf("\n   Size of Huffman trees of all the variants = %d bytes", sizeHuffVariants(&(wcsa->HV)));
		printf("\n         [Huff.fstbits/int = %4d, Huff.numbits/int = %4d ]", wcsa->HV.sizeFstbits, 	wcsa->HV.sizeNumbits);
		printf("\n   Size Of Separators structure (%d bytes):",sizeSeparators(separators));
		printf("\n      [     Dirb +  HuffZone + maxLenSep + HuffTreeSeps]");
		printf("\n      [ %8d + %9d + %9d + %12d]", 
							( (((separators->dirbSize)+W-1)/W) * sizeof (uint)) ,
							( ((separators->sizeZoneHuff + W-1)/W) * sizeof(uint)) ,
							( ((separators->maxLenSep) +1) * sizeof(byte)) , 
							sizeHuff (separators->H));
		printf("\n   Huffman tree of separators (word-oriented) = %d bytes",2 * ((wcsa->separators2.huffman.depth+1)*sizeof(uint)));						
		printf("\n\n Total = **  %u bytes (in RAM) **",presentationSize);		
		//printf("\n\n @@ Summary of self-index on Integers:");
		err = printInfoIntIndex(wcsa->myicsa, " ");
		if (err!=0) return err;
 			
		printf("\n ===================================================:");		
		printf("\n");
		return 0;
	}



/*
int sizeSeparators(tSeparators2 *separators) {
	return  ( (((separators->dirbSize)+W-1)/W) * sizeof (uint)) + 
			( ((separators->sizeZoneHuff + W-1)/W) * sizeof(uint)) +
			( ((separators->maxLenSep) +1) * sizeof(byte)) + 
			sizeHuff (separators->H);
}
*/

//shows a char* of len "len" not ended in '\0'
void printWord(uchar *str, uint len) {
		uint i;
		fflush(stdout);fflush(stderr);
		fprintf(stderr,"<<<<<");
		for (i=0;i<len;i++)
			fprintf(stderr,"%c",str[i]);
		fprintf(stderr,">>>>>");
		fflush(stdout);fflush(stderr);
}

void printInfoReduced(twcsa *wcsa) {
/*
	{
		printf("\n Summary:");		
		printf("\n  Number of valid words = %ld",wcsa->seSize);
		printf("\n  Size of vectors: TOH= %ld",wcsa->tohSize);
		printf("\n  Size of  B= %ld, SE = %ld, Z=%ld, validWords =%ld\n",wcsa->bSize,wcsa->seSize,wcsa->zSize,wcsa->seSize);
		printf("  TOH:");
		showBitVector(wcsa->TOH,wcsa->tohSize);
		printf("\n    B:");
		showBitVector(wcsa->B,wcsa->bSize);
		printf("\n");
	}
		
		//slo para imprimir un "1" cada 10 posiciones del bitmap (sirve como referencia)
		printf("\n      ");
		{int i;
			for (i=0;i<wcsa->bSize;i++) if (!(i%10)) printf("1"); else printf(" ");						
		}	
			
		printf("\n  Z   ");	
		{int i;
			for (i=0;i<wcsa->zSize;i++) printf("%ld ",wcsa->Z[i]);						
		}
*/
//		printf("\n  SE  ");
//		{int i;
//			for (i=0;i<seSize;i++) printf("%ld ",SE[i]);						
//		}								
//		printf("   countValidWords is  = %ld\n",seSize);			
		
//		printf("\n  words in SE ");
//		{int i;
//			for (i=0;i<seSize;i++) printf("[%s] ",wordsData[SE[i]].word);						
//		}								

//		printf("\n  canonicalwords  ");
//		{int i;
//			for (i=0;i<wcsa->numberOfCanonicalWords;i++) printf("[%s] ",wcsa->wordsData[i].word);						
//		}								
//		printf("   numberOfCanonicalWords is  = %ld\n",wcsa->numberOfCanonicalWords);	
//	

	
}

/**------------------------------------------------------------------
 * structsSize.
 *	Counts the memory amount needed by the Facade (Presentation Layer).
 * skipping the stop_words hash table 
 ----------------------------------------------------------------- */
/*
uint structsSizeMem(twcsa *wcsa) {
	uint i;
	uint total=0;

	//the words
#ifdef KEEP_CANONICALS_UNCOMPRESSED		
	//total += wcsa->numberOfCanonicalWords * sizeof(twords);  //number
	//total += wcsa->canonicalsZoneMem.size;

	total += ((((wcsa->numberOfCanonicalWords+1)* (wcsa->wordsData.elemSize))+W-1) /W) * sizeof(uint);
	total += wcsa->wordsData.wordsZoneMem.size * sizeof(byte) ;
	
#endif
#ifdef COMPRESS_HUFFMAN_CANONICALS		//huffman over the canonical words.
	total += sizeCanonicals(&(wcsa->HC));
#endif
#ifdef COMPRESS_FCHT_CANONICALS		//FT-HU-TUCKER over the canonical words.
	total += sizeCanonicalsFCHT(wcsa->fcht)
#endif	
	
	//huffman of variants
	total += sizeHuffVariants(&(wcsa->HV));
	total += sizeVariants(&(wcsa->variants));
	total += sizeSeparators(&(wcsa->separators2));

	total += (wcsa->tohSize) /8;
	total += (wcsa->bSize) /8;
	total += (wcsa->zSize)* sizeof (uint);
	return total;					
}

uint structsSizeDisk(twcsa *wcsa) {
	uint i;
	uint total=0;

	//the words
#ifdef KEEP_CANONICALS_UNCOMPRESSED		
	total += wcsa->numberOfCanonicalWords * sizeof(twords);  //number
	total += wcsa->canonicalsZoneMem.size;
#endif
#ifdef COMPRESS_HUFFMAN_CANONICALS	
	total += sizeCanonicals(&(wcsa->HC));
#endif
#ifdef COMPRESS_FCHT_CANONICALS		//FT-HU-TUCKER over the canonical words.
	total += sizeCanonicalsFCHT(wcsa->fcht)
#endif	
	
	//huffman of variants
	total += sizeHuffVariants(&(wcsa->HV));
	total += sizeVariants(&(wcsa->variants));
	total += sizeSeparators(&(wcsa->separators2));


	total += (wcsa->tohSize) /8;
	total += (wcsa->bSize) /8;
	total += (wcsa->zSize)* sizeof (uint);
	return total;					
}
*/
	/** saves the content of the file SE (ids of the source words) **/
int saveSEfile (char *basename, uint *v, uint n) {
	char outfilename[255];
	int file;
	sprintf(outfilename,"%s.%s",basename,SE_FILE_EXT);
	unlink(outfilename);
	if( (file = open(outfilename, O_WRONLY|O_CREAT,S_IRWXG | S_IRWXU)) < 0) {
		printf("Cannot open file %s\n", outfilename);
		exit(0);
	}

	write(file, v, sizeof(uint) * n ); 
	close(file);
}





/**------------------------------------------------------------------ 
  *  MAIN PROGRAM.
  *------------------------------------------------------------------ */
#ifdef FACADEWITHMAIN
	int main(int argc, char* argv[])
	{
		
			
	//{
	
	// 	uint gapsFreqCount[256];
	// 	THuff gapsHuffman;
	//	uint i;
	//	uint *bitvector;
	//	uint bitvectorSize;
	//	uint ptr;
	//	bitmap bG,bE;
	//	uint len;
	//	len = 1000; //number of bits
	//	bitvector = (uint *) malloc ((len/32 +1)* sizeof(uint));
	//	
	//	byte texto[100] = "LA_BELLA_ROSA_ROJA_DE_ROSA,_ES_BELLA_PERO_NO_ROSA";
	//	uint freqs[256];
	//	
	//	//fprintf(stderr,"\n este es el texto a codificar: %s",texto);
	//	for (i=0;i<256;i++) freqs[i]=0;
	//	for (i=0;i<strlen(texto);i++) freqs[texto[i]]++;
	//	gapsHuffman = createHuff (freqs,255,UNSORTED);
	//		
	//	ptr=0;
	//	for (i=0;i<strlen(texto);i++) {
	//		fprintf(stderr,"\n ENCODING seprators !!\n");
	//		fprintf(stderr,"%d. \n",ptr=encodeHuff(gapsHuffman, texto[i],bitvector,ptr) );		
	//	}		
	//}	
	
		
		char *infile, *outbasename, *stopwordsfile;	// Name of in/out files
		byte *inputBuffer;
		ulong finsize;
	
	 	int f_in;
		void *Index;

		
		printf("\n*Presentation level for CSA");
		printf("\n*CopyRight (c) 2007 [LBD & G.N.]\n\n");
	
		// Reads input parameters from command line.
		if(argc < 4) {
			printf("Use: %s <in file> <out basename> <stopwords file> \n", argv[0]);
			exit(0);
		}
	
		// Reads params (input file, output basename, and stopwords file)
		infile = argv[1];
		outbasename = argv[2];
		stopwordsfile = argv[3];
		
		finsize= fileSize(infile);
		
		if (! finsize) {
			printf( "\nFILE EMPTY OR FILE NOT FOUND %s !!\nSkipping processement ...\n",infile);
			exit(0);
		}	
	
		// Opening the input text file.
		if( (f_in = open(infile, O_RDONLY)) < 0) {
			printf("Cannot read file %s\n", infile);
			exit(0);
		}	
		inputBuffer = (byte *) malloc(finsize *sizeof(byte));// +1);
		read (f_in,inputBuffer,finsize);	
		close (f_in);	
		
		
	{
		//printf("\n parametros <<%s>>\n\n",stopwordsfile);	
		//	build_WCSA (inputBuffer, finsize, stopwordsfile, NULL,outbasename);								
		build_index (inputBuffer, finsize, stopwordsfile, &Index);  /** building the index */

		/** recovering the source text from the index */
			{
				double start, end;
				start = getTime2();
				ulong size;
				get_length(Index, &size);
			//	recoverSourceText1((twcsa*) Index, outbasename,".source", size);
			//	recoverSourceText2((twcsa*) Index, outbasename,".source2",size);
				end = getTime2();	
				fprintf(stderr, "\nRecovering source file time: %.3f secs\n", end-start );	
			}
		
		// DISPLAYING THE OCCURRENCES OF A TEXT PATTERN (word/phrase).
			{unsigned char textPattern[MAX_TEXT_PATTERN_SIZE];
			 int error = 0;
			ulong numocc,numc, length, i, *snippet_len, tot_numcharext = 0, numpatt;
			uchar *pattern, *snippet_text;
				 
				 pattern = textPattern;
			 printf("\nSEARCH TEST for DISPLAY (pizzachili interface)\n");
				while(1) {	
					printf("Intro string: ");
					fgets((char*)textPattern, MAX_TEXT_PATTERN_SIZE, stdin);
					if (!strcmp((char*)textPattern,"\n") ) break;
					 textPattern[strlen((char*)textPattern)-1] = '\0';
		
					length = strlen( (char*)textPattern);
					numc=50;
         
//					error =	display (Index, textPattern, length, numc, &numocc, 
//							    	 &snippet_text, &snippet_len);
					error =	displayWords (Index, textPattern, length, numc, &numocc, 
							    	 &snippet_text, &snippet_len,1);
					
					if (error){ fprintf(stderr, "%s\n", "Hubo un error durante display");exit(0);}
		
						fprintf(stderr,"\n acabou display");fflush(stderr);			
					{//show the results
						ulong j, len = length + 2*numc;
					    char blank = '\0';
						fprintf(stderr,"\n length = %d",length);
						fprintf(stderr,"\n pattern = %s",pattern);fflush(stderr);
						fprintf(stderr,"\n numocc = %d",numocc);fflush(stderr);
						fprintf(stderr,"\n snippet len = %d",len);fflush(stderr);
						fprintf(stderr,"\n =========");fflush(stderr);		
						for (i = 0; i < numocc; i++){
							fprintf(stderr,"\n[%2d][len=%3d]<<",i+1,snippet_len[i]);fflush(stderr);
							fwrite(snippet_text+len*i,sizeof(uchar),snippet_len[i],stderr);fflush(stderr);
							fprintf(stderr,">>");fflush(stderr);
						}
					}
					numpatt--;
					
					for(i=0; i<numocc; i++) {
						tot_numcharext += snippet_len[i];
					}
						
					if (numocc) {
						free (snippet_len);
						free (snippet_text);
					}
					
					printf("Ocurrences = %d\n", numocc);
					if (!strcmp((char*)textPattern,"\n") ) break;
				}
			}
	


	// SEARCHING FOR A TEXT PATTERN (word/phrase).
	{unsigned char textPattern[MAX_TEXT_PATTERN_SIZE];
	 int occ;
	 int len;
	 uint *occs;
	 int i;
	 printf("\nSEARCH TEST for LOCATE\n");
		while(1) {	
			printf("Intro string: ");
			fgets((char*)textPattern, MAX_TEXT_PATTERN_SIZE, stdin);
			len = strlen((char*)textPattern);
			if (!strcmp((char*)textPattern,"\n") ) break;
			textPattern[len-1] = '\0';
			len --;
			
			//occs = locateTextOcurrences(wcsa,textPattern,&occ);
			// locate(Index, textPattern, len, (ulong **)&occs, (ulong *)&occ);
			  locateWord(Index, textPattern, len, (ulong **)&occs, (ulong *)&occ, 0);
			
			printf("\n*** %s occurs %d times: In the source text in positions:\n\t",textPattern,occ);
/* 			for (i=0;i<occ;i++) 
				printf("[%u]",occs[i]);
			fflush(stderr);	 */
			if (occ >0) free(occs);		
			
			if (!strcmp((char*)textPattern,"\n") ) break;
		}
	}	
	

		/** saving the index to disk*/
		
		save_index (Index, outbasename);		
		fprintf(stderr,"Index saved !! ");

		/** tells the mem used by the index */
		ulong indexsize;		
		index_size(Index, &indexsize);
		fprintf(stderr,"Index occupied %d bytes, 2 extra mallocs = %d",indexsize,2* sizeof(uint));

		{twcsa *wcsa = (twcsa*) Index;
		fprintf(stderr," 2 extra mallocs = %d", 2* sizeof(uint) *  (wcsa->numberOfCanonicalWords)) ;fflush(stderr);
		}

		/** freeing the index */		 			
		free_index(Index);	 			
	 		 
	}
}
	
#endif





























//{	
//	bG = createBitmap (B,len);
//	bE = createBitmapEdu (B,len);
//	//fprintf(stderr,"\n RANK DE GONZALO DA %u, de EDU DA %u\n",rank(bG,33),rank1Edu(bE,33));
//	//fprintf(stderr,"\n SELECT1(2) DE GONZALO DA %u, de EDU DA %u\n",bselect(bG,4),bselect(bE,4));
//	
//	showBitVector(bitvector,34);	
//}



/*	

	//USING A HASH TABLE
	//	{
	//	char a[20]="beginnings";char b[20]="HOSTIAS";
	//	char *w;	
	//	int i;
	//	w=a;
	//	i = inHashTable(stopwordshash,w, strlen(w), &addrInTH );
	//	if (!i) insertElement (stopwordshash, w, strlen(w), &addrInTH);
	//	else stopwordshash->hash[addrInTH].freq++;
	//	//fprintf(stderr,"\n i = %ld, addrInTh = %ld ",i,addrInTH);
	//	//fprintf(stderr,"\n word in hash[%ld]= %s , freq = %ld, posinvoc =%ld",addrInTH, stopwordshash->hash[addrInTH].word, stopwordshash->hash[addrInTH].freq, stopwordshash->hash[addrInTH].posInVoc);	
	//  }



	/// ENCODING THE separators  ...
{	
	freeHuff(gapsHuffman);
	uint i;
	uint *bitvector;
	uint bitvectorSize;
	uint ptr;
	bitmap bG,bE;
	uint len;
	len = 1000; //number of bits
	bitvector = (uint *) malloc ((len/32 +1)* sizeof(uint));
	
	byte texto[100] = "####@?*";
	uint freqs[256];
	
	//fprintf(stderr,"\n este es el texto a codificar: %s",texto);
	for (i=0;i<256;i++) freqs[i]=0;
	for (i=0;i<strlen(texto);i++) freqs[texto[i]]++;
	gapsHuffman = createHuff (freqs,255,UNSORTED);
		
	ptr=0;
	for (i=0;i<strlen(texto);i++) {
		//fprintf(stderr,"\n ENCODING seprators !!\n");
		//fprintf(stderr,"%d. \n",ptr=encodeHuff(gapsHuffman, texto[i],bitvector,ptr) );		
	}	

	prepareToDecode(&(gapsHuffman));
	bitvectorSize = ptr;
	showBitVector(bitvector,bitvectorSize);	
	uint pos;
	//fprintf(stderr,"\n DECODING !!\n");
	ptr=0;
	while (ptr < bitvectorSize) {
		ptr=decodeHuff (gapsHuffman, &pos, bitvector, ptr);
		//fprintf(stderr,"\n DECODING pos is %ld!!\n",pos);
		//fprintf(stderr,"%c. \n",pos);
	}
	exit(0);	
}															
	
/// ENCODING THE CANONICAL WORDS ...
{	uint i;
	uint *bitvector;
	uint bitvectorSize;
	uint ptr;
	bitmap bG,bE;
	uint len;
	len = 1000; //number of bits
	bitvector = (uint *) malloc ((len/32 +1)* sizeof(uint));
	
	
	ptr=0;
	//fprintf(stderr,"\n ENCODING VARIANTS !!\n");
	//fprintf(stderr,"%d. \n",ptr=encodeHuff(posInHT[0].huffman, 0,bitvector,ptr) );
	//fprintf(stderr,"%d. \n",ptr=encodeHuff(posInHT[0].huffman, 1,bitvector,ptr) );
	//fprintf(stderr,"%d. \n",ptr=encodeHuff(posInHT[0].huffman, 2,bitvector,ptr) );
	//fprintf(stderr,"%d. \n",ptr=encodeHuff(posInHT[0].huffman, 3,bitvector,ptr) );
	//fprintf(stderr,"%d. \n",ptr=encodeHuff(posInHT[0].huffman, 4,bitvector,ptr) );
	//fprintf(stderr,"%d. \n",ptr=encodeHuff(posInHT[0].huffman, 4,bitvector,ptr) );
			
	//	FILE *f;
	//	f = fopen("huff","w");
	//	saveHuff(posInHT[0].huffman,f);
	//	fclose(f);
	//	
	//	f = fopen("huff","r");
	//	posInHT[0].huffman = loadHuff (f,0);
	//	fclose(f);
	//	
	
	prepareToDecode(&(posInHT[0].huffman));

	FILE *f;
	f = fopen("huff","w");
	saveHuffAfterDecode(posInHT[0].huffman,f);
	fclose(f);
	
	f = fopen("huff","r");
	//posInHT[0].huffman = loadHuffAfterDecode(f,0);
	loadHuffAfterDecode2 (&(posInHT[0].huffman),f,0);
	fclose(f);
	


	bitvectorSize = ptr;
	showBitVector(bitvector,bitvectorSize);	
	uint pos;
	//fprintf(stderr,"\n DECODING !!\n");
	ptr=0;
	while (ptr < bitvectorSize) {
		ptr=decodeHuff (posInHT[0].huffman, &pos, bitvector, ptr);
		//fprintf(stderr,"\n DECODING pos is %ld!!\n",pos);
		//fprintf(stderr,"%s. \n",posInHT[0].variants[pos]);
	}
	exit(0);
}	
	
	
			
{	uint i;
	uint *bitvector;
	bitmap bG,bE;
	uint len;
	len = 101; //number of bits
	bitvector = (uint *) malloc ((len/32 +1)* sizeof(uint));
	//bitvector[0]=0;
	//bitvector[1]=0;
	bitzero (bitvector,0,101-1);
	for (i=0; i<len;i++) setBit (bitvector,len,i,0);
	 


	bitset(bitvector,1); 
	bitset(bitvector,10);
	bitset(bitvector,12);
	//activateBit(bitvector,1); 
	//activateBit(bitvector,10); 
	
	bG = createBitmap (bitvector,len);
	bE = createBitmapEdu (bitvector,len);
	//fprintf(stderr,"\n RANK DE GONZALO DA %u, de EDU DA %u\n",rank(bG,33),rank1Edu(bE,33));
	//fprintf(stderr,"\n SELECT1(2) DE GONZALO DA %u, de EDU DA %u\n",bselect(bG,4),bselect(bE,4));
	
	showBitVector(bitvector,34);
}	
*/		
	
