/* buildFacade.h
 * Copyright (C) 2011, Antonio Fariña, all rights reserved.
 *
 * BuildFacade.h: Definition of the functions that the word-based self-index
 *   must implement following the "../pizza/interface.h" guidelines. We also 
 *   define here the data types required by the WCSA self-index, as well as the
 *   twcsa type (that will be referred to as an opaque type by "../pizza" programs.
 *
 * See more details in:
 * Antonio Fariña, Nieves Brisaboa, Gonzalo Navarro, Francisco Claude, Ángeles Places, 
 * and Eduardo Rodríguez. Word-based Self-Indexes for Natural Language Text. ACM 
 * Transactions on Information Systems (TOIS), 2012. 
 * http://vios.dc.fi.udc.es/indexing/wsi/publications.html
 * http://www.dcc.uchile.cl/~gnavarro/ps/tois11.pdf	  

 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */


#include <sys/time.h>
#include <sys/resource.h>

#include "utils/valstring.h"
#include "utils/defValues.h"
#include "utils/MemoryManager.h"
#include "utils/fileInfo.h"

#include "utils/hash.h"
#include "utils/hashStopWords.h"

#include "utils/huff.h"
#include "utils/parameters.h"

//from SEARCHER FACADE 
#include "utils/filterWord.h"
#include "utils/huffDec.h"
//#include "icsa/icsa.h"
#include "intIndex/interfaceIntIndex.h"

#include "FCHT/FCHTinterface.h"

#ifndef uchar
#define uchar unsigned char
#endif
#ifndef uint
#define uint  unsigned int
#endif
#ifndef ulong
#define ulong unsigned long
#endif

#define STRLEN(str,len) \
{len=0; \
 byte *ptr = str; \
 while(*ptr++) len++; \
}

#define ADDLEN(str,len) \
{byte *ptr = str; \
	while(*ptr++) len++; \
}

/** Some data types used **ONLY**during construction process */

	typedef struct {   //aux structure used to sort vectors (for words of separators)
		uint slot;     //depending on the values of the associated hash table.
		uint freq;
	} tauxslots;

//  Separators dataStructure.
	typedef struct {
		uint *slots;   //vector of pointers to the hash table
		uint count;	   //its size
		THuff huffman; //Huffman tree of separators
		t_hash hash;   //Hash table
		uint totallen; //the number of bytes occupied by all the "separators"
	} tSepsBuild;
	
// Words, both the canonical words and their variants
	typedef struct {
		unsigned long slot;   // the position in the hash table of the canonical word
		unsigned int number;
		byte *word;   //the canonical word
		//unsigned long len;    //of the canonical word
		byte **variants;  		// array of variants;
		//unsigned long *lens;	// the size of the variants.
		unsigned long *freqs;	// the freq of the variants.
		unsigned long *slots;   // the position in the hash table of any variant		
		THuff huffman;          //stores the Huffman data needed to encode and decode.
	} tposInHT;


/** Some data types used during searches */
	//  Separators dataStructure.
// 	typedef struct {
// 		uint count;	   //its size
// 		THuffDec huffman; //Huffman tree of separators
// 		byte **words;   //vector of words
// 	} tSeparators;
	
	typedef struct {
		uint count;		// # of separators in the word-oriented Huffman tree of seps.
		THuffDec huffman; //Huffman tree of separators (word-oriented)

		uint *Dirb;		
		uint dirbSize; //in bits
		uint dirbElemSize; // kbits (offset into zoneMem)
		
		uint *zoneMemHuff; 
		uint sizeZoneHuff; //in bits	
		uint maxLenSep; //the max len in bytes of a separator :: temp zone !!!
		byte *decodeBuffer;	//a buffer for decoding variants.
		THuff H;
	} tSeparators2;
	
	
    // **********************************************************	
	//  Canonical words dataStructure.	
	typedef struct SzoneMem {        //a large block of memory to load a file into mem.
		byte *zone;  //block of mem.
		uint size;  //number of bytes	 	 
	} tZoneMem;

	typedef struct {
		uint *words;   // kbitVector. 1 extra bucket to point to the end
		uint elemSize;  //the size (in bits) of each pointer.
		tZoneMem wordsZoneMem; // a block of memory where the canonical words are loaded (from file).
	} twords;  // '\0' char is not needed at the end. --> len[i] = words[i+1] - words[i];

#ifdef COMPRESS_HUFFMAN_CANONICALS
	typedef struct {
		uint nwords;    //number of canonical words.
		uint *words;   // kbitVector. 1 extra bucket to point to the end
		uint elemSize;  //the size (in bits) of each pointer.
		uint *zoneMemHuff;
		uint sizeZoneHuff;  //size in bits!!
		
		uint maxLenAsciiCanonical; //the max len in bytes of a canonical word :: temp zone !!!
		byte *decodeBuffer;	//a buffer for decoding canonicals.		
		THuff H;
	} tCanonicalsHuff;                // '\0' char is not needed at the end. --> len[i] = words[i+1] - words[i];
#endif	
    // **********************************************************	



	typedef struct {		
		//t_kBitArray Dirb;  /** 1st level. Offsets to variants through kbitvector */
		uint *Dirb;
		uint dirbSize; //in bits
		uint dirbElemSize; // kbits (offset into zoneMem)

		uint *Subb;
		uint subbSize;	//in bits
		uint subbAbsSize; // kbits (offset into zoneMem)
		uint subbJumpSize; // kbits (bits to represent a jump)

		uint *zoneMemHuff;
		uint sizeZoneHuff; //in bits	
		uint maxLenAsciiVariant; //the max len in bytes of a variant :: temp zone !!!
		byte *decodeBuffer;	//a buffer for decoding variants.
		THuff H;
	}tVariants;

 	typedef struct {
 		uint *Dirb;    //Directory: offsets into zoneMem			
 		uint dirbSize; //in bits
 		uint dirbElemSize; // kbits (offset into zoneMem)
 		
 		uint *zoneMem;  // [Fst|Num][Fst|Num]...
 		uint zoneMemSize; //in bits
 		uint sizeBuckbits; //in bits	
 		uint sizeFstbits; //in bits	
 		uint sizeNumbits; //in bits	
 	} tHuffVariants;


	/**the WCSA index structures... */
	typedef struct {
		stt_hash stopwordshash;	/* A hash table with the stopwords */	
		char *stopwordsfilename;// not needed during searches !! (only for save_index) 
			
		/**valid words */
		unsigned long numberOfCanonicalWords;  /* number of different canonical words. */				

		//needed for compression even if HUFFMAN_CANONICALS is SWITCH ON.
		twords wordsData;		/* Canonical words and variants (and Huffman trees) */
		tZoneMem canonicalsZoneMem; // a block of memory where the canonical words are loaded (from file).
#ifdef COMPRESS_HUFFMAN_CANONICALS
		tCanonicalsHuff HC;
#endif

#ifdef COMPRESS_FCHT_CANONICALS
		void *fcht;
		uint blocksizeFCHuTucker;
#endif
		/**Huffman of variants and variants */
		tHuffVariants HV;
		tVariants variants;

		/** separators */
		tSeparators2 separators2;
			
		/* The bitmaps and other vectors holding data to recover the source text. */
		uint *TOH;    //the compressed data.
		uint tohSize;
		uint *B;      //bitmap that indicates the beginning of a valid word in TOH 
		uint bSize;   //it's size
		uint *Z;      //samples of the original text
		uint zSize;   //it's size = (tohSize + (DEFAULT_SAMPLE_PERIOD_Z -1)) /DEFAULT_SAMPLE_PERIOD_Z
		uint samplePeriodZ;
		uint samplePeriodB;	
		uint seSize;  //it's size == "numberOfValidWords". 
		uint sourceTextSize; //the size of the source text.
		void *myicsa; //the WiCSA on SE words
	
		//#ifndef CSA_ON
		uint *se;
		//#endif
		
	}twcsa;


/** ******************************************************************************
    * Interface (from pizza chili) for using the WCSA index
*********************************************************************************/

/* Error management */

        /* Returns a string describing the error associated with error number
          e. The string must not be freed, and it will be overwritten with
          subsequent calls. */

char *error_index (int e);

/* Building the index */

        /* Creates index from text[0..length-1]. Note that the index is an 
          opaque data type. Any build option must be passed in string 
          build_options, whose syntax depends on the index. The index must 
          always work with some default parameters if build_options is NULL. 
          The returned index is ready to be queried. */

int build_index (uchar *text, ulong length, char *build_options, void **index);

        /*  Saves index on disk by using single or multiple files, having 
          proper extensions. */

int save_index (void *index, char *filename);

        /*  Loads index from one or more file(s) named filename, possibly 
          adding the proper extensions. */

int load_index (char *filename, void **index);

        /* Frees the memory occupied by index. */

int free_index (void *index);

        /* Gives the memory occupied by index in bytes. */

int index_size(void *index, ulong *size);

/* Querying the index */

        /* Writes in numocc the number of occurrences of the substring 
          pattern[0..length-1] found in the text indexed by index. */

int count (void *index, uchar *pattern, ulong length, ulong *numocc);

        /* Writes in numocc the number of occurrences of the substring 
          pattern[0..length-1] in the text indexed by index. It also allocates
          occ (which must be freed by the caller) and writes the locations of 
          the numocc occurrences in occ, in arbitrary order.  */

int locate (void *index, uchar *pattern, ulong length, ulong **occ, 
        ulong *numocc);

        /* Gives the length of the text indexed */

int get_length(void *index, ulong *length);

/* Accessing the indexed text  */

        /*  Allocates snippet (which must be freed by the caller) and writes 
          the substring text[from..to] into it. Returns in snippet_length the 
          length of the text snippet actually extracted (that could be less 
          than to-from+1 if to is larger than the text size). */

int extract (void *index, ulong from, ulong to, uchar **snippet, 
        ulong *snippet_length);

        /* Displays the text (snippet) surrounding any occurrence of the 
          substring pattern[0..length-1] within the text indexed by index. 
          The snippet must include numc characters before and after the 
          pattern occurrence, totalizing length+2*numc characters, or less if 
          the text boundaries are reached. Writes in numocc the number of 
          occurrences, and allocates the arrays snippet_text and 
          snippet_lengths (which must be freed by the caller). The first is a 
          character array of numocc*(length+2*numc) characters, with a new 
          snippet starting at every multiple of length+2*numc. The second 
          gives the real length of each of the numocc snippets. */

int display (void *index, uchar *pattern, ulong length, ulong numc, 
        ulong *numocc, uchar **snippet_text, ulong **snippet_lengths);

        /*  Obtains the length of the text indexed by index. */

int length (void *index, ulong *length);

		/* Shows summary info of the index */
int printInfo(void *index);

/** *******************************************************************************************/
/** Building part of the index ****************************************************************/
//int build_WCSA (uchar *text, ulong lenth, char *build_options, void **index, char *outbasename);
int build_WCSA (uchar *text, ulong length, char *stopwordsfile, void **index);
int build_iCSA (char  *build_options, void *index); 

/** *******************************************************************************************/
/** Search part of the index ******************************************************************/
// Definitions of some PUBLIC function prototipes.

		//loading/freeing the data structures into memory.
	
    void loadStructs(twcsa *wcsa, char *basename);	
	twcsa *loadWCSA(char *filename);	
	void destroyStructsFacade(twcsa *wcsa);
	
		//Gives the offset in TOH of the code of the i^th valid word
	uint BSelect(twcsa *wcsa, uint i);
		
 		//uncompress nValid words starting in a given offset in TOH. Returns the size of the decoded text		
	uint unCompressNValidWords (twcsa *wcsa, uint posSE, uint startOffset, int n);

		//locate for a unique word/phrase
	uint locateOne (twcsa *wcsa, uint indexInSE);  

		//returns the source text from given [offsetIni, offsetFin] offsets.
	//byte *displayFacade (twcsa *wcsa, uint offsetIni, uint offsetFin);
 	byte *displayFacadeMalloc (twcsa *wcsa, uint offsetIni, uint offsetFin, ulong *length);	 
	int displayFacade (twcsa *wcsa, uint offsetIni, uint offsetFin, ulong *length, byte *dstptr);

		//locate all the ocurrences of a word/phrase 
	int locateFacade (twcsa *wcsa, uint *sourceTextPositions,uint *sePositions, uint number);
	
		//show text around the occurrences of a word. 
	int locateAllAndDisplay (twcsa *wcsa, uint *sePositions, uint number, int radix);

		//recovers the source text by calling display (either only once or "len" times)
	void recoverSourceText1(twcsa *wcsa, char *basename, char *ext, uint sourceTextSize);
	void recoverSourceText2(twcsa *wcsa, char *basename, char *ext, uint sourceTextSize);


	//***Searching for a TEXT pattern ...
	
		//extracts the ids of the valid words of a "plain text".
	void parseTextIntoIntegers(twcsa *wcsa, byte *textPattern, uint patLen, uint *integerPattern, uint *sizeIntegers) ;
		
 		//counts the occurrences of a given text pattern.
	int countTextOcurrences(twcsa *wcsa, byte *textPattern);
	
 		//returns the offsets (to the source text) where a given text pattern appears.	
	uint *locateTextOcurrences(twcsa *wcsa, byte *textPattern, int *numberOccurrences);
	
 		//shows a snippet with the text around the ocurrences of a pattern.	
	int displayTextOcurrences(twcsa *wcsa, byte *textPattern, uint radixDisplay);


/** ***********************************************************************************
  * WORD-ORIENTED QUERY FUNCTIONS: LocateWord and DisplayWord
  * ***********************************************************************************/  
	/** Writes in numocc the number of occurrences of the substring 
	  pattern[0..length-1] in the text indexed by index. It also allocates
	  occ (which must be freed by the caller) and writes the locations of 
	  the numocc occurrences in occ, in arbitrary order. These occurrences
	  refer to the offsets in TOH where the caller could start a display
	  operation. So locateWord implies synchronization using B.
	  ** Parameter kbefore sets locateWord not to obtain the offset in TOH of the
	     searched word, but the offset in TOH of k-before words before.	  
	*/	  
	  
int locateWord(void *index, uchar *pattern, ulong length, ulong **occ, ulong *numocc, uint kbefore);

  /** Displays the text (snippet) surrounding any occurrence of the 
    substring pattern[0..length-1] within the text indexed by index. 
    The snippet must include numc characters before and after the 
    pattern occurrence, totalizing length+2*numc characters, or less if 
    the text boundaries are reached. Writes in numocc the number of 
    occurrences, and allocates the arrays snippet_text and 
    snippet_lengths (which must be freed by the caller). The first is a 
    character array of numocc*(length+2*numc) characters, with a new 
    snippet starting at every multiple of length+2*numc. The second 
    gives the real length of each of the numocc snippets. */

 int displayWords (void *index, uchar *pattern, ulong length, ulong numc, 
         ulong *numocc, uchar **snippet_text, ulong **snippet_lengths, uint kbefore);


/** simulates extration of text process, but does not actually returns anything at all 
   Extracts upto <=2K words from K=wordsbefore words before each occurrence of a pattern.
   Less than 2K words can be extracted if more than numc characters have been already obtained.
   Do nothing else... do not return the text */
int  displayTextOcurrencesNoShow(void *index, uchar *pattern, ulong length, uint wordsbefore, uint maxnumc);



// Other private functions ************************************
	// binary search algorithm... needed by display().
	int binsearchUintArray (uint *V, uint aSize, uint value);

	//Auxiliary functions

	//uint structsSizeDisk(twcsa *wcsa);
	//uint structsSizeMem(twcsa *wcsa);
	//void printInfoReduced(twcsa *wcsa);
	double getTime2 (void); 
	int saveSEfile (char *basename, uint *v, uint n);
	void printWord(uchar *str, uint len) ;
