
/* canonicalsHuff.c
 * Copyright (C) 2011, Antonio Fariña, all rights reserved.
 *
 * canonicalsHuff.c: Handling the array of canonical words (the vocabulary).
 *   compressed with canonical Huffman (char-based, bit-oriented)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */


#ifdef COMPRESS_HUFFMAN_CANONICALS  

/** *******************************************************************
  * creating the tCanonicalsHuff structure during construction of WCSA 
  * *******************************************************************/  	
int buildHuffCanonicals(tCanonicalsHuff *HC, tposInHT *posInHT, uint numberOfCanonicalWords) {
	uint maxLenAsciiCanonical =0;
	register uint i;
	uint len;
	byte *src;

	/** creating Huffman over ZoneMem (the text with all the canonical words) */
	uint freqs[256];
	for (i=0;i<256;i++) freqs[i]=0;
	for (i=0;i<numberOfCanonicalWords;i++) {
			src = posInHT[i].word;		 //calculates the len of the variant
			len=0;
			while (*src) {freqs[*src]++;src++; len++;}  
			//freqs[*src]++;  //no \0 is kept.
			if (maxLenAsciiCanonical < len) maxLenAsciiCanonical = len;
	}
	HC->nwords = numberOfCanonicalWords;
	HC->maxLenAsciiCanonical = maxLenAsciiCanonical;
	HC->decodeBuffer = (byte*) malloc((maxLenAsciiCanonical +1)*sizeof(byte));

	HC->H = createHuff (freqs,255,UNSORTED);

	/** allocating the memory needed for all the canonicals (Huffman); */
	HC->zoneMemHuff = (uint *) malloc ((HC->H.total + W -1)/W * sizeof(uint));
	HC->zoneMemHuff[(HC->H.total + W -1)/W -1] = 0000;
	HC->sizeZoneHuff = HC->H.total;
	
	/** allocating the memory for the pointers to the words **/
	HC->elemSize = bits(HC->sizeZoneHuff);
	HC->words = (uint *) malloc ( ((HC->elemSize * (HC->nwords+1)) + W -1)/W * sizeof(uint));	
	HC->words[(((HC->elemSize * (HC->nwords+1) ) + W -1)/W)  -1] = 0000;


	/** Computing the max jump that can be needed for the 2nd, 3rd, etc variants of a word
		in order to know the number of bits that have to be reserved.
		!! And generating the compressed text in zoneMemHuff (it will be repeated later).
    */
	uint ptr=0;
	uint ptrprev;

	for (i=0;i<numberOfCanonicalWords;i++) {
		src = posInHT[i].word;		//encodes the canonical word
		ptrprev=ptr;		
		while (*src) {
			ptr=encodeHuff(HC->H, *src, HC->zoneMemHuff, ptr);
			src++; 
		}  
		len=ptr-ptrprev;					            
		bitwrite (HC->words, i*(HC->elemSize), (HC->elemSize), ptrprev);		
	}
	bitwrite (HC->words, numberOfCanonicalWords*(HC->elemSize), (HC->elemSize), ptr);	/* writes ptr to end of buffer */
	
	/** set the Huffman tree to permit decoding */
	prepareToDecode(&(HC->H));
} 

int saveCanonicals (tCanonicalsHuff *HC, int file) {
	/** the max length of a variant **/
	write(file, &(HC->maxLenAsciiCanonical), sizeof(uint));	

	/** compressed canonicals */
	write(file, &(HC->sizeZoneHuff), sizeof(uint));
	write(file,  (HC->zoneMemHuff), sizeof(uint) * ((HC->sizeZoneHuff + W-1)/W) );

	/** the huffman tree */
	THuff H = HC->H;
	write(file,&H.max, sizeof(int));
	write(file,&H.lim, sizeof(int));
	write(file,&H.depth, sizeof(int));
	write(file, H.s.symb, (H.lim+1)*sizeof(int));	
	write(file, H.num, (H.depth+1)*sizeof(int));
	write(file, H.fst, (H.depth+1)*sizeof(int));

	/** ptrs to the beginning of each word-code. */
	write(file, &(HC->nwords), sizeof(uint));	
	write(file, &(HC->elemSize), sizeof(uint));
	write(file,   HC->words, ( ((HC->elemSize * (HC->nwords+1)) + W -1)/W )*sizeof(uint));
}

int loadCanonicals (tCanonicalsHuff *HC, int file) {

	/** the max length of a variant */
	read(file, &(HC->maxLenAsciiCanonical), sizeof(uint));	
	HC->decodeBuffer = (byte*) malloc((HC->maxLenAsciiCanonical +1)*sizeof(byte)); //tmp buffer

	/** compressed canonicals */
	read(file, &(HC->sizeZoneHuff), sizeof(uint));
	HC->zoneMemHuff = (uint *) malloc (sizeof(uint) * ((HC->sizeZoneHuff + W-1)/W));
	read(file, (HC->zoneMemHuff), sizeof(uint) * ((HC->sizeZoneHuff + W-1)/W) );

	/** the huffman tree */
	THuff H;
	read(file, &H.max, sizeof(int));
	read(file, &H.lim, sizeof(int));
	read(file, &H.depth, sizeof(int));
	H.s.symb = (unsigned int *) malloc((H.lim+1)*sizeof(int));
	H.num = (unsigned int *) malloc((H.depth+1)*sizeof(int));	
	H.fst = (unsigned int *) malloc((H.depth+1)*sizeof(int));	

	read(file, H.s.symb, (H.lim+1)*sizeof(int));	
	read(file, H.num, (H.depth+1)*sizeof(int));
	read(file, H.fst, (H.depth+1)*sizeof(int));	
	HC->H = H;

	/** ptrs to the beginning of each word-code. */

	read(file, &(HC->nwords), sizeof(uint));
	read(file, &(HC->elemSize), sizeof(uint));
	HC->words =(uint *) malloc ( (( (HC->elemSize * (HC->nwords+1)) +W-1)/W) * sizeof (uint));	
	read(file, HC->words, (( (HC->elemSize * (HC->nwords+1)) + W-1)/W) * sizeof(uint));

}

int freeCanonicals (tCanonicalsHuff *HC) {
	free(HC->zoneMemHuff);
	free(HC->decodeBuffer);
	free(HC->words);
	freeHuff(HC->H);
	return 0;
}

int sizeCanonicals(tCanonicalsHuff *HC) {
	return  ( (( (HC->elemSize * (HC->nwords+1)) +W-1)/W) * sizeof (uint)) + 
			( ((HC->sizeZoneHuff + W-1)/W) * sizeof(uint)) +
			( ((HC->maxLenAsciiCanonical) +1) * sizeof(byte)) +
			sizeHuff (HC->H);
}



/** lenVariant 
	returns the length of a canonical word (in characters)
	*/
int lenCanonical(tCanonicalsHuff *HC, uint i) {
	register uint len;
	uint ini,end;
	uint returnValue;
		
	ini =     i * HC->elemSize;
	end = (i+1) * HC->elemSize;
	
	ini  = bitread ( HC->words, ini, HC->elemSize);
	end  = bitread ( HC->words, end, HC->elemSize);

	{	//Decoding the whole word 
		uint canonical;
		register uint ptr;
		THuff *huff = &(HC->H);
		len=0;
		ptr = ini;	
		while (ptr < end) {
			decodeNormalHuffMacro(huff,canonical,(HC->zoneMemHuff),ptr); 
			len++;
		}		
	}
	return (len);
}

/** asciiVariant 
	returns a pointer to a decoded variant (ended in '\0')
	*/
byte * asciiCanonical(tCanonicalsHuff *HC, uint i) {
	byte *dst;
	register uint len=0;		
	uint ini,end;
	uint returnValue;
		
	ini =     i * HC->elemSize;
	end = (i+1) * HC->elemSize;	
	ini  = bitread ( HC->words, ini, HC->elemSize);
	end  = bitread ( HC->words, end, HC->elemSize);
	dst = HC->decodeBuffer;
	
	{	//Decoding the whole word 
		uint canonical;
		register uint ptr;
		THuff *huff = &(HC->H);
		len=0;
		ptr = ini;	
		while (ptr < end) {
			decodeNormalHuffMacro(huff,canonical,(HC->zoneMemHuff),ptr); 
			*dst=(byte)canonical; dst++;
			len++;
		}
		*dst='\0';		
	}
	return ((byte*) (HC->decodeBuffer));
}


/** asciiVariant 
	returns a pointer to a decoded variant (ended in '\0')
	*/
byte * asciiCanonicalLen(tCanonicalsHuff *HC, uint i, uint *length) {
	byte *dst;
	register uint len=0;		
	uint ini,end;
	uint returnValue;
		
	ini =     i * HC->elemSize;
	end = (i+1) * HC->elemSize;	
	ini  = bitread ( HC->words, ini, HC->elemSize);
	end  = bitread ( HC->words, end, HC->elemSize);
	dst = HC->decodeBuffer;
	
	{	//Decoding the whole word 
		uint canonical;
		register uint ptr;
		THuff *huff = &(HC->H);
		len=0;
		ptr = ini;	
		while (ptr < end) {
			decodeNormalHuffMacro(huff,canonical,(HC->zoneMemHuff),ptr); 
			*dst=(byte)canonical; dst++;
			len++;
		}
		*dst='\0';		
	}
	
	*length=len;
	return ((byte*) (HC->decodeBuffer));
}


//permits to compare 2 strings of len ws1 and ws2 that do not end in '\0'
int strcompHuff(tCanonicalsHuff *HC, uint i, const byte *s1, register ulong ws1) {
	register ulong iters = 0;	 
	ulong ws2=0;   

	register uint ini,end;
	ini =     i * HC->elemSize;
	end = (i+1) * HC->elemSize;	
	ini  = bitread (HC->words, ini, HC->elemSize);
	end  = bitread (HC->words, end, HC->elemSize);
	
	uint canonical;
	register uint ptr=ini;
	register byte value;
	THuff *huff = &(HC->H);

	while (ptr < end) {
		decodeNormalHuffMacro(huff,canonical,(HC->zoneMemHuff),ptr); 
		value=(byte)canonical; 		
		ws2++;
		if (!(iters<ws1 && value == *s1 )) {
			if (ptr<end) ws2=ws1+1;
			break;
		}
		s1++;
		iters++;
	}


	if (ws1 == iters) {
			if (ws2 == ws1)
				return 0;  // s1 equals to s2
			else 
				return -1;   // w1 < w2  and  *s1_i == *s2_i for i=0 to iters-1
	}
	else 
	if (ws2 == iters) {
			if (ws2 == ws1)
				return 0;  // s1 equals to s2
			else 
				return +1;   // w2 < w1  and  *ws1 is '\0'
	}
	else { //*s1 != *s2
		return (*s1-value);
	}	
	
}





/** DEBUGING ONLY****************/
void showCanonicals(tCanonicalsHuff *HC){
	uint i;
	fprintf(stderr,"\n Decoded canonicals:\n");
	for (i=0;i<HC->nwords; i++) {
		byte *dst;
		dst = asciiCanonical(HC, i);
		fprintf(stderr,"\n[Canonical word[%6u]: %s]",i, dst);
		fprintf(stderr,"[Len:%d]", lenCanonical(HC, i)) ;
	}
	fflush(stderr);
}


#endif
